
(* 
   		Lemme de Dixon:
		dans toute suite infinie de monomes, il existe un monome qui
		divise un de ses successeurs.

		[Dixon's Lemma:
		in every infinite sequence of monomials, there exists a
		monomial dividing one of its successors.
		]

Loic Pottier, 25 avril 96.
Modifi le 21 juillet 98.

*)
Require Export Nat_Sequences.
Require Export Monomials.

(* 	
			Sequences of monomials.
*)

Definition sequence_mon := [d:nat] nat ->(mon d).
Section monseqdef.
Variable d:nat.

Record monseq: Type := {
   mseq:> nat -> (mon d) ->Prop;
   mtotal: (n:nat)(ex (mon d) [m:(mon d)](mseq n m));
   mfunct: (n:nat) (m, m':(mon d)) (mseq n m) -> (mseq n m') ->m=m' }.
End monseqdef.

Definition proj1_sequence_mon: (d:nat) (monseq d) ->natseq.
Intros d u.
Apply (Build_natseq
         [n, m:nat](ex (mon d) [m':(mon d)](u n m') /\ (pmon1 d m')=m)).
Intros.
Case (mtotal d u n).
Intros m' H.
Exists (pmon1 d m').
Exists m'.
Auto.
Intros.
Case H; Clear H.
Case H0; Clear H0.
Intros m1 H m2 H0.
Case H; Clear H; Intros.
Case H0; Clear H0; Intros.
Generalize (mfunct d u n m1 m2).
Intros.
Rewrite (H3 H H0) in H1.
Rewrite <- H1; Rewrite <- H2; Auto.
Defined.

Definition proj2_sequence_mon: (d:nat) (monseq d) ->(monseq (pred d)).
Intros d u.
Apply (Build_monseq
         (pred d)
         [n:nat] [m2:(mon (pred d))]
         (ex (mon d) [m1:(mon d)](u n m1) /\ (pmon2 d m1)=m2)).
Intros n.
Case (mtotal d u n).
Intros m1 H.
Exists (pmon2 d m1).
Exists m1.
Auto.
Intros.
Case H; Clear H.
Case H0; Clear H0.
Intros m1 H m2 H0.
Case H; Clear H; Intros.
Case H0; Clear H0; Intros.
Generalize (mfunct d u n m1 m2).
Intros.
Rewrite (H3 H H0) in H1.
Rewrite <- H1; Rewrite <- H2; Auto.
Defined.

Lemma mnil_unique:
  (d:nat) (m:(mon d)) d=O ->
  (existS nat [d:nat](mon d) d m)=(existS nat [d:nat](mon d) O n_0).
Induction m.
Auto.
Intros.
Inversion H0.
Qed.
Hints Resolve mnil_unique.

Definition mcompose: (d:nat) (monseq d) -> natseq ->(monseq d).
Intros d f u.
Apply (Build_monseq d [n:nat] [m:(mon d)](ex nat [p:nat](u n p) /\ (f p m))).
Intros.
Case (total u n).
Intros p H.
Case (mtotal d f p).
Intros m H0.
Exists m.
Exists p; Auto.
Intros.
Case H; Clear H.
Case H0; Clear H0.
Intros m1 H m2 H0.
Case H; Clear H; Intros.
Case H0; Clear H0; Intros.
Generalize (funct u n m1 m2).
Intros.
Rewrite (H3 H H0) in H1.
Apply (mfunct d f m2); Auto.
Defined.
(*
		Every sequence has a sub_sequence with not_decreasing first
		coordinate. 
*)

Lemma Dixon1:
  (d:nat) (f:(monseq d))
  (exT
     natseq
     [u:natseq]
     (sincr u) /\
     ((p, q:nat) (lt p q) ->
      (m, m':(mon d)) ((mcompose d f u) p m) -> ((mcompose d f u) q m') ->
      (le (pmon1 d m) (pmon1 d m')))).
Intros.
Exists (s1seq (proj1_sequence_mon d f)).
Split.
Apply s1seq_sincr.
Generalize (s2seq_incr (proj1_sequence_mon d f)).
Unfold incr.
Simpl.
Intros.
Apply H with p q.
Auto with arith.
Case H1; Clear H1.
Case H2; Clear H2.
Intros b H1 b' H2.
Case H1; Clear H1; Intros.
Case H2; Clear H2; Intros.
Exists b'.
Split; Auto.
Exists m.
Auto.
Case H1; Clear H1.
Case H2; Clear H2.
Intros b H1 b' H2.
Case H1; Clear H1; Intros.
Case H2; Clear H2; Intros.
Exists b.
Split; Auto.
Exists m'.
Auto.
Qed.
Require Import Eqdep.
(*
		Every sequence has a sub_sequence not_decreasing for the
		division. 
*)

Definition idnatseq: natseq.
Apply (Build_natseq [n, m:nat]n=m).
Intros.
Exists n; Auto.
Intros.
Rewrite <- H; Rewrite <- H0; Auto.
Defined.

Lemma compose_sincr:
  (u, v:natseq) (sincr u) -> (sincr v) ->(sincr (compose u v)).
Unfold sincr compose; Simpl.
Intros.
Case H2; Clear H2; Intros.
Case H2; Clear H2; Intros.
Case H3; Clear H3; Intros.
Case H3; Clear H3; Intros.
Apply H with x x0; Auto.
Apply H0 with n m; Auto.
Qed.

Theorem Dixon:
  (d:nat) (f:(monseq d))
  (exT
     natseq
     [h:natseq]
     (sincr h) /\
     ((p, q:nat) (lt p q) ->
      (m, m':(mon d)) ((mcompose d f h) p m) -> ((mcompose d f h) q m') ->
      (mdiv d m m'))).
Induction d.
Intro.
Exists idnatseq.
Split.
Unfold sincr.
Simpl.
Intros.
Rewrite H0 in H; Rewrite H1 in H; Auto.
Simpl.
Intros.
Cut (existS nat [d:nat](mon d) O m)=(existS nat [d:nat](mon d) O n_0); Auto.
Cut (existS nat [d:nat](mon d) O m')=(existS nat [d:nat](mon d) O n_0); Auto.
Intros.
Rewrite (inj_pair2 nat [x:nat](mon x) O m n_0); Auto.
Rewrite (inj_pair2 nat [x:nat](mon x) O m' n_0); Auto.
Apply mdiv_nil; Auto.
Intros.
Elim (Dixon1 (S n) f); Auto.
Intros g H0.
Case H0; Clear H0; Intros.
Elim (H (proj2_sequence_mon (S n) (mcompose (S n) f g))).
Clear H.
Intros g1 H.
Case H; Clear H; Intros H2 H.
Exists (compose g g1).
Split.
Apply compose_sincr; Auto.
Intros.
Apply mdiv_proj; Auto.
Case (total g1 p).
Case (total g1 q).
Intros q1 h6 p1 H7.
Apply (H1 p1 q1); Auto.
Unfold sincr in H2.
Apply H2 with p q; Auto.
Simpl in H4.
Simpl.
Case H4; Clear H4; Intros.
Case H4; Clear H4; Intros.
Case H4; Clear H4; Intros.
Case H4; Clear H4; Intros.
Exists x; Auto.
Split; Auto.
Replace p1 with x0; Auto.
Apply (funct g1 p); Auto.
Simpl in H5.
Case H5; Clear H5; Intros.
Case H5; Clear H5; Intros.
Case H5; Clear H5; Intros.
Case H5; Clear H5; Intros.
Simpl.
Exists x.
Split; Auto.
Replace q1 with x0; Auto.
Apply (funct g1 q); Auto.
Simpl.
Apply (H p q); Auto.
Simpl.
Simpl in H4.
Case H4; Clear H4; Intros.
Case H4; Clear H4; Intros.
Case H4; Clear H4; Intros.
Case H4; Clear H4; Intros.
Exists x0.
Split; Auto.
Exists m.
Split; Auto.
Exists x; Auto.
Simpl.
Simpl in H5.
Case H5; Clear H5; Intros.
Case H5; Clear H5; Intros.
Case H5; Clear H5; Intros.
Case H5; Clear H5; Intros.
Exists x0.
Split; Auto.
Exists m'.
Split; Auto.
Exists x; Auto.
Qed.
(*
		In every sequence of monomials, there exists a
		monomial dividing one of its successors.
*)

Definition fun_monseq: (d:nat) (nat ->(mon d)) ->(monseq d).
Intros d f.
Apply (Build_monseq d [n:nat] [m:(mon d)]m=(f n)).
Intros.
Exists (f n); Auto.
Intros.
Rewrite H; Rewrite H0; Auto.
Defined.

Theorem Dixon2:
  (d:nat) (f:nat ->(mon d))
  (Ex [i:nat] (Ex [j:nat] (lt i j) /\ (mdiv d (f i) (f j)))).
Intros.
Elim (Dixon d (fun_monseq d f)).
Intros g H.
Case H; Clear H; Intros.
Case (total g O).
Case (total g (S O)).
Intros j H1 i H2.
Exists i.
Exists j.
Split.
Apply H with O (1); Auto.
Apply H0 with O (1); Auto.
Simpl.
Exists i; Auto.
Simpl.
Exists j; Auto.
Qed.

