
(****************************************************************************
                                                                           
          Buchberger : the monomials                                       
                                                                           
          Laurent Thery 	                                             
                                                                           
****************************************************************************)
Section Monomials.
Require Import Arith.
Require Import Compare.
Require Import Compare_dec.
Require Import Peano_dec.

Mutual Inductive mon: nat ->Set :=
     n_0: (mon O)
    | c_n: (d:nat) nat -> (mon d) ->(mon (S d)) .

Definition pmon1: (d:nat) (mon d) ->nat.
Realizer [_:nat] [m:mon](<nat>Case m of O [_:nat] [n:nat] [_:mon]n end).
Program_all.
Defined.

Definition pmon2: (d:nat) (mon d) ->(mon (pred d)).
Realizer [_:nat] [m:mon](<mon>Case m of n_0 [_:nat] [_:nat] [m':mon]m' end).
Program_all.
Defined.

Definition recomp: (d:nat) ~ d=O -> (mon d) ->(mon d).
Realizer [d:nat] [m:mon]
         (<mon>Case d of
           (False_rec mon) [d':nat](c_n d' (pmon1 (S d') m) (pmon2 (S d') m))
           end).
Program_all.
Defined.

Lemma recomp_ok: (d:nat) (h:~ d=O) (m:(mon d))(recomp d h m)=m.
Intros d h m.
Generalize h; Clear h.
Elim m.
Intros h; Elim h; Auto.
Intros; Simpl; Unfold recomp; Auto.
Qed.

Lemma proj_ok: (d:nat) (m:(mon (S d)))(c_n d (pmon1 (S d) m) (pmon2 (S d) m))=m.
Intros d m; Cut ~ (S d)=O; Auto.
Intros H; Cut (recomp (S d) H m)=m; Auto.
Apply recomp_ok; Auto.
Qed.
(*Generator
*)

Definition gen_mon: (d:nat) nat ->(mon d).
Intros d; Elim d.
Intros n; Exact n_0.
Intros n H' n'; Case n'.
Exact (c_n n (S O) (H' n)).
Intros n''; Exact (c_n n O (H' n'')).
Defined.
(*
	Multiplication of monomials.
*)

Definition mult_mon: (d:nat) (mon d) -> (mon d) ->(mon d).
Intros d; Elim d.
Intros H' H'0; Exact n_0.
Intros n Rec S1 S2.
Exact (c_n
         n (plus (pmon1 (S n) S1) (pmon1 (S n) S2))
         (Rec (pmon2 (S n) S1) (pmon2 (S n) S2))).
Defined.

Theorem mult_mon_com: (d:nat) (a, b:(mon d))(mult_mon d a b)=(mult_mon d b a).
Intros d; Elim d; Simpl; Auto.
Intros n H' a b.
Rewrite (H' (pmon2 (S n) a) (pmon2 (S n) b)); Auto.
Rewrite (plus_sym (pmon1 (S n) a) (pmon1 (S n) b)); Auto.
Qed.

Theorem mult_mon_assoc:
  (d:nat) (a, b, c:(mon d))
  (mult_mon d a (mult_mon d b c))=(mult_mon d (mult_mon d a b) c).
Intros d; Elim d; Simpl; Auto.
Intros n H' a b c.
Rewrite (H' (pmon2 (S n) a) (pmon2 (S n) b) (pmon2 (S n) c)); Auto.
Rewrite (plus_assoc_l (pmon1 (S n) a) (pmon1 (S n) b) (pmon1 (S n) c)); Auto.
Qed.
(*
Zero
*)

Definition zero_mon: (d:nat)(mon d).
Intros d; Elim d.
Exact n_0.
Intros n Rec; Exact (c_n n O Rec).
Defined.

Theorem mult_mon_zero_r: (d:nat) (a:(mon d))(mult_mon d a (zero_mon d))=a.
Intros d a; Elim a; Auto.
Simpl.
Intros d0 n m H'; Rewrite H'; Auto.
Rewrite (plus_sym n O); Auto.
Qed.

Theorem mult_mon_zero_l: (d:nat) (a:(mon d))(mult_mon d (zero_mon d) a)=a.
Intros d a; Elim a; Auto.
Simpl.
Intros d0 n m H'; Rewrite H'; Auto.
Qed.
(* 
			Division of monomials.
*)

Mutual Inductive mdiv: (d:nat) (mon d) -> (mon d) ->Prop :=
     mdiv_nil: (mdiv O n_0 n_0)
    | mdiv_cons:
        (d:nat) (v, v':(mon d)) (n, n':nat) (le n n') -> (mdiv d v v') ->
        (mdiv (S d) (c_n d n v) (c_n d n' v')) .
Hints Resolve mdiv_nil mdiv_cons.

Lemma mdiv_proj:
  (d:nat)
  (m, m':(mon (S d)))
  (le (pmon1 (S d) m) (pmon1 (S d) m')) ->
  (mdiv d (pmon2 (S d) m) (pmon2 (S d) m')) ->(mdiv (S d) m m').
Intros d m m' H' H'0; Rewrite <- (proj_ok d m); Rewrite <- (proj_ok d m'); Auto.
Qed.
Require Import Relation_Definitions.
Require Import Eqdep.
(*
		Division is transitive.
*)

Lemma mdiv_trans: (d:nat)(transitive (mon d) (mdiv d)).
Intros d; Elim d; Unfold transitive; Auto.
Intros x y z H'; Inversion H'.
Intros H0; Inversion H0; Auto.
Rewrite <- (inj_pair2 nat [x:nat](mon x) O n_0 z); Auto.
Rewrite <- (inj_pair2 nat [x:nat](mon x) O n_0 x); Auto.
Intros n Rec x y z H'; Inversion H'.
Rewrite <- (inj_pair2 nat [x:nat](mon x) (S n) (c_n n n' v') y); Auto.
Rewrite <- (inj_pair2 nat [x:nat](mon x) (S n) (c_n n n0 v) x); Auto.
Intros H'0; Inversion H'0; Auto.
Rewrite <- (inj_pair2 nat [x:nat](mon x) (S n) (c_n n n'0 v'0) z); Auto.
Apply mdiv_cons; Auto.
Apply le_trans with m := n'; Auto.
Apply Rec with y := v0; Auto.
Rewrite (inj_pair2 nat [x:nat](mon x) n v0 v'); Auto.
Qed.
(*Division  of monomials (total function!).
*)

Definition div_mon: (d:nat) (mon d) -> (mon d) ->(mon d).
Intros d; Elim d.
Intros H' H'0; Exact n_0.
Intros n Rec S1 S2.
Exact (c_n
         n (minus (pmon1 (S n) S1) (pmon1 (S n) S2))
         (Rec (pmon2 (S n) S1) (pmon2 (S n) S2))).
Defined.

Theorem mdiv_div:
  (d:nat) (a, b:(mon d)) (mdiv d b a) ->(mult_mon d (div_mon d a b) b)=a.
Intros d a b H'; Elim H'; Simpl; Auto.
Intros d0 v v' n n' H'0 H'1 H'2.
Rewrite (plus_sym (minus n' n) n).
Rewrite <- (le_plus_minus n n'); Auto.
Rewrite H'2; Auto.
Qed.
(*Division  of monomials (total function!).
*)

Definition div_mon_clean: (d:nat) (mon d) -> (mon d) ->(mon d) * bool.
Intros d; Elim d.
Intros H' H'0; Exact (n_0, true).
Intros n Rec S1 S2.
Case (le_lt_dec (pmon1 (S n) S2) (pmon1 (S n) S1)).
Case (Rec (pmon2 (S n) S1) (pmon2 (S n) S2)).
Intros res b H'1;
 Exact ((c_n n (minus (pmon1 (S n) S1) (pmon1 (S n) S2)) res), b).
Intros H'; Exact (S1, false).
Defined.

Definition is_nil: (d:nat) (mon d) ->(mon d).
Intros d; Case d.
Intros H'; Exact n_0.
Intros n H'; Exact H'.
Defined.

Theorem is_nil_id: (d:nat) (a:(mon d))a=(is_nil d a).
Intros d a; Elim a; Simpl; Auto.
Qed.

Theorem mon_0: (a:(mon O))a=n_0.
Intros a; Generalize (is_nil_id O a); Simpl; Auto.
Qed.
Hints Resolve mon_0.

Theorem eqmon_dec: (d:nat) (x, y:(mon d)){x=y}+{~ x=y}.
Intros d; Elim d; Auto.
Intros x y; Left.
Rewrite (mon_0 x); Rewrite (mon_0 y); Auto.
Intros n H' x y.
Elim (eq_nat_dec (pmon1 (S n) x) (pmon1 (S n) y)); Intros H'2.
Elim (H' (pmon2 (S n) x) (pmon2 (S n) y)); Intros H'3.
Left.
Rewrite <- (proj_ok n x).
Rewrite <- (proj_ok n y).
Rewrite H'3; Rewrite H'2; Auto.
Right; Rewrite <- (proj_ok n x); Rewrite <- (proj_ok n y); Red; Intros H'0.
Apply H'3.
Change <(mon n)>
         (pmon2 (S n) (c_n n (pmon1 (S n) x) (pmon2 (S n) x)))=
         (pmon2 (S n) (c_n n (pmon1 (S n) y) (pmon2 (S n) y))).
Rewrite H'0; Auto.
Right; Red; Rewrite <- (proj_ok n x); Rewrite <- (proj_ok n y); Intros H'0;
 Apply H'2; Injection H'0; Auto.
Qed.

Theorem mult_div_com: (d:nat) (a, b:(mon d))(div_mon d (mult_mon d a b) b)=a.
Intros d; Elim d; Simpl; Auto.
Intros n H' a b.
Rewrite (H' (pmon2 (S n) a) (pmon2 (S n) b)); Auto.
Rewrite (plus_sym (pmon1 (S n) a) (pmon1 (S n) b)).
Rewrite (minus_plus (pmon1 (S n) b) (pmon1 (S n) a)); Auto.
Apply proj_ok; Auto.
Qed.

Theorem mult_div_id: (d:nat) (a:(mon d))(div_mon d a a)=(zero_mon d).
Intros d a; Elim a; Simpl; Auto.
Intros d0 n m H'; Rewrite H'; Auto.
Rewrite <- (minus_n_n n); Auto.
Qed.

Local gb: (d:nat) (mon d) * bool ->bool.
Intros d H'; Case H'; Auto.
Defined.

Local gm: (d:nat) (mon d) * bool ->(mon d).
Intros d H'; Case H'; Auto.
Defined.

Theorem minus_lt_0: (m, n:nat) (lt n m) ->(minus n m)=O.
Intros m; Elim m; Simpl; Auto.
Intros n H; Absurd (lt n O); Auto with arith.
Intros n H' n0; Case n0; Simpl; Auto with arith.
Qed.

Theorem div_clean_dec2:
  (d:nat) (a, b:(mon d)) (gb d (div_mon_clean d a b))=false ->
  ~ (mult_mon d (div_mon d a b) b)=a.
Intros d; Elim d; Simpl; Auto.
Intros a H' H'0; Inversion H'0; Auto.
Intros n H' a b.
Case (le_lt_dec (pmon1 (S n) b) (pmon1 (S n) a)); Simpl; Auto.
Intros l; Generalize (H' (pmon2 (S n) a) (pmon2 (S n) b)).
Case (div_mon_clean n (pmon2 (S n) a) (pmon2 (S n) b)); Simpl; Auto.
Intros H'0 b0 H'1 H'2; Red; Intros H'3.
LApply H'1; [Intros H'4; Apply H'4; Clear H'1 | Clear H'1]; Auto.
Change <(mon n)>
         (pmon2
            (S n)
            (c_n
               n (plus (minus (pmon1 (S n) a) (pmon1 (S n) b)) (pmon1 (S n) b))
               (mult_mon
                  n (div_mon n (pmon2 (S n) a) (pmon2 (S n) b)) (pmon2 (S n) b))))=
         (pmon2 (S n) a).
Rewrite H'3; Auto.
Intros H'0 H'1; Red; Intros H'2;
 Absurd (plus (minus (pmon1 (S n) a) (pmon1 (S n) b)) (pmon1 (S n) b))=
        (pmon1 (S n) a); Auto.
2:Change <nat>
           (pmon1
              (S n)
              (c_n
                 n
                 (plus (minus (pmon1 (S n) a) (pmon1 (S n) b)) (pmon1 (S n) b))
                 (mult_mon
                    n (div_mon n (pmon2 (S n) a) (pmon2 (S n) b))
                    (pmon2 (S n) b))))=(pmon1 (S n) a).
2:Rewrite H'2; Auto.
Rewrite (minus_lt_0 (pmon1 (S n) b) (pmon1 (S n) a)); Simpl; Auto.
Red; Intros H'3; Absurd (lt (pmon1 (S n) a) (pmon1 (S n) b)); Auto.
Apply le_not_lt; Auto.
Rewrite H'3; Auto.
Qed.

Theorem div_clean_dec1:
  (d:nat) (a, b:(mon d)) (gb d (div_mon_clean d a b))=true ->
  (gm d (div_mon_clean d a b))=(div_mon d a b) /\
  (mult_mon d (div_mon d a b) b)=a.
Intros d; Elim d; Simpl; Auto.
Intros n H' a b; Case (le_lt_dec (pmon1 (S n) b) (pmon1 (S n) a)); Simpl; Auto.
2:Intros H'0 H'1; Inversion H'1.
Intros l; Generalize (H' (pmon2 (S n) a) (pmon2 (S n) b)).
Case (div_mon_clean n (pmon2 (S n) a) (pmon2 (S n) b)); Simpl; Auto.
Intuition.
Rewrite H1; Auto.
Rewrite H2; Auto.
Rewrite (plus_sym (minus (pmon1 (S n) a) (pmon1 (S n) b)) (pmon1 (S n) b)).
Rewrite <- (le_plus_minus (pmon1 (S n) b) (pmon1 (S n) a)); Auto.
Try Exact (proj_ok n a).
Qed.
Require Import Max.

Definition ppcm_mon: (d:nat) (mon d) -> (mon d) ->(mon d).
Intros d; Elim d.
Intros m1 m2; Exact n_0.
Intros n Rec S1 S2.
Exact (c_n
         n (max (pmon1 (S n) S1) (pmon1 (S n) S2))
         (Rec (pmon2 (S n) S1) (pmon2 (S n) S2))).
Defined.

Theorem ppcm_com: (d:nat) (a, b:(mon d))(ppcm_mon d a b)=(ppcm_mon d b a).
Intros d; Elim d; Simpl; Auto.
Intros n H' a b; Rewrite (H' (pmon2 (S n) a) (pmon2 (S n) b)).
Rewrite (max_sym (pmon1 (S n) a) (pmon1 (S n) b)); Auto.
Qed.

Theorem ppcm_prop_l:
  (d:nat) (a, b:(mon d))
  (ppcm_mon d a b)=(mult_mon d (div_mon d (ppcm_mon d a b) a) a).
Intros d; Elim d; Simpl; Auto.
Intros n H' a b; Pattern 1 (ppcm_mon n (pmon2 (S n) a) (pmon2 (S n) b));
 Rewrite (H' (pmon2 (S n) a) (pmon2 (S n) b)).
Pattern 1 (max (pmon1 (S n) a) (pmon1 (S n) b));
 Rewrite (le_plus_minus (pmon1 (S n) a) (max (pmon1 (S n) a) (pmon1 (S n) b)));
 Auto.
Rewrite (plus_sym
           (pmon1 (S n) a)
           (minus (max (pmon1 (S n) a) (pmon1 (S n) b)) (pmon1 (S n) a))); Auto.
Apply max_intro_l; Auto.
Qed.

Theorem ppcm_prop_r:
  (d:nat) (a, b:(mon d))
  (ppcm_mon d a b)=(mult_mon d (div_mon d (ppcm_mon d a b) b) b).
Intros d; Elim d; Simpl; Auto.
Intros n H' a b; Pattern 1 (ppcm_mon n (pmon2 (S n) a) (pmon2 (S n) b));
 Rewrite (H' (pmon2 (S n) a) (pmon2 (S n) b)).
Pattern 1 (max (pmon1 (S n) a) (pmon1 (S n) b));
 Rewrite (le_plus_minus (pmon1 (S n) b) (max (pmon1 (S n) a) (pmon1 (S n) b)));
 Auto.
Rewrite (plus_sym
           (pmon1 (S n) b)
           (minus (max (pmon1 (S n) a) (pmon1 (S n) b)) (pmon1 (S n) b))); Auto.
Apply max_intro_r; Auto.
Qed.

Theorem plus_minus_le: (a, b:nat) (plus (minus a b) b)=a ->(le b a).
Intros a; Elim a; Simpl; Auto.
Intros b H'; Rewrite H'; Auto.
Intros n H' b; Case b; Simpl; Auto with arith.
Intros n0; Rewrite (plus_sym (minus n n0) (S n0)); Simpl; Auto.
Rewrite (plus_sym n0 (minus n n0)); Auto with arith.
Qed.

Theorem ppcm_mom_is_ppcm:
  (d:nat)
  (a, b, c:(mon d))
  c=(mult_mon d (div_mon d c a) a) -> c=(mult_mon d (div_mon d c b) b) ->
  c=(mult_mon d (div_mon d c (ppcm_mon d a b)) (ppcm_mon d a b)).
Intros d; Elim d; Simpl; Auto.
Intros n H' a b c H'0 H'1.
Rewrite <- (H' (pmon2 (S n) a) (pmon2 (S n) b) (pmon2 (S n) c)); Auto.
2:Change <(mon n)>
           (pmon2 (S n) c)=
           (pmon2
              (S n)
              (c_n
                 n
                 (plus (minus (pmon1 (S n) c) (pmon1 (S n) a)) (pmon1 (S n) a))
                 (mult_mon
                    n (div_mon n (pmon2 (S n) c) (pmon2 (S n) a))
                    (pmon2 (S n) a)))); Auto.
2:Rewrite <- H'0; Auto.
2:Change <(mon n)>
           (pmon2 (S n) c)=
           (pmon2
              (S n)
              (c_n
                 n
                 (plus (minus (pmon1 (S n) c) (pmon1 (S n) b)) (pmon1 (S n) b))
                 (mult_mon
                    n (div_mon n (pmon2 (S n) c) (pmon2 (S n) b))
                    (pmon2 (S n) b)))); Auto.
2:Rewrite <- H'1; Auto.
Rewrite (plus_sym
           (minus (pmon1 (S n) c) (max (pmon1 (S n) a) (pmon1 (S n) b)))
           (max (pmon1 (S n) a) (pmon1 (S n) b))).
Rewrite <- (le_plus_minus (max (pmon1 (S n) a) (pmon1 (S n) b)) (pmon1 (S n) c));
 Auto.
Rewrite <- (proj_ok n c); Auto.
Apply max_is_lub; Auto.
Apply plus_minus_le; Auto.
Change <nat>
         (pmon1
            (S n)
            (c_n
               n (plus (minus (pmon1 (S n) c) (pmon1 (S n) a)) (pmon1 (S n) a))
               (mult_mon
                  n (div_mon n (pmon2 (S n) c) (pmon2 (S n) a)) (pmon2 (S n) a))))=
         (pmon1 (S n) c); Auto.
Rewrite <- H'0; Auto.
Apply plus_minus_le; Auto.
Change <nat>
         (pmon1
            (S n)
            (c_n
               n (plus (minus (pmon1 (S n) c) (pmon1 (S n) b)) (pmon1 (S n) b))
               (mult_mon
                  n (div_mon n (pmon2 (S n) c) (pmon2 (S n) b)) (pmon2 (S n) b))))=
         (pmon1 (S n) c); Auto.
Rewrite <- H'1; Auto.
Qed.
End Monomials.

