
(* 
		Suites d'entiers naturels.

*)
Require Export Classical.
Require Export Arith.
Require Export Compare.
Require Export Compare_dec.

Record natseq: Type := {
   seq:> nat -> nat ->Prop;
   total: (n:nat)(ex nat [m:nat](seq n m));
   funct: (n, m, m':nat) (seq n m) -> (seq n m') ->m=m' }.

Lemma not_imp: (P, Q:Prop) ~ (P ->Q) ->P /\ ~ Q.
Intros P Q.
Apply NNPP.
Intuition.
Qed.

Lemma not_le_lt: (x, y:nat) ~ (le x y) ->(lt y x).
Intros.
Case (le_lt_dec x y); Intros.
Absurd (le x y); Auto.
Auto.
Qed.

Lemma ex_bound_subset:
  (P:nat ->Prop) ((ex nat [a:nat](P a))) ->
  (ex nat [b:nat](P b) /\ ((n:nat) (P n) ->(le b n))).
Intros.
Case H; Clear H.
Intros a.
Generalize (well_founded_ind
              nat lt lt_wf
              [a:nat] (P a) ->
              (ex nat [b:nat](P b) /\ ((n:nat) (P n) ->(le b n)))).
Intros.
Apply H with a; Auto.
Intros.
Case (classic (n:nat) (P n) ->(le x n)); Intros.
Exists x; Auto.
Cut (ex nat [n:nat]~ ((P n) ->(le x n))).
Intro.
Case H4; Clear H4.
Intros y H4.
Cut (P y).
Cut (lt y x).
Intros.
Case (H1 y); Auto.
Intros b H7.
Exists b; Auto.
Generalize (not_imp (P y) (le x y)).
Intros.
Case H5; Auto.
Intros.
Apply not_le_lt; Auto.
Generalize (not_imp (P y) (le x y)).
Intros.
Case H5; Auto.
Apply not_all_ex_not with P := [n:nat] (P n) ->(le x n).
Auto.
Qed.

Definition in_image := [u:natseq] [a:nat](ex nat [n:nat](u n a)).

Lemma image_inhabited: (u:natseq)(ex nat [a:nat](in_image u a)).
Intro.
Generalize (total u O).
Intro.
Case H.
Intros a H0.
Exists a.
Unfold in_image.
Exists O; Auto.
Qed.

Definition index_of_lbound :=
   [u:natseq] [p:nat] [m:nat]
   (ex
      nat
      [b:nat]
      (le p m) /\ ((u m b) /\ ((m', a:nat) (le p m') -> (u m' a) ->(le b a)))).

Definition shift: natseq -> nat ->natseq.
Intros u p.
Apply (Build_natseq [n, a:nat](u (plus p n) a)).
Case u; Intros.
Simpl.
Auto.
Case u; Simpl; Intros.
Apply funct0 with (plus p n); Auto.
Defined.

Lemma ex_index_of_lbound:
  (u:natseq) (p:nat)(ex nat [m:nat](index_of_lbound u p m)).
Intros.
Generalize (ex_bound_subset
              (in_image (shift u p)) (image_inhabited (shift u p))).
Unfold in_image index_of_lbound.
Simpl.
Intros.
Case H; Clear H.
Intros b H.
Case H; Clear H.
Intros.
Case H; Clear H.
Intros m H1.
Exists (plus p m).
Exists b.
Split; Auto with arith.
Split; Auto.
Intros.
Apply H0.
Exists (minus m' p); Auto.
Replace (plus p (minus m' p)) with m'; Auto with arith.
Qed.

Lemma ex_lbound:
  (u:natseq) (p:nat)
  (ex
     nat
     [m:nat]
     (ex
        nat
        [b:nat]
        (le p m) /\
        ((u m b) /\
         (((m', a:nat) (le p m') -> (u m' a) ->(le b a)) /\
          ((m', a:nat) (le p m') -> (u m' a) -> (lt m' m) ->(lt b a)))))).
Intros u p.
Generalize (ex_bound_subset (index_of_lbound u p) (ex_index_of_lbound u p)).
Unfold index_of_lbound; Simpl.
Intros.
Case H; Clear H.
Intros m H.
Case H; Clear H.
Intros.
Case H; Clear H.
Intros b H.
Exists m.
Exists b.
Case H; Clear H.
Intros.
Case H1; Clear H1.
Intros.
Split; Auto.
Split; Auto.
Split; Auto.
Intros.
Cut (le b a).
Intro.
Cut ~ b=a.
Intro.
Generalize (le_lt_or_eq b a); Intros.
Case H8.
Auto.
Auto.
Intros.
Absurd b=a; Auto.
Unfold not; Intros.
Absurd (le m m'); Auto with arith.
Apply H0.
Exists a.
Split; Auto.
Split; Auto.
Rewrite <- H7.
Auto.
Apply H2 with m'; Auto.
Qed.

Mutual Inductive s1[u:natseq]: nat -> nat ->Prop :=
     s11:
       (m, b:nat)
       (u m b) ->
       ((m', a:nat) (u m' a) ->(le b a)) ->
       ((m', a:nat) (u m' a) -> (lt m' m) ->(lt b a)) ->(s1 u O m)
    | s12:
        (n, m1:nat) (s1 u n m1) ->
        (m, b:nat)
        (lt m1 m) ->
        (u m b) ->
        ((m', a:nat) (lt m1 m') -> (u m' a) ->(le b a)) ->
        ((m', a:nat) (lt m1 m') -> (u m' a) -> (lt m' m) ->(lt b a)) ->
        (s1 u (S n) m) .

Definition s1seq: natseq ->natseq.
Intros u.
Apply (Build_natseq (s1 u)).
Induction n.
Generalize (ex_lbound u O).
Intros.
Case H; Clear H.
Intros m H.
Case H; Clear H.
Intros b H.
Exists m.
Case H; Clear H; Intros.
Case H0; Clear H0; Intros.
Case H1; Clear H1; Intros.
Apply s11 with b; Auto with arith.
Intros.
Apply H1 with m'; Auto with arith.
Intros.
Apply H2 with m'; Auto with arith.
Intros.
Case H; Clear H.
Intros m1 H.
Generalize (ex_lbound u (S m1)).
Intros.
Case H0; Clear H0.
Intros m H0.
Case H0; Clear H0.
Intros b H0.
Case H0; Clear H0; Intros.
Case H1; Clear H1; Intros.
Case H2; Clear H2; Intros.
Exists m.
Apply s12 with m1 b; Auto.
Induction n.
Intros.
Inversion H.
Inversion H0.
Case (lt_eq_lt_dec m m'); Intros.
Case s; Intros.
Cut (lt b0 b).
Intro.
Cut (le b b0).
Intro.
Absurd (lt b0 b); Auto with arith.
Apply (H2 m'); Auto.
Apply (H8 m); Auto.
Auto.
Cut (lt b b0).
Cut (le b0 b).
Intros.
Absurd (lt b b0); Auto with arith.
Apply (H7 m); Auto.
Apply (H3 m'); Auto.
Intros.
Inversion H0.
Inversion H1.
Case (lt_eq_lt_dec m m'); Intros.
Case s; Intros.
Cut (lt b0 b).
Cut (le b b0).
Intros.
Absurd (lt b0 b); Auto with arith.
Apply H6 with m'; Auto.
Replace m1 with m2; Auto.
Apply H14 with m; Auto.
Replace m2 with m1; Auto.
Auto.
Cut (lt b b0).
Cut (le b0 b).
Intros.
Absurd (lt b b0); Auto with arith.
Apply H13 with m; Auto.
Replace m2 with m1; Auto.
Apply H7 with m'; Auto.
Replace m1 with m2; Auto.
Defined.

Definition sincr :=
   [u:natseq] (n, m:nat) (lt n m) ->(a, b:nat) (u n a) -> (u m b) ->(lt a b).

Lemma s1seq_sincr1:
  (u:natseq) (n:nat) (a, b:nat) (s1 u n a) -> (s1 u (S n) b) ->(lt a b).
Intros.
Inversion H0.
Replace a with m1.
Auto.
Apply (funct (s1seq u)) with n.
Simpl.
Auto.
Simpl.
Auto.
Qed.

Theorem s1seq_sincr: (u:natseq)(sincr (s1seq u)).
Intros u.
Unfold sincr.
Simpl.
Intros n m.
Induction 1.
Intros.
Apply s1seq_sincr1 with u n; Auto.
Intros.
Generalize (total (s1seq u) m0).
Intro.
Case H4; Clear H4.
Intros c H4.
Simpl in H4.
Apply lt_trans with c; Auto.
Apply s1seq_sincr1 with u m0; Auto.
Qed.

Definition eq_natseq := [u, v:natseq] (n, a:nat)(iff (u n a) (v n a)).

Definition compose: natseq -> natseq ->natseq.
Intros u v.
Apply (Build_natseq [n, a:nat](ex nat [b:nat](u b a) /\ (v n b))).
Intros.
Generalize (total v n).
Intro.
Case H; Intros b H0.
Generalize (total u b).
Intro.
Case H1; Intros m H2.
Exists m.
Exists b; Auto.
Intros.
Case H; Intros.
Case H0; Intros.
Case H1; Case H2; Intros.
Apply (funct u) with x; Auto.
Replace x with x0; Auto.
Apply (funct v) with n; Auto.
Defined.

Definition s2seq := [u:natseq](compose u (s1seq u)).

Definition incr :=
   [u:natseq] (n, m:nat) (le n m) ->(a, b:nat) (u n a) -> (u m b) ->(le a b).

Lemma s2seq_incr1:
  (u:natseq) (n:nat) (a, b:nat) ((s2seq u) n a) -> ((s2seq u) (S n) b) ->
  (le a b).
Simpl.
Intros.
Case H; Intros.
Case H0; Intros.
Case H1; Clear H1; Intros.
Case H2; Clear H2; Intros.
Inversion H3.
Replace a with b0.
Apply H6 with x0; Auto.
Apply (funct u x); Auto.
Replace a with b0.
Apply H8 with x0; Auto.
Apply lt_trans with x; Auto.
Apply s1seq_sincr1 with u n; Auto.
Apply (funct u x); Auto.
Qed.
(*
   s2seq est une sous-suite croissante de u.
*)

Theorem s2seq_incr: (u:natseq)(incr (s2seq u)).
Unfold incr.
Intros u n m.
Induction 1.
Intros.
Replace a with b; Auto.
Apply (funct (s2seq u) n); Auto.
Intros.
Generalize (total (s2seq u) m0); Intros.
Case H4; Intros.
Apply le_trans with x; Auto.
Apply s2seq_incr1 with u m0; Auto.
Qed.

