Section Optimized.

(* Require *)
Require Export PolyList.
Require Import Inclusion.
Require Import Inverse_Image.
Require Import Wf_nat.
Require Import Relation_Definitions.

(* Hypothesis *)
Variable poly:Set.
Variable spoly:poly -> poly ->poly.
Variable divP:poly -> poly ->Prop.
Hypothesis divP_trans:(transitive poly divP).
Variable divP_dec:(a, b:poly){(divP a b)}+{~ (divP a b)}.
Variable addL:poly -> (list poly) ->(list poly).
Hypothesis addL_cons:
           (a, b:poly) (aL:(list poly)) (In a (addL b aL)) ->a=b \/ (In a aL).
Hypothesis addL_id1:(a:poly) (aL:(list poly))(In a (addL a aL)).
Hypothesis addL_id2:(a, b:poly) (aL:(list poly)) (In a aL) ->(In a (addL b aL)).
Variable addS:poly -> (list poly) ->(list poly).
Hypothesis addS_cons:
           (a, b:poly) (aL:(list poly)) (In a (addS b aL)) ->a=b \/ (In a aL).
Hypothesis addS_id1:(a:poly) (aL:(list poly))(In a (addS a aL)).
Hypothesis addS_id2:(a, b:poly) (aL:(list poly)) (In a aL) ->(In a (addS b aL)).
Variable red:poly -> (list poly) ->Prop.
Hypothesis red_com:
           (a, b:poly) (aL:(list poly)) (red (spoly a b) aL) ->
           (red (spoly b a) aL).
Hypothesis red_incl:
           (a:poly) (p, q:(list poly)) (incl p q) -> (red a p) ->(red a q).
Hypothesis red_cons:(a:poly) (p:(list poly)) (In a p) ->(red a p).
Hypothesis red_id:(a:poly) (aL:(list poly))(red (spoly a a) aL).
Variable zerop:poly ->Prop.
Hypothesis zerop_dec:(a:poly){(zerop a)}+{~ (zerop a)}.
Variable ppc:poly -> poly ->poly.
Hypothesis divp_ppc:(a, b, c:poly) (divP (ppc a b) c) ->(divP (ppc b a) c).
Variable Cb:poly -> (list poly) ->Prop.
Hypothesis Cb_id:(a:poly) (Q:(list poly)) (In a Q) ->(Cb a Q).
Hypothesis Cb_in:(a, b:poly) (Q:(list poly)) (Cb a Q) ->(Cb a (addL b Q)).
Hypothesis Cb_in1:(a, b:poly) (Q:(list poly)) (Cb a Q) ->(Cb a (cons b Q)).
Hypothesis Cb_trans:
           (a, b:poly) (Q:(list poly)) (Cb a (addL b Q)) -> (Cb b Q) ->(Cb a Q).
Hypothesis Cb_incl:
           (a:poly)
           (P, Q:(list poly)) ((a:poly) (In a P) ->(In a Q)) -> (Cb a P) ->
           (Cb a Q).
Hypothesis Cb_sp:
           (a, b:poly) (Q:(list poly)) (Cb a Q) -> (Cb b Q) ->(Cb (spoly a b) Q).
Variable nf:poly -> (list poly) ->poly.
Hypothesis nf_red:
           (a:poly) (aP, aQ:(list poly)) (incl aP aQ) -> (red (nf a aP) aQ) ->
           (red a aQ).
Hypothesis red_zerop:
           (a:poly) (P:(list poly)) (red (nf a P) P) ->(zerop (nf a P)).
Hypothesis zerop_red:(a:poly) (aP:(list poly)) (zerop (nf a aP)) ->(red a aP).
Hypothesis nf_Cb:(a:poly) (aP:(list poly)) (Cb a aP) ->(Cb (nf a aP) aP).
Hypothesis divP_ppc:(a, b, c:poly) (divP a b) -> (divP a c) ->(divP a (ppc b c)).
Hypothesis RO:(list poly) -> (list poly) ->Prop.
Hypothesis wf_RO:(well_founded (list poly) RO).
Hypothesis RO_lem:
           (a:poly) (P:(list poly)) ~ (zerop (nf a P)) ->
           (RO (addL (nf a P) P) P).
Variable foreigner:poly -> poly ->Prop.
Hypothesis foreigner_red:
           (a, b:poly)
           (P:(list poly)) (foreigner a b) -> (In a P) -> (In b P) ->
           (red (spoly a b) P).
Variable foreigner_dec:(a, b:poly){(foreigner a b)}+{~ (foreigner a b)}.
Hypothesis zerop_ddivp_ppc:
           (a, b:poly) ~ (zerop a) -> ~ (zerop b) ->(divP (ppc a b) b).
Hypothesis divp_nzeropl:(a, b:poly) (divP a b) ->~ (zerop a).
Hypothesis divp_nzeropr:(a, b:poly) (divP a b) ->~ (zerop b).
Hypothesis zerop_red_spoly_l:
           (a, b:poly) (zerop a) ->(Q:(list poly))(red (spoly a b) Q).
Hypothesis zerop_red_spoly_r:
           (a, b:poly) (zerop b) ->(Q:(list poly))(red (spoly a b) Q).
(* Useful definitions and theorems *)

Mutual Inductive stable: (list poly) -> (list poly) ->Prop :=
     stable0:
       (P, Q:(list poly))
       ((a:poly) (Cb a P) ->(Cb a Q)) -> ((a:poly) (Cb a Q) ->(Cb a P)) ->
       (stable P Q) .
Hints Resolve stable0.

Theorem stable_refl: (Q:(list poly))(stable Q Q).
Auto.
Qed.

Theorem stable_trans:
  (Q, y, R:(list poly)) (stable Q y) -> (stable y R) ->(stable Q R).
Intros Q y R H' H'0; Inversion H'; Inversion H'0; Auto.
Qed.

Theorem stable_sym: (Q, R:(list poly)) (stable R Q) ->(stable Q R).
Intros Q R H'; Elim H'; Auto.
Qed.

Theorem Cb_stable: (a:poly) (Q:(list poly)) (Cb a Q) ->(stable Q (addL a Q)).
Intros a Q H'0; Apply stable0; Auto.
Intros a0 H'1; Try Assumption.
Apply Cb_trans with b := a; Auto.
Qed.

Theorem in_incl:
  (A:Set) (p, q:(list A)) (a, b:A) (incl p q) -> (In a p) ->(In a q).
Auto.
Qed.

Mutual Inductive reds: poly -> poly -> (list poly) ->Prop :=
     reds0: (P:(list poly)) (a, b:poly) (red (spoly a b) P) ->(reds a b P)
    | reds1:
        (P:(list poly))
        (a, b, c:poly)
        (In c P) -> (reds a c P) -> (reds c b P) -> (divP (ppc a b) c) ->
        (reds a b P) .

Theorem reds_com: (P:(list poly)) (a, b:poly) (reds a b P) ->(reds b a P).
Intros P a b H'; Elim H'; Simpl; Auto.
Intros P0 a0 b0 H'0.
Apply reds0; Auto.
Intros P0 a0 b0 c H'0 H'1 H'2 H'3 H'4 H'5.
Apply reds1 with c := c; Auto.
Qed.
(* Now we are ready!! We start with the definition of genCpC *)

Mutual Inductive cpRes: Set :=
     Keep: (P:(list poly))cpRes
    | DontKeep: (P:(list poly))cpRes .

Definition getRes: cpRes ->(list poly).
Intros H'; Case H'; Auto.
Defined.

Definition addRes: poly -> cpRes ->cpRes.
Intros i H'; Case H'.
Intros H'0; Exact (Keep (cons i H'0)).
Intros H'0; Exact (DontKeep (cons i H'0)).
Defined.

Definition slice: poly -> poly -> (list poly) ->cpRes.
Intros i a q; Elim q; Clear q.
Case (foreigner_dec i a).
Intros H; Exact (DontKeep (nil ?)).
Intros H; Exact (Keep (nil ?)).
Intros b q1 Rec.
Case (divP_dec (ppc i a) b).
Intros divP0; Exact (DontKeep (cons b q1)).
Intros divP0.
Case (divP_dec (ppc i b) a).
Intros divP1; Exact Rec.
Intros divP1; Exact (addRes b Rec).
Defined.

Definition slicef: poly -> poly -> (list poly) ->(list poly).
Intros i a q; Case (slice i a q); Auto.
Defined.

Theorem slicef_incl: (a, b:poly) (P:(list poly))(incl (slicef a b P) P).
Intros a b P; Elim P; Simpl; Auto.
Unfold slicef; Simpl; Auto.
Case (foreigner_dec a b); Intros H; Apply incl_refl; Auto with datatypes.
Intros c; Unfold slicef; Simpl.
Case (divP_dec (ppc a b) c); Auto with datatypes.
Case (divP_dec (ppc a c) b); Auto with datatypes.
Intros H' H'0 l; Case (slice a b l); Simpl; Auto with datatypes.
Qed.

Theorem slice_inv:
  (a, b:poly) (P:(list poly)) (c:poly) (In c P) ->
  (In c (getRes (slice a b P))) \/ (divP (ppc a c) b).
Intros a b P; Elim P; Simpl; Auto.
Intros c H'; Elim H'.
Intros p aP1; Case (divP_dec (ppc a b) p); Simpl; Auto.
Case (divP_dec (ppc a p) b); Auto.
Intros H' H'0 H'1 c H'2; Elim H'2;
 [Intros H'3; Rewrite <- H'3; Clear H'2 | Intros H'3; Clear H'2]; Auto.
Case (slice a b aP1); Simpl; Auto.
Intros P0 H' H'0 H'1 c H'2; Elim H'2;
 [Intros H'3; Rewrite <- H'3; Clear H'2 | Intros H'3; Clear H'2]; Auto.
Elim (H'1 c); [Intros H'5; Try Exact H'5 | Intros H'5 | Idtac]; Auto.
Intros P0 H' H'0 H'1 c H'2; Elim H'2;
 [Intros H'3; Rewrite <- H'3; Clear H'2 | Intros H'3; Clear H'2]; Auto.
Elim (H'1 c); [Intros H'5; Try Exact H'5 | Intros H'5 | Idtac]; Auto.
Qed.

Theorem slice_cons:
  (i, a:poly) (aP, Q:(list poly)) (slice i a aP)=(DontKeep Q) ->
  (Ex [c:poly] (In c Q) /\ (divP (ppc i a) c)) \/ (foreigner i a).
Intros i a aP; Elim aP.
Simpl; Case (foreigner_dec i a); Auto.
Intros H' Q H'0; Inversion H'0.
Intros a0 l H' Q; Simpl.
Case (divP_dec (ppc i a) a0); Auto.
Intros H'0 H'1; Inversion H'1; Auto.
Left; Exists a0; Split; Simpl; Auto.
Case (divP_dec (ppc i a0) a); Auto.
Generalize H'; Clear H'; Case (slice i a l); Simpl; Auto.
Intros P H' H'0 H'1 H'2; Inversion H'2.
Intros P H' H'0 H'1 H'2; Inversion H'2.
Elim (H' P);
 [Intros H'5; Elim H'5; Intros c E; Elim E; Intros H'6 H'7; Clear E H'5 |
   Intros H'5 | Idtac]; Auto.
Left; Exists c; Split; Simpl; Auto.
Qed.

Definition Tl: (list poly) -> (list poly) ->Prop.
Exact [x, y:(list poly)](lt (length x) (length y)).
Defined.

Theorem wf_Tl: (well_founded (list poly) Tl).
Apply (wf_inverse_image ? ? lt (!length poly)); Auto.
Generalize lt_wf; Auto.
Qed.

Scheme Sdep := Induction for prod Sort Prop.
Require Import Arith.

Theorem slice_Tl: (a, ia:poly) (L:(list poly))(Tl (slicef a ia L) (cons a L)).
Intros a b P; Elim P; Simpl; Auto.
Unfold slicef; Simpl; Auto.
Case (foreigner_dec a b); Unfold Tl; Simpl; Auto.
Intros c l.
Unfold slicef; Simpl; Auto.
Case (divP_dec (ppc a b) c); Auto.
Unfold Tl; Simpl; Auto.
Case (divP_dec (ppc a c) b); Auto.
Intros H' H'0; Case (slice a b l); Simpl; Auto.
Unfold Tl; Simpl; Auto.
Unfold Tl; Simpl; Auto.
Unfold Tl; Intros H' H'0; Case (slice a b l); Simpl; Auto with arith.
Qed.

Mutual Inductive genPcP:
                   poly -> (list poly) -> (list poly) -> (list poly) ->Prop :=
     genPcP0: (i:poly) (L:(list poly))(genPcP i (nil ?) L L)
    | genPcP1:
        (L, L1, L2, L3:(list ?))
        (a, i:poly) (slice i a L1)=(Keep L2) -> (genPcP i L2 L L3) ->
        (genPcP i (cons a L1) L (addS (spoly i a) L3))
    | genPcP2:
        (L, L1, L2, L3:(list ?))
        (a, i:poly) (slice i a L1)=(DontKeep L2) -> (genPcP i L2 L L3) ->
        (genPcP i (cons a L1) L L3) .
Hints Resolve genPcP0.

Theorem genPcP_spoly:
  (i:poly) (L, L1, L2:(list ?)) (genPcP i L1 L L2) ->
  (a:poly) (In a L2) ->(Ex [b:poly] (In b L1) /\ a=(spoly i b)) \/ (In a L).
Intros i L L1 L2 H'; Elim H'; Clear H'; Simpl; Auto.
Intros L0 L3 L4 L5 a i0 H' H'0 H'1 a0 H'2.
Case (addS_cons a0 (spoly i0 a) L5); Auto; Intros H'7.
Rewrite H'7; Auto.
Left; Exists a; Split; Simpl; Auto.
Elim (H'1 a0); Auto.
Intros H'3; Case H'3; Intros b E; Case E; Intros H'4 H'5; Rewrite H'5;
 Clear E H'3.
Left; Exists b; Split; Auto.
Right; Try Assumption.
Generalize (slicef_incl i0 a L3); Unfold slicef; Rewrite H'; Auto.
Intros L0 L3 L4 L5 a i0 H' H'0 H'1 a0 H'2.
Elim (H'1 a0);
 [Intros H'5; Elim H'5; Intros b E; Elim E; Intros H'6 H'7; Rewrite H'7;
   Clear E H'5 | Intros H'5 | Idtac]; Auto.
Left; Exists b; Split; [Right | Idtac]; Auto.
Generalize (slicef_incl i0 a L3); Unfold slicef; Rewrite H'; Auto.
Qed.

Theorem genPcP_incl:
  (i:poly) (L, L1, L2:(list ?)) (genPcP i L1 L L2) ->(incl L L2).
Intros i L L1 L2 H'; Elim H'; Simpl; Auto with datatypes.
Intros L0 L3 L4 L5 a i0 H'0 H'1 H'2.
Unfold incl; Simpl; Auto.
Qed.

Lemma spoly_cons_genPcP0:
  (aP, R, Q:(list ?)) (i:poly) (genPcP i aP R Q) -> ~ (zerop i) ->
  (b:poly) (In b aP) -> ~ (zerop b) ->
  (Ex [c:poly]
  (In c aP) /\ (((In (spoly i c) Q) \/ (foreigner i c)) /\ (divP (ppc i b) c))).
Intros aP R Q i H'; Elim H'; Clear H' i aP R Q; Simpl; Auto.
Intros i L H' b H'0; Elim H'0.
Intros L L1 L2 L3 a i H' H'0 H'1 H'2 b H'3 H'4.
Cut (incl L2 L1);
 [Intros incl0 | Generalize (slicef_incl i a L1); Unfold slicef; Rewrite H'];
 Auto.
Elim H'3; [Intros H'5; Rewrite <- H'5; Clear H'3 | Intros H'5; Clear H'3]; Auto.
Exists a; Split; [Idtac | Split; [Left | Idtac]]; Auto.
Rewrite H'5; Auto.
Elim (slice_inv i a L1 b); [Intros H'10 | Intros H'10 | Idtac]; Auto.
Rewrite H' in H'10; Simpl in H'10; Auto.
LApply H'1;
 [Intros H'3; Elim (H'3 b);
   [Intros c E; Elim E; Intros H'9 H'11; Elim H'11; Intros H'12 H'13; Elim H'12;
     [Intros H'14; Clear H'12 H'11 E H'1 | Intros H'14; Clear H'12 H'11 E H'1] |
     Clear H'1 | Clear H'1] | Clear H'1]; Auto.
Exists c; Split; [Right | Idtac]; Auto.
Exists c; Split; [Idtac | Split; [Right | Idtac]]; Auto.
Exists a; Split; [Idtac | Split]; Auto.
Intros L L1 L2 L3 a i H' H'0 H'1 H'2 b H'3 H'4.
Cut (incl L2 L1);
 [Intros incl0 | Generalize (slicef_incl i a L1); Unfold slicef; Rewrite H'];
 Auto.
Elim H'3; [Intros H'5; Rewrite <- H'5; Clear H'3 | Intros H'5; Clear H'3]; Auto.
Elim (slice_cons i a L1 L2);
 [Intros H'8; Elim H'8; Intros c E; Elim E; Intros H'9 H'10; Clear E H'8 |
   Intros H'8 | Idtac]; Auto.
LApply H'1;
 [Intros H'3; Elim (H'3 c);
   [Intros c0 E; Elim E; Intros H'11 H'12; Elim H'12; Intros H'13 H'14;
     Elim H'13;
     [Intros H'15; Clear H'13 H'12 E H'1 | Intros H'15; Clear H'13 H'12 E H'1] |
     Clear H'1 | Clear H'1] | Clear H'1]; Auto.
Exists c0; Split; [Idtac | Split; [Left | Idtac]]; Auto.
Apply divP_trans with y := (ppc i c); Auto.
Apply divP_ppc; Auto.
Apply divp_ppc.
Apply zerop_ddivp_ppc; Auto.
Rewrite H'5; Auto.
Exists c0; Split; [Idtac | Split; [Right | Idtac]]; Auto.
Apply divP_trans with y := (ppc i c); Auto.
Rewrite H'5; Auto.
Apply divP_ppc; Auto.
Rewrite <- H'5; Auto.
Apply divp_nzeropr with a := (ppc i a); Auto.
Exists a; Split; [Idtac | Split; [Right | Idtac]]; Auto.
Rewrite H'5; Auto.
Elim (slice_inv i a L1 b); [Intros H'10 | Intros H'10 | Idtac]; Auto.
Rewrite H' in H'10; Simpl in H'10; Auto.
LApply H'1;
 [Intros H'3; Elim (H'3 b);
   [Intros c E; Elim E; Intros H'9 H'11; Elim H'11; Intros H'12 H'13; Elim H'12;
     [Intros H'14; Clear H'12 H'11 E H'1 | Intros H'14; Clear H'12 H'11 E H'1] |
     Clear H'1 | Clear H'1] | Clear H'1]; Auto.
Exists c; Split; [Right | Idtac]; Auto.
Exists c; Split; [Idtac | Split; [Right | Idtac]]; Auto.
Elim (slice_cons i a L1 L2);
 [Intros H'8; Elim H'8; Intros c E; Elim E; Intros H'9 H'11; Clear E H'8 |
   Intros H'8 | Idtac]; Auto.
LApply H'1;
 [Intros H'3; Elim (H'3 c);
   [Intros c0 E; Elim E; Intros H'12 H'13; Elim H'13; Intros H'14 H'15;
     Elim H'14;
     [Intros H'16; Clear H'14 H'13 E H'1 | Intros H'16; Clear H'14 H'13 E H'1] |
     Clear H'1 | Clear H'1] | Clear H'1]; Auto.
Exists c0; Split; [Idtac | Split; [Left | Idtac]]; Auto.
Apply divP_trans with y := (ppc i c); Auto.
Apply divP_trans with y := (ppc i a); Auto.
Apply divP_ppc; Auto.
Apply divp_ppc.
Apply zerop_ddivp_ppc; Auto.
Apply divp_nzeropr with a := (ppc i b); Auto.
Exists c0; Split; [Idtac | Split; [Right | Idtac]]; Auto.
Apply divP_trans with y := (ppc i a); Auto.
Apply divP_trans with y := (ppc i c); Auto.
Apply divP_ppc; Auto.
Apply divp_ppc.
Apply zerop_ddivp_ppc; Auto.
Apply divp_nzeropr with a := (ppc i b); Auto.
Apply divp_nzeropr with a := (ppc i a); Auto.
Exists a; Split; [Idtac | Split; [Right | Idtac]]; Auto.
Qed.

Lemma spoly_cons_genPcP:
  (aP, R, Q:(list ?)) (i:poly) (genPcP i aP R Q) -> ~ (zerop i) ->
  (b:poly) (In b aP) -> ~ (zerop b) ->
  (Ex [c:poly]
  (In c aP) /\
  (((In (spoly i c) Q) \/ (red (spoly i c) (addL i aP))) /\ (divP (ppc i b) c))).
Intros aP R Q i H' H'0 b H'1 H'2.
LApply (spoly_cons_genPcP0 aP R Q i);
 [Intros H'7; LApply H'7;
   [Intros H'8; Elim (H'8 b);
     [Intros c E; Elim E; Intros H'12 H'13; Elim H'13; Intros H'14 H'15;
       Elim H'14;
       [Intros H'16; Clear H'14 H'13 E H'7 | Intros H'16; Clear H'14 H'13 E H'7] |
       Clear H'7 | Clear H'7] | Clear H'7] | Idtac]; Auto.
Exists c; Split; [Idtac | Split; [Left | Idtac]]; Auto.
Exists c; Split; [Idtac | Split; [Right | Idtac]]; Auto.
Qed.

Theorem Cb_genPcP:
  (i:poly)
  (P, Q, R, S:(list poly))
  (genPcP i P R Q) ->
  (Cb i S) ->
  ((a:poly) (In a P) ->(Cb a S)) -> ((a:poly) (In a R) ->(Cb a S)) ->
  (a:poly) (In a Q) ->(Cb a S).
Intros i P Q R S H'; Elim H'; Simpl; Auto.
Intros L L1 L2 L3 a i0 H'0 H'1 H'2 H'3 H'4 H'5 a0 H'6.
Case (addS_cons a0 (spoly i0 a) L3); Auto; Intros H'7.
Rewrite H'7; Auto.
Apply H'2; Auto.
Intros a1 H'6.
Apply H'4; Auto.
Right.
Generalize (slicef_incl i0 a L1); Unfold slicef; Rewrite H'0; Auto.
Intros L L1 L2 L3 a i0 H'0 H'1 H'2 H'3 H'4 H'5 a0 H'6; Auto.
Apply H'2; Auto.
Intros a1 H'7.
Apply H'4; Auto.
Generalize (slicef_incl i0 a L1); Unfold slicef; Rewrite H'0; Auto.
Qed.

Definition genPcPf0:
  (i:poly) (aP, R:(list poly)){Q:(list poly) | (genPcP i aP R Q)}.
Intros i aP; Pattern aP.
Apply well_founded_induction with A := (list poly) R := Tl; Clear aP; Auto.
Try Exact wf_Tl.
Intros aP; Case aP.
Intros H' R; Exists R; Auto.
Intros a L1 Rec L; Generalize (refl_equal ? (slice i a L1));
 Pattern 2 (slice i a L1); Case (slice i a L1).
Intros L2 H'.
LApply (Rec L2); [Intros H'1; Elim (H'1 L); Intros L3 E | Idtac]; Auto.
Exists (addS (spoly i a) L3); Auto.
Apply genPcP1 with L2 := L2; Auto.
Generalize (slice_Tl i a L1); Unfold slicef; Rewrite H'; Simpl; Auto.
Intros L2 H'.
LApply (Rec L2); [Intros H'1; Elim (H'1 L); Intros L3 E | Idtac]; Auto.
Exists L3; Auto.
Apply genPcP2 with L2 := L2; Auto.
Generalize (slice_Tl i a L1); Unfold slicef; Rewrite H'; Simpl; Auto.
Qed.

Definition genPcPf: poly -> (list poly) -> (list poly) ->(list poly).
Intros i aP Q; Case (genPcPf0 i aP Q).
Intros x H'; Exact x.
Defined.
(* The proof will carry on if we have the following 3 properties for
   the function genPcPf *)

Theorem Cb_genPcPf:
  (b:poly)
  (P, Q, R:(list poly))
  (Cb b R) ->
  ((a:poly) (In a P) ->(Cb a R)) -> ((a:poly) (In a Q) ->(Cb a R)) ->
  (a:poly) (In a (genPcPf b P Q)) ->(Cb a R).
Intros b P Q R; Unfold genPcPf; Case (genPcPf0 b P Q).
Intros x H' H'0 H'1 H'2 a H'3.
Apply Cb_genPcP with i := b P := P Q := x R := Q; Auto.
Qed.

Theorem genPcPf_incl: (a:poly) (aL, Q:(list poly))(incl Q (genPcPf a aL Q)).
Intros a aL Q; Unfold genPcPf; Case (genPcPf0 a aL Q).
Intros x H'.
Apply genPcP_incl with i := a L1 := aL; Auto.
Qed.
Hints Resolve genPcPf_incl.

Theorem spoly_addL_genPcPf:
  (aP, R, Q:(list poly)) (a, b:poly) ~ (zerop a) -> ~ (zerop b) -> (In b aP) ->
  (Ex [c:poly]
  (In c aP) /\
  (((In (spoly a c) (genPcPf a aP Q)) \/ (red (spoly a c) (addL a aP))) /\
   (divP (ppc a b) c))).
Intros aP H' Q a b H'0 H'1 H'2.
Unfold genPcPf.
Case (genPcPf0 a aP Q).
Intros x H'3.
Apply spoly_cons_genPcP with R := Q; Auto.
Qed.
(* Now we can define the optimized version of Buchberger *)

Definition genOCPf: (list poly) ->(list poly).
Intros H'; Elim H'.
Exact (nil poly).
Intros a l rec; Exact (genPcPf a l rec).
Defined.

Theorem genOCPf_stable: (a:poly) (P:(list poly)) (In a (genOCPf P)) ->(Cb a P).
Intros a P; Generalize a; Elim P; Clear a P; Simpl; Auto.
Intros a l H' a0 H'0.
Apply Cb_genPcPf with b := a P := l Q := (genOCPf l); Auto with datatypes.
Qed.

Mutual Inductive OBuch: (list poly) -> (list poly) -> (list poly) ->Prop :=
     OBuch0: (aL:(list poly))(OBuch aL (nil poly) aL)
    | OBuch1:
        (a:poly)
        (aP, Q, R:(list poly))
        (OBuch (addL (nf a aP) aP) (genPcPf (nf a aP) aP Q) R) ->
        ~ (zerop (nf a aP)) ->(OBuch aP (cons a Q) R)
    | OBuch2:
        (a:poly) (aP, Q, R:(list poly)) (OBuch aP Q R) -> (zerop (nf a aP)) ->
        (OBuch aP (cons a Q) R) .
Hints Resolve OBuch0 OBuch2.
Hints Resolve incl_refl incl_tl.

Theorem incl_addL:
  (a:poly) (L1, L2:(list poly)) (incl (addL a L1) L2) ->(incl (cons a L1) L2).
Unfold incl; Simpl; Auto.
Intros a L1 L2 H' a0 H'0; Case H'0;
 [Intros H'1; Rewrite <- H'1; Clear H'0 | Intros H'1; Clear H'0]; Auto.
Qed.

Theorem ObuchPincl: (aP, R, Q:(list poly)) (OBuch aP Q R) ->(incl aP R).
Intros aP R Q H'; Elim H'; Simpl; Auto.
Intros a aP0 Q0 R0 H'0 H'1 H'2; Try Assumption.
Apply incl_tran with m := (cons (nf a aP0) aP0); Simpl; Auto.
Apply incl_addL; Auto.
Qed.

Theorem ObuchPred:
  (aP, R, Q:(list poly)) (OBuch aP Q R) ->(a:poly) (In a aP) ->(red a R).
Intros aP R Q H'; Elim H'; Simpl; Auto.
Qed.

Theorem ObuchQred:
  (aP, R, Q:(list poly)) (OBuch aP Q R) ->(a:poly) (In a Q) ->(red a R).
Intros aP R Q H'; Elim H'; Simpl; Auto.
Intros aL a H'0; Elim H'0.
Intros a aP0 Q0 R0 H'0 H'1 H'2 a0 H'3; Elim H'3;
 [Intros H'4; Rewrite <- H'4; Clear H'3 | Intros H'4; Clear H'3]; Auto.
Apply red_incl with p := (addL (nf a aP0) aP0); Auto.
Apply ObuchPincl with Q := (genPcPf (nf a aP0) aP0 Q0); Auto.
Apply nf_red with aP := aP0; Simpl; Auto.
Unfold incl; Auto.
Apply H'1; Auto.
Apply (genPcPf_incl (nf a aP0) aP0 Q0); Auto.
Intros a aP0 Q0 R0 H'0 H'1 H'2 a0 H'3; Elim H'3;
 [Intros H'4; Rewrite <- H'4; Clear H'3 | Intros H'4; Clear H'3]; Auto.
Apply red_incl with p := aP0; Auto.
Apply ObuchPincl with Q := Q0; Auto.
Qed.

Theorem OBuch_Stable:
  (P, Q, R:(list poly)) (OBuch P Q R) -> ((a:poly) (In a Q) ->(Cb a P)) ->
  (stable P R).
Intros P Q R H'; Elim H'; Simpl; Auto.
Intros a aP Q0 R0 H'0 H'1 H'2 H'3.
Apply stable_trans with y := (addL (nf a aP) aP); Auto.
Apply stable0; Auto.
Intros a0 H'4.
Apply Cb_trans with b := (nf a aP); Auto.
Apply H'1; Auto.
Intros a0 H'4.
Apply Cb_genPcPf with b := (nf a aP) P := aP Q := Q0; Auto.
Qed.

Mutual Inductive redIn:
                   poly -> poly -> (list poly) -> (list poly) -> (list poly) ->
                   Prop :=
     redIn0b:
       (P, Q, R:(list poly)) (a, b:poly) (redIn b a P Q R) ->(redIn a b P Q R)
    | redIn0:
        (P, Q, R:(list poly)) (a, b:poly) (In (spoly a b) Q) ->(redIn a b P Q R)
    | redIn1:
        (P, Q, R:(list poly)) (a, b:poly) (red (spoly a b) R) ->
        (redIn a b P Q R)
    | redIn2:
        (P, Q, R:(list poly))
        (a, b, c:poly)
        (In c P) ->
        (redIn a c P Q R) -> (redIn b c P Q R) -> (divP (ppc a b) c) ->
        (redIn a b P Q R) .
Hints Resolve redIn1 redIn0.

Remark lem_redIn_nil:
  (aP, Q, R:(list poly))
  (a, b:poly)
  (In a R) -> (In b R) -> (redIn a b aP Q R) -> Q=(nil poly) -> aP=R ->
  (reds a b R).
Intros aP Q R a b H' H'0 H'1; Elim H'1; Auto.
Intros P Q0 R0 a0 b0 H'2 H'3 H'4 H'5.
Apply reds_com; Auto.
Intros P Q0 R0 a0 b0 H'2 H'3 H'4.
Rewrite H'3 in H'2; Elim H'2.
Intros P Q0 R0 a0 b0 H'2 H'3 H'4; Rewrite <- H'4.
Rewrite H'4.
Apply reds0; Auto.
Intros P Q0 R0 a0 b0 c H'2 H'3 H'4 H'5 H'6 H'7 H'8 H'9.
Apply reds1 with c := c; Auto.
Rewrite <- H'9; Auto.
Apply reds_com; Auto.
Qed.

Theorem redIn_nil:
  (R:(list poly))
  (a, b:poly) (In a R) -> (In b R) -> (redIn a b R (nil poly) R) ->(reds a b R).
Intros R a b H' H'0 H'1.
Apply lem_redIn_nil with aP := R Q := (nil poly); Auto.
Qed.

Remark lem_redln_cons:
  (aP, R, Q:(list poly))
  (a, b:poly) (In a aP) -> (In b aP) -> (redIn a b aP Q R) ->
  (c:poly) (Q1:(list poly)) Q=(cons c Q1) -> (red c R) ->(redIn a b aP Q1 R).
Intros aP R Q a b H' H'0 H'1; Elim H'1; Auto.
Intros P Q0 R0 a0 b0 H'2 H'3 c Q1 H'4 H'5.
Apply redIn0b; Auto.
Apply H'3 with c := c; Auto.
Intros P Q0 R0 a0 b0 H'2 c Q1 H'3 H'4.
Rewrite H'3 in H'2; Elim H'2; Auto.
Intros H'5; Rewrite H'5 in H'4; Auto.
Intros P Q0 R0 a0 b0 c H'2 H'3 H'4 H'5 H'6 H'7 c0 Q1 H'8 H'9.
Apply redIn2 with c := c; Auto.
Apply H'4 with c0 := c0; Auto.
Apply H'6 with c0 := c0; Auto.
Qed.

Theorem redln_cons:
  (aP, R, Q:(list poly))
  (a, b, c:poly)
  (In a aP) -> (In b aP) -> (redIn a b aP (cons c Q) R) -> (red c R) ->
  (redIn a b aP Q R).
Intros aP R Q a b c H' H'0 H'1 H'2; Try Assumption.
Apply lem_redln_cons with Q := (cons c Q) c := c; Auto.
Qed.

Theorem redInclP:
  (P, Q, R:(list poly)) (a, b:poly) (redIn a b P Q R) ->
  (P1:(list poly)) (incl P P1) ->(redIn a b P1 Q R).
Intros P Q R a b H'; Elim H'; Auto.
Intros P0 Q0 R0 a0 b0 H'0 H'1 P1 H'2.
Apply redIn0b; Auto.
Intros P0 Q0 R0 a0 b0 c H'0 H'1 H'2 H'3 H'4 H'5 Q1 H'6.
Apply redIn2 with c := c; Auto.
Qed.

Theorem redInInclQ:
  (P:(list poly))
  (Q:(list poly)) (R:(list poly)) (a, b:poly) (redIn a b P Q R) ->
  (Q1:(list poly)) (incl Q Q1) ->(redIn a b P Q1 R).
Intros P Q R a b H'; Elim H'; Auto.
Intros P0 Q0 R0 a0 b0 H'0 H'1 Q1 H'2; Try Assumption.
Apply redIn0b; Auto.
Intros P0 Q0 R0 a0 b0 c H'0 H'1 H'2 H'3 H'4 H'5 Q1 H'6; Try Assumption.
Apply redIn2 with c := c; Auto.
Qed.

Theorem redInclR:
  (P, Q, R:(list poly)) (a, b:poly) (redIn a b P Q R) ->
  (R1:(list poly)) (incl R R1) ->(redIn a b P Q R1).
Intros P Q R a b H'; Elim H'; Simpl; Auto.
Intros P0 Q0 R0 a0 b0 H'0 H'1 R1 H'2; Try Assumption.
Apply redIn0b; Auto.
Intros P0 Q0 R0 a0 b0 H'0 R1 H'1; Try Assumption.
Apply redIn1; Auto.
Apply red_incl with p := R0; Auto.
Intros P0 Q0 R0 a0 b0 c H'0 H'1 H'2 H'3 H'4 H'5 R1 H'6.
Apply redIn2 with c := c; Auto.
Qed.

Remark lem_redln_cons_gen:
  (aP, R, Q:(list poly))
  (a, b:poly) (In a aP) -> (In b aP) -> (redIn a b aP Q R) ->
  (c:poly)
  (Q1:(list poly))
  (incl (addL (nf c aP) aP) R) -> <(list poly)> Q=(cons c Q1) ->
  (redIn a b (addL (nf c aP) aP) Q1 R).
Intros aP R Q a b H' H'0 H'1; Elim H'1; Auto.
Intros P Q0 R0 a0 b0 H'2 H'3 c Q1 H'4 H'5.
Apply redIn0b; Auto.
Intros P Q0 R0 a0 b0 H'2 c Q1 H'3 H'4.
Rewrite H'4 in H'2; Elim H'2; Auto.
Intros H'5; Rewrite H'5.
Apply redIn1; Auto.
Apply nf_red with aP := P; Auto.
Apply incl_tran with m := (cons (nf c P) P); Simpl; Auto.
Apply incl_addL; Auto.
Apply red_cons; Auto.
Apply in_incl with p := (cons (nf c P) P); Auto.
Apply incl_addL; Auto.
Rewrite H'5; Simpl; Auto.
Intros P Q0 R0 a0 b0 c H'2 H'3 H'4 H'5 H'6 H'7 c0 Q1 H'8 H'9.
Apply redIn2 with c := c; Auto.
Qed.

Theorem redln_cons_gen:
  (aP, R, Q:(list poly))
  (a, b, c:poly)
  (In a aP) ->
  (In b aP) -> (redIn a b aP (cons c Q) R) -> (incl (addL (nf c aP) aP) R) ->
  (redIn a b (addL (nf c aP) aP) Q R).
Intros aP R Q a b c H' H'0 H'1 H'2.
Apply lem_redln_cons_gen with Q := (cons c Q); Auto.
Qed.
Hints Resolve redln_cons_gen.
Require Import Relation_Operators.

Theorem red_gen_in:
  (a:poly)
  (aP, R, Q:(list poly))
  ~ (zerop (nf a aP)) ->
  (OBuch (addL (nf a aP) aP) (genPcPf (nf a aP) aP Q) R) ->
  ((b, c:poly) (In b aP) -> (In c aP) ->(redIn b c aP (cons a Q) R)) ->
  (b:poly) (In b aP) -> ~ (zerop b) ->
  (redIn (nf a aP) b (addL (nf a aP) aP) (genPcPf (nf a aP) aP Q) R).
Intros a aP R Q H' H'0 H'1 b H'2 H'3.
LApply (spoly_addL_genPcPf aP);
 [Intros H'5; Elim (H'5 Q (nf a aP) b);
   [Intros c E; Elim E; Intros H'12 H'13; Elim H'13; Intros H'14 H'15;
     Elim H'14;
     [Intros H'16; Clear H'14 H'13 E | Intros H'16; Clear H'14 H'13 E] | Idtac |
     Idtac | Idtac] | Idtac]; Auto.
Apply redIn2 with c := c; Simpl; Auto.
Apply redln_cons_gen; Auto.
Apply redInInclQ with Q := (cons a Q); Auto with datatypes.
Apply ObuchPincl with Q := (genPcPf (nf a aP) aP Q); Auto.
Apply redIn2 with c := c; Simpl; Auto.
Apply redIn1.
Apply red_incl with p := (addL (nf a aP) aP); Auto.
Apply ObuchPincl with Q := (genPcPf (nf a aP) aP Q); Auto.
Apply redln_cons_gen; Auto.
Apply redInInclQ with Q := (cons a Q); Auto with datatypes.
Apply ObuchPincl with Q := (genPcPf (nf a aP) aP Q); Auto.
Qed.

Theorem OBuch_Inv:
  (aP, R, Q:(list poly))
  (OBuch aP Q R) -> ((a, b:poly) (In a aP) -> (In b aP) ->(redIn a b aP Q R)) ->
  (a, b:poly) (In a R) -> (In b R) ->(reds a b R).
Intros aP R Q H'; Elim H'; Simpl; Auto.
Intros aL H'0 a b H'1 H'2; Try Assumption.
Apply redIn_nil; Auto.
Intros a aP0 Q0 R0 H'0 H'1 H'2 H'3 a0 b H'4 H'5.
Apply H'1; Auto.
Intros a1 b0 H'6.
Case (addL_cons a1 (nf a aP0) aP0); Auto.
Intros H'7; Rewrite <- H'7; Auto.
Intros H'8.
Case (addL_cons b0 a1 aP0); Auto.
Intros H'9; Rewrite <- H'9; Auto.
Intros H'9.
Case (zerop_dec b0); Intros Z; Auto.
Rewrite H'7; Auto.
Apply red_gen_in; Auto.
Intros H'7 H'8.
Case (addL_cons b0 (nf a aP0) aP0); Auto.
Intros H'9; Rewrite <- H'9; Auto.
Apply redIn0b.
Case (zerop_dec a1); Intros Z.
Apply redIn1; Auto.
Rewrite H'9.
Apply red_gen_in; Auto.
Intros H'9.
Apply redln_cons with c := a; Simpl; Auto.
Apply redInclP with P := aP0; Auto.
Apply redInInclQ with Q := (cons a Q0); Auto with datatypes.
Unfold incl; Auto.
Apply nf_red with aP := aP0; Auto.
Apply incl_tran with m := (addL (nf a aP0) aP0); Auto.
Unfold incl; Auto.
Apply ObuchPincl with Q := (genPcPf (nf a aP0) aP0 Q0); Auto.
Apply red_cons; Auto.
Apply in_incl with p := (addL (nf a aP0) aP0); Simpl; Auto.
Apply ObuchPincl with Q := (genPcPf (nf a aP0) aP0 Q0); Auto.
Intros a aP0 Q0 R0 H'0 H'1 H'2 H'3 a0 b H'4 H'5.
Apply H'1; Auto.
Intros a1 b0 H'6 H'7.
Apply redln_cons with c := a; Auto.
Apply red_incl with p := aP0; Auto.
Apply ObuchPincl with Q := Q0; Auto.
Qed.

Theorem addL_incl:
  (a:poly) (L1, L2:(list poly)) (incl (cons a L1) L2) ->(incl (addL a L1) L2).
Unfold incl; Simpl; Auto.
Intros a L1 L2 H' a0 H'0.
Case (addL_cons a0 a L1); Auto.
Qed.

Theorem genOCp1_redln:
  (aL1:(list poly)) (R:(list poly)) (incl aL1 R) ->
  (a, b:poly) (In a aL1) -> (In b aL1) ->(redIn a b aL1 (genOCPf aL1) R).
Intros aL1; Elim aL1; Simpl; Auto.
Intros a l H' R H'0 a0 b H'1 H'2.
Elim H'2; [Intros H'3; Rewrite <- H'3; Clear H'2 | Intros H'3; Clear H'2]; Auto.
Elim H'1; [Intros H'2; Rewrite <- H'2; Clear H'1 | Intros H'2; Clear H'1]; Auto.
Apply redIn0b.
Case (zerop_dec a); Intros Z; Auto.
Case (zerop_dec a0); Intros Z1; Auto.
LApply (spoly_addL_genPcPf l);
 [Intros H'4; Elim (H'4 (genOCPf l) a a0);
   [Intros c E; Elim E; Intros H'11 H'12; Elim H'12; Intros H'13 H'14;
     Elim H'13;
     [Intros H'15; Clear H'13 H'12 E | Intros H'15; Clear H'13 H'12 E] | Idtac |
     Idtac | Idtac] | Idtac]; Auto.
Apply redIn2 with c := c; Auto.
Simpl; Auto.
Apply redInInclQ with Q := (genOCPf l); Auto.
Apply redInclP with P := l; Auto.
Apply H'; Auto.
Apply incl_tran with m := (cons a l); Simpl; Auto.
Apply redIn2 with c := c; Auto.
Simpl; Auto.
Apply redIn1; Auto.
Apply red_incl with p := (addL a l); Auto.
Apply addL_incl; Auto.
Apply redInclP with P := l; Auto.
Apply redInInclQ with Q := (genOCPf l); Auto.
Apply H'; Auto.
Apply incl_tran with m := (cons a l); Auto.
Elim H'1; [Intros H'2; Rewrite <- H'2; Clear H'1 | Intros H'2; Clear H'1]; Auto.
Case (zerop_dec a); Intros Z; Auto.
Case (zerop_dec b); Intros Z1; Auto.
LApply (spoly_addL_genPcPf l);
 [Intros H'4; Elim (H'4 (genOCPf l) a b);
   [Intros c E; Elim E; Intros H'11 H'12; Elim H'12; Intros H'13 H'14;
     Elim H'13;
     [Intros H'15; Clear H'13 H'12 E | Intros H'15; Clear H'13 H'12 E] | Idtac |
     Idtac | Idtac] | Idtac]; Auto.
Apply redIn2 with c := c; Simpl; Auto.
Apply redInInclQ with Q := (genOCPf l); Auto.
Apply redInclP with P := l; Auto.
Apply H'; Auto.
Apply incl_tran with m := (cons a l); Simpl; Auto.
Apply redIn2 with c := c; Simpl; Auto.
Apply redIn1; Auto.
Apply red_incl with p := (addL a l); Auto.
Apply addL_incl; Auto.
Apply redInclP with P := l; Auto.
Apply redInInclQ with Q := (genOCPf l); Auto.
Apply H'; Auto.
Apply incl_tran with m := (cons a l); Auto.
Apply redInclP with P := l; Auto.
Apply redInInclQ with Q := (genOCPf l); Auto.
Apply H'; Auto.
Apply incl_tran with m := (cons a l); Auto.
Qed.

Theorem OBuch_Stable_f:
  (P, Q:(list poly)) (OBuch P (genOCPf P) Q) ->(stable P Q).
Intros P Q H'; Try Assumption.
Apply OBuch_Stable with Q := (genOCPf P); Auto.
Intros a H'0; Try Assumption.
Apply genOCPf_stable; Auto.
Qed.

Theorem OBuch_Inv_f:
  (P, Q:(list poly)) (OBuch P (genOCPf P) Q) ->
  (a, b:poly) (In a Q) -> (In b Q) ->(reds a b Q).
Intros P Q H' a b H'0 H'1; Try Assumption.
Apply OBuch_Inv with aP := P Q := (genOCPf P); Auto.
Intros a0 b0 H'3 H'4.
Apply genOCp1_redln; Auto.
Apply ObuchPincl with Q := (genOCPf P); Auto.
Qed.
Require Import Lexicographic_Product.

Local FPset := [A:(list poly)](list poly).

Definition Fl: (x:(list poly)) (FPset x) -> (FPset x) ->Prop.
Unfold FPset; Simpl.
Intros H' P1 P2.
Exact (Tl P1 P2).
Defined.

Theorem wf_Fl: (x:(list poly))(well_founded (FPset x) (Fl x)).
Unfold FPset; Simpl.
Intros x; Generalize wf_Tl; Auto.
Qed.

Local Co := (lexprod (list poly) FPset RO Fl).

Theorem wf_Co: (well_founded (sigS (list poly) FPset) Co).
LApply (wf_lexprod ? ? RO Fl);
 [Intros H'3; LApply H'3; [Intros H'4; Try Exact H'4; Clear H'3 | Clear H'3] |
   Idtac]; Auto.
Exact wf_Fl.
Qed.

Definition PtoS: (list poly) * (list poly) ->(sigS (list poly) FPset).
Intros H'; Case H'.
Intros P1 P2.
Exact (existS (list poly) FPset P1 P2).
Defined.

Definition RL: (list poly) * (list poly) -> (list poly) * (list poly) ->Prop :=
   [x, y:(list poly) * (list poly)](Co (PtoS x) (PtoS y)).

Theorem wf_RL: (well_founded (list poly) * (list poly) RL).
Apply (wf_inverse_image ? ? Co PtoS); Auto.
Try Exact wf_Co.
Qed.
Require Import LetP.

Definition pbuchf:
  (PQ:(list poly) * (list poly)){R:(list poly) | (OBuch (Fst PQ) (Snd PQ) R)}.
Intros pq; Pattern pq.
Apply well_founded_induction with A := (list poly) * (list poly) R := RL.
Try Exact wf_RL.
Intros x; Elim x.
Intros P Q; Case Q; Simpl.
Intros H'; Exists P; Auto.
Intros a Q2 Rec.
Apply LetP with A := poly h := (nf a P).
Intros a0 H'.
Case (zerop_dec a0); Intros red0.
Elim (Rec (P, Q2)); Simpl; [Intros R E | Idtac]; Auto.
Exists R; Auto.
Apply OBuch2; Auto.
Rewrite <- H'; Auto.
Red; Unfold Co; Unfold PtoS.
Apply (right_lex ? ? RO Fl); Auto.
Red; Red; Simpl; Auto.
Elim (Rec ((addL a0 P), (genPcPf a0 P Q2))); Simpl;
 [Intros R E0; Try Exact E0 | Idtac].
Exists R; Auto.
Apply OBuch1; Auto.
Rewrite <- H'; Auto.
Rewrite <- H'; Auto.
Rewrite H'.
Red; Unfold Co; Unfold PtoS.
Apply (left_lex ? ? RO Fl); Auto.
Apply RO_lem; Auto.
Rewrite <- H'; Auto.
Defined.

Definition strip: (P:(list poly) ->Prop) (sig (list poly) P) ->(list poly).
Intros P H'; Case H'.
Intros x H'0; Try Assumption.
Defined.

Theorem pbuchf_Stable:
  (P, R:(list poly)) R=(strip ? (pbuchf (P, (genOCPf P)))) ->(stable P R).
Simpl.
Intros P R H'; Try Assumption.
Apply OBuch_Stable_f; Auto.
Rewrite H'.
Case (pbuchf <(list poly), (list poly)> (P, (genOCPf P))); Simpl; Auto.
Qed.

Theorem pbuchf_Inv:
  (P, R:(list poly)) R=(strip ? (pbuchf (P, (genOCPf P)))) ->
  (a, b:poly) (In a R) -> (In b R) ->(reds a b R).
Intros P R H' a b H'0 H'1; Simpl.
Apply OBuch_Inv_f with P := P; Auto.
Rewrite H'; Simpl; Auto.
Case (pbuchf <(list poly), (list poly)> (P, (genOCPf P))); Simpl; Auto.
Qed.

Definition buch: (list poly) ->(list poly).
Intros P; Exact (strip ? (pbuchf (P, (genOCPf P)))).
Defined.

Theorem buch_Stable: (P:(list poly))(stable P (buch P)).
Intros P; Apply pbuchf_Stable; Auto.
Qed.

Theorem buch_reds:
  (P:(list poly)) (a, b:poly) (In a (buch P)) -> (In b (buch P)) ->
  (reds a b (buch P)).
Intros P a b H' H'0.
Apply pbuchf_Inv with P := P; Auto.
Qed.
End Optimized.

