
(****************************************************************************
                                                                           
          Buchberger : spolynomial                                         
                                                                           
          Laurent Thery June 97 (revised Mai 98)                           
                                                                           
****************************************************************************)
Section def_spoly.
Load "shyp_def_spoly".
Load "mod_sreduce".

Theorem divP_on_eqM:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) -> (eqM a b) ->(divP a b).
Intros a b H' nZb H'1; Apply divTerm_def with nZb := nZb; Auto.
Qed.

Theorem divP_on_eqM_eqM:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b)) (eqM a b) ->(eqM (divTerm a b nZb) T1).
Auto.
Qed.
Hints Resolve divP_on_eqM divP_on_eqM_eqM.

Mutual Inductive ppcm[p, q:Term]: Term ->Prop :=
     ppcm0:
       (s:Term)
       ((r:Term) (divP r p) -> (divP r q) ->(divP r s)) ->
       (divP s p) -> (divP s q) ->(ppcm p q s) .
Hints Resolve ppcm0.

Theorem pcc_is_ppcm:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->(ppcm a b (ppc a b)).
Intros a b nZa nZb; Apply ppcm0; Auto.
Intros r H'1 H'2; Inversion H'1; Inversion H'2.
Cut ~ (zeroP (ppc a b)); [Intros nZppab | Auto].
Apply divTerm_def with nZb := nZppab; Auto.
Apply divTerm_ppc with nZa := nZa nZb := nZb; Auto.
Apply divTerm_def with nZb := nZa; Auto.
Apply divTerm_def with nZb := nZb; Auto.
Qed.

Theorem ppc_multTerm_divP:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->(divP (multTerm a b) (ppc a b)).
Intros a b H' H'0.
Elim pcc_is_ppcm; Auto.
Qed.
Hints Resolve ppc_multTerm_divP.

Theorem divP_comp:
  (a, b, c, d:Term) (divP a c) -> (eqTerm a b) -> (eqTerm c d) ->(divP b d).
Intros a b c d H'; Generalize b d; Elim H'.
Intros a0 b0 nZa0 nZb0 H'2 b1 d0 H'3 H'4; Auto.
Cut ~ (zeroP d0); [Intros nZd0 | Auto].
Apply divTerm_def with nZb := nZd0; Auto.
Red; Intros nz1; Absurd (zeroP a0); Auto.
Apply zeroP_comp_eqTerm with a := b1; Auto.
Apply eqTerm_trans with y := a0; Auto.
Apply eqTerm_trans with y := (multTerm (divTerm a0 b0 nZb0) b0); Auto.
Red; Intros nz1; Absurd (zeroP b0); Auto.
Apply zeroP_comp_eqTerm with a := d0; Auto.
Qed.

Theorem divP_multTerm_l:
  (a, b, c:Term) ~ (zeroP a) -> ~ (zeroP b) -> (eqTerm (multTerm a b) c) ->
  (divP c a).
Intros a b c H' H'0 H'1.
Apply divP_comp with a := (multTerm a b) c := a; Auto.
Qed.

Theorem divP_multTerm_r:
  (a, b, c:Term) ~ (zeroP a) -> ~ (zeroP b) -> (eqTerm (multTerm a b) c) ->
  (divP c b).
Intros a b c H' H'0 H'1.
Apply divP_comp with a := (multTerm a b) c := b; Auto.
Qed.
Hints Resolve divP_multTerm_r divP_multTerm_l.

Theorem divP_pccl: (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->(divP (ppc a b) a).
Intros a b H' H'0; Try Assumption.
LApply (pcc_is_ppcm a b);
 [Intros H'3; LApply H'3; Clear H'3; [Intros H'4 | Idtac] | Idtac]; Auto; Auto.
Inversion H'4; Auto.
Qed.

Theorem divP_pccr: (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->(divP (ppc a b) b).
Intros a b H' H'0; Try Assumption.
LApply (pcc_is_ppcm a b);
 [Intros H'3; LApply H'3; Clear H'3; [Intros H'4 | Idtac] | Idtac]; Auto; Auto.
Inversion H'4; Auto.
Qed.
Require Import LetP.

Definition spolyf:
  (p, q:(list Term)) (Cp:(canonical p)) (Cq:(canonical q))(list Term).
Intros p1; Case p1.
Intros p2 H' H'0; Exact pO.
Intros a p11 p2; Case p2.
Intros H' H'0; Exact pO.
Intros b p22 H' H'0.
Apply LetP with A := Term h := (ppc a b).
Intros u H'.
Cut ~ (zeroP a); [Intros nZa | Apply canonical_nzeroP with p := p11; Auto].
Cut ~ (zeroP b); [Intros nZb | Apply canonical_nzeroP with p := p22; Auto].
Exact (minuspf
         (multm_lm (divTerm u a nZa) p11) (multm_lm (divTerm u b nZb) p22)).
Defined.

Theorem spolyf_canonical:
  (p, q:(list Term)) (Cp:(canonical p)) (Cq:(canonical q))
  (canonical (spolyf p q Cp Cq)).
Intros p; Case p; Simpl; Auto.
Intros a l q; Case q; Simpl; Auto.
Intros a0 l0 H' H'0; Try Assumption.
Cut (canonical l0); [Intros Op1 | EAuto].
Cut (canonical l); [Intros Op2 | EAuto].
Cut ~ (zeroP a); [Intros Z0 | Apply canonical_nzeroP with p := l; Auto].
Cut ~ (zeroP a0); [Intros Z1 | Apply canonical_nzeroP with p := l0; Auto].
Unfold LetP; Auto.
Qed.
Hints Resolve spolyf_canonical.

Theorem divTerm_compo:
  (a, b, c:Term) (nZb:~ (zeroP b)) (nZc:~ (zeroP c)) (divP a b) -> (divP b c) ->
  (eqTerm (divTerm a c nZc) (multTerm (divTerm a b nZb) (divTerm b c nZc))).
Intros a b c nZb nZc H'; Inversion H'.
Intros H'0; Inversion H'0.
Apply eqTerm_trans with y := (divTerm (multTerm (divTerm a b nZb) b) c nZc);
 Auto.
Qed.
Hints Resolve divTerm_compo.

Theorem spolyf_def:
  (a, b:Term)
  (nZa:~ (zeroP a))
  (nZb:~ (zeroP b))
  (p, q:(list Term)) (Cpxa:(canonical (pX a p))) (Cpxb:(canonical (pX b q)))
  (eqP
     (spolyf (pX a p) (pX b q) Cpxa Cpxb)
     (minuspf
        (multm_lm (divTerm (ppc a b) a nZa) p)
        (multm_lm (divTerm (ppc a b) b nZb) q))).
Intros a b nZa nZb p q Cpxa Cpxb; Simpl; Auto.
Cut (canonical p); [Intros Op1 | Auto]; Auto.
Cut (canonical q); [Intros Op2 | Auto]; Unfold LetP; Simpl; Auto.
Apply canonical_imp_canonical with a := b; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Qed.
Hints Resolve spolyf_def.

Theorem eqTerm_spolyf_red2:
  (a, b, c:Term)
  (nZa:~ (zeroP a))
  (nZb:~ (zeroP b)) (nZppab:~ (zeroP (ppc a b))) (divP c a) -> (divP c b) ->
  (p, q, r:(list Term))
  (Cpxa:(canonical (pX a p))) (Cpxb:(canonical (pX b q))) (canonical r) ->
  (eqP
     (multm_lm
        (divTerm c (ppc a b) nZppab) (spolyf (pX a p) (pX b q) Cpxa Cpxb))
     (minuspf (spminusf c b nZb r q) (spminusf c a nZa r p))).
Intros a b c nZa nZb nZppab H' H'0 p q r Cpxa Cpxb H'4.
Cut (canonical p); [Intros Op | EAuto].
Cut (canonical q); [Intros Oq | EAuto].
Cut (divP (ppc a b) a); [Intros div1 | Apply divP_pccl; Auto].
Cut (divP (ppc a b) b); [Intros div2 | Apply divP_pccr; Auto].
Cut (divP c (ppc a b)); [Intros div3 | Elim pcc_is_ppcm; Auto].
Apply eqp_trans
     with y :=
          (multm_lm
             (divTerm c (ppc a b) nZppab)
             (pluspf
                (multm_lm (divTerm (ppc a b) a nZa) p)
                (multm_lm (invTerm T1) (multm_lm (divTerm (ppc a b) b nZb) q)))).
Simpl; Unfold LetP; Simpl; Auto.
Apply multm_lm_comp; Auto.
Apply eqp_trans
     with y :=
          (minuspf
             (multm_lm (divTerm (ppc a b) a nZa) p)
             (multm_lm (divTerm (ppc a b) b nZb) q)); Auto.
Simpl; Auto.
Cut ~ (zeroP c); [Intros Z2 | Inversion H'0; Auto].
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm
                (divTerm c (ppc a b) nZppab)
                (multm_lm (divTerm (ppc a b) a nZa) p))
             (multm_lm
                (divTerm c (ppc a b) nZppab)
                (multm_lm (invTerm T1) (multm_lm (divTerm (ppc a b) b nZb) q))));
 [Auto | Idtac].
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm
                (multTerm
                   (divTerm c (ppc a b) nZppab) (divTerm (ppc a b) a nZa)) p)
             (multm_lm
                (multTerm (divTerm c (ppc a b) nZppab) (invTerm T1))
                (multm_lm (divTerm (ppc a b) b nZb) q))).
Apply eqp_sym; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (divTerm c a nZa) p)
             (multm_lm
                (multTerm (invTerm T1) (divTerm c (ppc a b) nZppab))
                (multm_lm (divTerm (ppc a b) b nZb) q))); [Auto | Idtac].
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (divTerm c a nZa) p)
             (multm_lm
                (invTerm T1)
                (multm_lm
                   (divTerm c (ppc a b) nZppab)
                   (multm_lm (divTerm (ppc a b) b nZb) q)))).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (divTerm c a nZa) p)
             (multm_lm
                (invTerm T1)
                (multm_lm
                   (multTerm
                      (divTerm c (ppc a b) nZppab) (divTerm (ppc a b) b nZb)) q))).
Apply eqp_sym; Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (divTerm c a nZa) p)
             (multm_lm (invTerm T1) (multm_lm (divTerm c b nZb) q))).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (invTerm T1) (multm_lm (divTerm c b nZb) q))
             (multm_lm (divTerm c a nZa) p)); [Auto | Idtac].
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (invTerm T1) (multm_lm (divTerm c b nZb) q))
             (pluspf pO (multm_lm (divTerm c a nZa) p))).
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (invTerm T1) (multm_lm (divTerm c b nZb) q))
             (pluspf (minuspf r r) (multm_lm (divTerm c a nZa) p))).
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (invTerm T1) (multm_lm (divTerm c b nZb) q))
             (pluspf
                (pluspf r (multm_lm (invTerm T1) r))
                (multm_lm (divTerm c a nZa) p))).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (invTerm T1) (multm_lm (divTerm c b nZb) q))
             (pluspf
                r
                (pluspf
                   (multm_lm (invTerm T1) r) (multm_lm (divTerm c a nZa) p)))).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf (multm_lm (invTerm T1) (multm_lm (divTerm c b nZb) q)) r)
             (pluspf (multm_lm (invTerm T1) r) (multm_lm (divTerm c a nZa) p))).
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf r (multm_lm (invTerm T1) (multm_lm (divTerm c b nZb) q)))
             (pluspf (multm_lm (invTerm T1) r) (multm_lm (divTerm c a nZa) p))).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (pluspf (multm_lm (invTerm T1) r) (multm_lm (divTerm c a nZa) p))).
Apply eqp_pluspf_com; Auto.
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (pluspf
                (multm_lm (invTerm T1) r)
                (multm_lm T1 (multm_lm (divTerm c a nZa) p)))).
Apply eqp_sym; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (pluspf
                (multm_lm (invTerm T1) r)
                (multm_lm (invTerm (invTerm T1)) (multm_lm (divTerm c a nZa) p)))).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (pluspf
                (multm_lm (invTerm T1) r)
                (multm_lm
                   (invTerm (multTerm T1 (invTerm T1)))
                   (multm_lm (divTerm c a nZa) p)))); Auto.
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (pluspf
                (multm_lm (invTerm T1) r)
                (multm_lm
                   (multTerm (invTerm T1) (invTerm T1))
                   (multm_lm (divTerm c a nZa) p)))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (pluspf
                (multm_lm (invTerm T1) r)
                (multm_lm
                   (invTerm T1)
                   (multm_lm (invTerm T1) (multm_lm (divTerm c a nZa) p))))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (multm_lm
                (invTerm T1)
                (pluspf
                   r (multm_lm (invTerm T1) (multm_lm (divTerm c a nZa) p))))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply canonical_multlm_lm; Auto.
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (multm_lm (invTerm T1) (minuspf r (multm_lm (divTerm c a nZa) p)))).
Apply eqp_sym; Auto.
Apply eqp_pluspf_com; Auto.
Apply canonical_multlm_lm; Auto.
Apply eqp_trans
     with y :=
          (minuspf
             (minuspf r (multm_lm (divTerm c b nZb) q))
             (minuspf r (multm_lm (divTerm c a nZa) p))); Auto; Apply eqp_sym;
 Auto.
Qed.

Theorem eqTerm_spolyf_red3:
  (a, b:Term)
  (nZa:~ (zeroP a))
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (Cpxa:(canonical (pX a p))) (Cpxb:(canonical (pX b q))) (canonical r) ->
  (eqP
     (spolyf (pX a p) (pX b q) Cpxa Cpxb)
     (minuspf (spminusf (ppc a b) b nZb r q) (spminusf (ppc a b) a nZa r p))).
Intros a b nZa nZb p q r Cpxa Cpxb H'1.
Cut ~ (zeroP (ppc a b)); [Intros nZppab | Auto].
Apply eqp_trans with y := (multm_lm T1 (spolyf (pX a p) (pX b q) Cpxa Cpxb)).
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (multm_lm
             (divTerm (ppc a b) (ppc a b) nZppab)
             (spolyf (pX a p) (pX b q) Cpxa Cpxb)).
Apply eqp_sym; Auto.
Apply eqTerm_spolyf_red2; Auto.
Apply divP_pccl; Auto.
Apply divP_pccr; Auto.
Qed.

Theorem spoly_is_minus:
  (a, b:Term)
  (nZa:~ (zeroP a))
  (nZb:~ (zeroP b))
  (p, q:(list Term)) (Cpxa:(canonical (pX a p))) (Cpxb:(canonical (pX b q)))
  (eqP
     (spolyf (pX a p) (pX b q) Cpxa Cpxb)
     (minuspf
        (multm_lm (divTerm (ppc a b) a nZa) (pX a p))
        (multm_lm (divTerm (ppc a b) b nZb) (pX b q)))).
Simpl.
Intros a b nZa nZb p q Cpxa Cpxb.
Cut (canonical p); [Intros Op | EAuto].
Cut (canonical q); [Intros Oq | EAuto].
Cut ~ (zeroP (ppc a b)); [Intros nZppab | Auto].
Apply eqp_trans
     with y :=
          (minuspf
             (pX (ppc a b) (multm_lm (divTerm (ppc a b) a nZa) p))
             (pX (ppc a b) (multm_lm (divTerm (ppc a b) b nZb) q))).
Simpl; Unfold LetP; Simpl; Auto.
Apply eqp_trans
     with y :=
          (minuspf
             (multm_lm (divTerm (ppc a b) a nZa) p)
             (multm_lm (divTerm (ppc a b) b nZb) q)); Auto.
Apply eqp_sym.
Apply minuspf_zero; Auto.
Cut (eqTerm (ppc a b) (multTerm (divTerm (ppc a b) a nZa) a));
 [Intros Z2 | Idtac].
Cut (eqTerm (ppc a b) (multTerm (divTerm (ppc a b) b nZb) b));
 [Intros Z3 | Idtac].
Apply minuspf_comp; Auto.
Apply canonical_pX_eqM with a := (multTerm (divTerm (ppc a b) a nZa) a); Auto.
LApply (divP_pccl a b);
 [Intros H'5; LApply H'5; [Intros H'6; Inversion H'6; Clear H'5 | Clear H'5] |
   Idtac]; Auto.
Apply canonical_pX_eqM with a := (multTerm (divTerm (ppc a b) b nZb) b); Auto.
LApply (divP_pccr a b);
 [Intros H'5; LApply H'5; [Intros H'6; Inversion H'6; Clear H'5 | Clear H'5] |
   Idtac]; Auto.
LApply (divP_pccr a b);
 [Intros H'5; LApply H'5; [Intros H'6; Inversion H'6; Clear H'5 | Clear H'5] |
   Idtac]; Auto.
Qed.

Theorem spoly_div_is_minus:
  (a, b, c:Term)
  (nZa:~ (zeroP a))
  (nZb:~ (zeroP b)) (nZppab:~ (zeroP (ppc a b))) (divP c a) -> (divP c b) ->
  (p, q:(list Term)) (Cpxa:(canonical (pX a p))) (Cpxb:(canonical (pX b q)))
  (eqP
     (multm_lm
        (divTerm c (ppc a b) nZppab) (spolyf (pX a p) (pX b q) Cpxa Cpxb))
     (minuspf
        (multm_lm (divTerm c a nZa) (pX a p))
        (multm_lm (divTerm c b nZb) (pX b q)))).
Intros a b c nZa nZb nZppab H' H'0 p q Cpxa Cpxb.
Cut (canonical p); [Intros Op | EAuto].
Cut (canonical q); [Intros Oq | EAuto].
Cut ~ (zeroP c); [Intros Z2 | Idtac].
Apply eqp_trans
     with y :=
          (multm_lm
             (divTerm c (ppc a b) nZppab)
             (minuspf
                (multm_lm (divTerm (ppc a b) a nZa) (pX a p))
                (multm_lm (divTerm (ppc a b) b nZb) (pX b q)))).
Apply multm_lm_comp; Auto.
Apply spoly_is_minus; Auto.
Apply eqp_trans
     with y :=
          (minuspf
             (multm_lm
                (divTerm c (ppc a b) nZppab)
                (multm_lm (divTerm (ppc a b) a nZa) (pX a p)))
             (multm_lm
                (divTerm c (ppc a b) nZppab)
                (multm_lm (divTerm (ppc a b) b nZb) (pX b q)))).
Apply multlm_lm_dist_minuspf; Auto.
Apply eqp_trans
     with y :=
          (minuspf
             (multm_lm
                (multTerm
                   (divTerm c (ppc a b) nZppab) (divTerm (ppc a b) a nZa))
                (pX a p))
             (multm_lm
                (multTerm
                   (divTerm c (ppc a b) nZppab) (divTerm (ppc a b) b nZb))
                (pX b q))).
Apply minuspf_comp; Auto.
Apply eqp_sym; Auto.
Apply eqp_sym; Auto.
Cut (divP c (ppc a b)); [Intros div1 | Idtac].
Apply minuspf_comp; Auto.
Apply multm_lm_comp; Auto.
Apply eqTerm_sym; Apply divTerm_compo; Auto.
Apply divP_pccl; Auto.
Apply multm_lm_comp; Auto.
Apply eqTerm_sym; Apply divTerm_compo; Auto.
Apply divP_pccr; Auto.
LApply (pcc_is_ppcm a b);
 [Intros H'7; LApply H'7; [Intros H'8; Inversion H'8; Clear H'7 | Clear H'7] |
   Idtac]; Auto.
Apply divP_inv1 with b := a; Auto.
Qed.

Mutual Inductive ReduStarConfluent[Q:(list poly); p:(list Term)]: Prop :=
     ReduStarConfluent0:
       ((r, s:(list Term)) (reducestar Q p r) -> (reducestar Q p s) ->(eqP r s)) ->
       (ReduStarConfluent Q p) .

Theorem not_double_canonical:
  (a:Term) (p:(list Term))~ (canonical (pX a (pX a p))).
Intros a p; Red; Intros H'; Try Exact H'.
LApply (canonical_imp_order p a a); [Intros H'3; Try Exact H'3 | Idtac]; Auto.
Absurd (order a a); Auto.
Qed.

Theorem reduce_inv:
  (Q:(list poly))
  (a, b:Term)
  (p, q:(list Term))
  (reduce Q (pX a p) (pX b q)) -> (eqTerm a b) -> (canonical (pX a p)) ->
  (reduce Q p q).
Intros Q a b p q H'; Inversion H'; Auto.
Intros eq Cap; Simpl in H4; Simpl in H2; Simpl in H0; Simpl in H.
LApply (canonical_spminusf_full_t a b0 nZb p q0);
 [Intros H'5; LApply H'5;
   [Intros H'6; LApply H'6; [Intros H'7; Clear H'6 H'5 | Clear H'6 H'5] |
     Clear H'5] | Idtac]; Auto.
Cut (canonical (pX b (pX b q))); Auto.
Intros H'0; Elim (not_double_canonical b q); Auto.
Apply eqp_imp_canonical with p := (pX a (spminusf a b0 nZb p q0)); Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Qed.

Theorem reducestar_trans:
  (Q:(list poly))
  (x, y, z:(list Term))
  (canonical x) -> (reduceplus Q x y) -> (reducestar Q y z) ->(reducestar Q x z).
Intros Q x y z H' H'0 H'1; Inversion H'1.
Apply reducestar0; Auto.
Apply reduceplus_trans with y := y; Auto.
Qed.

Theorem confl_under:
  (Q:(list poly))
  (a:Term)
  (p:(list Term))
  (canonical (pX a p)) ->
  ((q:(list Term)) (canonical q) -> (ltP q (pX a p)) ->(ReduStarConfluent Q q)) ->
  (r, s:(list Term))
  (reduce Q (pX a p) (pX a r)) -> (reduce Q (pX a p) (pX a s)) ->
  (r1, s1:(list Term))
  (reducestar Q (pX a r) r1) -> (reducestar Q (pX a s) s1) ->(eqP r1 s1).
Intros Q a p Op1 H'0 r s H'1 H'2 r1 s1 H'3 H'4.
Cut (canonical p); [Intros Op2 | Apply (canonical_imp_canonical p a); Auto].
Cut (canonical (pX a r));
 [Intros Op3 | Apply canonical_reduce with Q := Q p := (pX a p); Auto].
Cut (canonical r); [Intros Op4 | Apply (canonical_imp_canonical r a); Auto].
Cut (canonical (pX a s));
 [Intros Op5 | Apply canonical_reduce with Q := Q p := (pX a p); Auto].
Cut (canonical s); [Intros Op6 | Apply (canonical_imp_canonical s a); Auto].
LApply (H'0 p);
 [Intros H'5; LApply H'5; [Intros H'6; Elim H'6; Clear H'5 | Apply ltP_refl_pX] |
   Idtac]; Auto.
Intros H'5.
Elim (reduce0_reducestar Q s); [Intros s0 E | Auto].
Elim (reduce0_reducestar Q r); [Intros r0 E0 | Auto].
Cut (reduceplus Q (pX a r) (pX a r0));
 [Intros red1 | Apply reduceplus_skip; Auto; Elim E0]; Auto.
Cut (canonical (pX a r0));
 [Intros Op7 | Apply canonical_reduceplus with Q := Q p := (pX a r)]; Auto.
Cut (reduceplus Q (pX a s) (pX a s0));
 [Intros red2 | Apply reduceplus_skip; Auto; Elim E]; Auto.
Cut (canonical (pX a s0));
 [Intros Op8 | Apply canonical_reduceplus with Q := Q p := (pX a s)]; Auto.
Cut (eqP (pX a r0) (pX a s0)); [Intros eqP1 | Auto].
Elim (reduce0_reducestar Q (pX a r0)); [Intros t0 E1 | Auto].
Elim (reduce0_reducestar Q (pX a s0)); [Intros t1 E2 | Auto].
Cut (reducestar Q (pX a r) t0);
 [Intros red3 | Apply (reducestar_trans Q) with y := (pX a r0)]; Auto.
Cut (reducestar Q (pX a s) t1);
 [Intros red4 | Apply (reducestar_trans Q) with y := (pX a s0)]; Auto.
Cut (eqP t0 t1); [Intros eqP2 | Auto].
Apply eqp_trans with y := t0; Auto.
LApply (H'0 (pX a r));
 [Intros H'7; LApply H'7; [Intros H'8; Elim H'8; Clear H'7 | Clear H'7] | Idtac];
 Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply eqp_trans with y := t1; Auto.
LApply (H'0 (pX a s));
 [Intros H'7; LApply H'7; [Intros H'8; Elim H'8; Clear H'7 | Clear H'7] | Idtac];
 Auto.
Apply ltP_reduce with Q := Q; Auto.
LApply (reducestar_eqp_com Q (pX a r0) t0 (pX a s0) t0);
 [Intros H'11; LApply H'11;
   [Intros H'12; LApply H'12;
     [Intros H'13; LApply H'13;
       [Intros H'14; Clear H'13 H'12 H'11 | Clear H'13 H'12 H'11] |
       Clear H'12 H'11] | Clear H'11] | Idtac]; Auto.
LApply (H'0 (pX a s0));
 [Intros H'7; LApply H'7; [Intros H'8; Elim H'8; Clear H'7 | Clear H'7] | Idtac];
 Auto.
Elim (order_reduceplus Q (pX a s) (pX a s0));
 [Intros H'13; Try Exact H'13 | Intros H'13 | Idtac | Idtac]; Auto.
Apply ltP_trans with y := (pX a s); Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply ltp_eqp_comp with p := (pX a s) q := (pX a p); Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply eqpP1; Auto.
Apply H'5; Auto.
Inversion E0; Auto.
LApply (reduce_inv Q a a p r);
 [Intros H'10; LApply H'10; [Intros H'11; Clear H'10 | Clear H'10] | Idtac];
 Auto.
Apply reducestar0; Auto.
EApply Rstar_n; EAuto.
Inversion E; Auto.
LApply (reduce_inv Q a a p s);
 [Intros H'10; LApply H'10; [Intros H'11; Clear H'10 | Clear H'10] | Idtac];
 Auto.
Apply reducestar0; Auto; EApply Rstar_n; EAuto.
Qed.

Mutual Inductive Spoly_1[Q:(list poly)]: (list Term) -> (list Term) ->Prop :=
     Spoly_10:
       (p, q:(list Term))
       (Cp:(canonical p))
       (Cq:(canonical q)) (reducestar Q (spolyf p q Cp Cq) pO) ->(Spoly_1 Q p q)
    | Spoly_11:
        (b, c, d:Term)
        (q, r, r0, s0, t:(list Term))
        (Spoly_1 Q (pX c r) (pX d t)) ->
        (Spoly_1 Q (pX d t) (pX b q)) ->
        (inPolySet (pX d t) Q) -> (divP (ppc c b) d) ->
        (Spoly_1 Q (pX c r) (pX b q)) .

Mutual Inductive SpolyQ[Q:(list poly)]: Prop :=
     SpolyQ0:
       ((p, q:(list Term))
        (inPolySet p Q) -> (canonical p) -> (inPolySet q Q) -> (canonical q) ->
        (Spoly_1 Q p q)) ->(SpolyQ Q) .

Mutual Inductive Reducep[Q:(list poly)]:
                   (list Term) -> (list Term) -> (list Term) ->Prop :=
     Reducep0:
       (a, b, c:Term)
       (nZb:~ (zeroP b))
       (nZc:~ (zeroP c))
       (p, q, r:(list Term))
       (reduce Q (pX a p) (spminusf a b nZb p q)) ->
       (reduce Q (pX a p) (spminusf a c nZc p r)) ->
       (divP a b) ->
       (inPolySet (pX b q) Q) -> (divP a c) -> (inPolySet (pX c r) Q) ->
       (Reducep Q (pX a p) (pX b q) (pX c r)) .

Mutual Inductive Confluent[Q:(list poly)]:
                   (list Term) -> (list Term) -> (list Term) ->Prop :=
     Confluent0:
       (a, b, c:Term)
       (nZb:~ (zeroP b))
       (nZc:~ (zeroP c))
       (p, q, r:(list Term))
       ((r1, s1:(list Term))
        (reducestar Q (spminusf a b nZb p q) r1) ->
        (reducestar Q (spminusf a c nZc p r) s1) ->(eqP r1 s1)) ->
       (Confluent Q (pX a p) (pX b q) (pX c r)) .

Theorem sp_rew:
  (a, b:Term) (nZ1, nZ2:~ (zeroP b)) (p, q:(list Term))
  (spminusf a b nZ1 p q)=(spminusf a b nZ2 p q).
Auto; Auto.
Intros a b nZ1 nZ2 p q; Unfold spminusf.
Change (minuspf p (multm_lm (divTerm a b nZ1) q))=
       (minuspf p (multm_lm (divTerm a b nZ2) q)).
Rewrite (divTerm_rew a b nZ1 nZ2); Auto.
Qed.

Theorem Conf_inv1:
  (Q:(list poly))
  (a, b, c:Term)
  (nZb:~ (zeroP b))
  (nZc:~ (zeroP c))
  (p, q, r, r1, s1:(list Term))
  (Confluent Q (pX a p) (pX b q) (pX c r)) ->
  (reducestar Q (spminusf a b nZb p q) r1) ->
  (reducestar Q (spminusf a c nZc p r) s1) ->(eqP r1 s1).
Intros Q a b c nZb nZc p q r r1 s1 H'0 H'1 H'2; Inversion H'0.
Apply H0.
Simpl in H; Simpl in H0; Simpl in H1; Simpl in H2; Simpl in H3; Simpl in H4.
Generalize nZb0; Rewrite H2; Rewrite H1; Rewrite H; Rewrite H3.
Intros nZb1.
Rewrite (sp_rew a b nZb1 nZb p q); Auto.
Simpl in H; Simpl in H0; Simpl in H1; Simpl in H2; Simpl in H3; Simpl in H4;
 Simpl in H5.
Generalize nZc0; Rewrite H4; Rewrite H; Rewrite H1; Rewrite H5.
Intros nZb1; Rewrite (sp_rew a c nZb1 nZc p r); Auto.
Qed.

Theorem Conf_trans:
  (Q:(list poly))
  (a, b, c, d:Term)
  (p0, q, r, t:(list Term))
  (Reducep Q (pX a p0) (pX c r) (pX d t)) ->
  (Reducep Q (pX a p0) (pX d t) (pX b q)) ->
  (Confluent Q (pX a p0) (pX c r) (pX d t)) ->
  (Confluent Q (pX a p0) (pX d t) (pX b q)) ->
  (canonical (pX a p0)) -> (canonical (pX d t)) ->
  (Confluent Q (pX a p0) (pX c r) (pX b q)).
Intros Q a b c d p q r t H' H'0 H'1 H'2 H'3 H'4.
Cut ~ (zeroP b); [Intros nZb | Idtac].
Cut ~ (zeroP c); [Intros nZc | Idtac].
Cut ~ (zeroP d); [Intros nZd | Idtac].
Apply Confluent0 with nZb := nZc nZc := nZb; Auto.
Intros r1 s1 H'5 H'6.
Elim (reduce0_reducestar Q (spminusf a d nZd p t)); [Intros t0 E | Idtac]; Auto.
Apply eqp_trans with y := t0.
Apply Conf_inv1
     with Q := Q a := a b := c c := d nZb := nZc nZc := nZd p := p q := r r := t;
 Auto.
Apply Conf_inv1
     with Q := Q a := a b := d c := b nZb := nZd nZc := nZb p := p q := t r := q;
 Auto.
Apply canonical_spminusf; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_imp_canonical with a := d; Auto.
Simple Inversion H'0; Auto.
Generalize nZb0; Try Rewrite (pX_invl a0 a p0 p);
 Try Rewrite (pX_invl b0 d q0 t); Auto.
Apply canonical_nzeroP with p := t; Auto.
Simple Inversion H'; Auto.
Rewrite H6.
Intros H'5 H'6 H'7 H'8 H'9 H'10.
Apply canonical_nzeroP with p := r; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Simple Inversion H'0; Auto.
Rewrite H7.
Intros H'5 H'6 H'7 H'8 H'9 H'10.
Apply canonical_nzeroP with p := q; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Qed.

Theorem confl_top_simpl:
  (Q:(list poly))
  (a, b, c:Term)
  (nZb:~ (zeroP b))
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (Cpxc:(canonical (pX c r)))
  (Cpxb:(canonical (pX b q)))
  (reducestar Q (spolyf (pX c r) (pX b q) Cpxc Cpxb) pO) ->
  (canonical (pX a p)) ->
  (divP a b) ->
  (inPolySet (pX b q) Q) ->
  (divP a c) ->
  (inPolySet (pX c r) Q) ->
  ((q:(list Term)) (canonical q) -> (ltP q (pX a p)) ->(ReduStarConfluent Q q)) ->
  (reduce Q (pX a p) (spminusf a b nZb p q)) ->
  (reduce Q (pX a p) (spminusf a c nZc p r)) ->
  (r1, s1:(list Term))
  (reducestar Q (spminusf a b nZb p q) r1) ->
  (reducestar Q (spminusf a c nZc p r) s1) ->(eqP r1 s1).
Intros Q a b c nZb nZc p q r Cpxc Cpxb red_r0 Cpxa divP_ab in_bp_Q divP_ac
       in_cr_Q Rec red1 red2 r1 s1 reds1 reds2.
Cut (canonical p); [Intros Op1 | EAuto].
Cut (canonical r); [Intros Op2 | EAuto].
Cut (canonical q); [Intros Op3 | EAuto].
Cut ~ (zeroP a); [Intros nZa | Idtac].
Cut ~ (zeroP (ppc c b)); [Intros nZppcd | Idtac].
Cut (eqP
       (multm_lm
          (divTerm a (ppc c b) nZppcd) (spolyf (pX c r) (pX b q) Cpxc Cpxb))
       (minuspf (spminusf a b nZb p q) (spminusf a c nZc p r)));
 [Intros red3 | Apply eqTerm_spolyf_red2; Auto].
2:Apply ppc_nZ; Auto.
2:Apply canonical_nzeroP with p := p; Auto.
Inversion red_r0.
LApply (reduceplus_multm
          Q (divTerm a (ppc c b) nZppcd) (spolyf (pX c r) (pX b q) Cpxc Cpxb) pO);
 [Intros H'3; LApply H'3;
   [Intros H'4; LApply H'4; [Intros H'5; Clear H'4 H'3 | Clear H'4 H'3] |
     Clear H'3] | Idtac]; Auto.
Elim (red_minus_zero_reduce Q (spminusf a b nZb p q) (spminusf a c nZc p r));
 [Intros r2 E; Elim E; Intros H'5 H'6; Clear E | Idtac | Idtac | Idtac]; Auto.
Elim (reduce0_reducestar Q r2); [Intros t0 E | Idtac]; Auto.
LApply (Rec (spminusf a b nZb p q));
 [Intros H'1; LApply H'1; [Intros H'2; Elim H'2; Clear H'1 | Clear H'1] | Idtac];
 Auto.
Intros H'.
LApply (Rec (spminusf a c nZc p r));
 [Intros H'1; LApply H'1; [Intros H'3; Elim H'3; Clear H'1 | Clear H'1] | Idtac];
 Auto.
Intros H'1.
Apply eqp_trans with y := t0.
Apply H'; Auto.
Auto; Apply (reducestar_trans Q) with y := r2; Auto.
Apply H'1; Auto.
Apply (reducestar_trans Q) with y := r2; Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply canonical_reduceplus with Q := Q p := (spminusf a b nZb p q); Auto.
Apply reduceplus_eqp_com
     with p :=
          (multm_lm
             (divTerm a (ppc c b) nZppcd) (spolyf (pX c r) (pX b q) Cpxc Cpxb))
          q := (multm_lm (divTerm a (ppc c b) nZppcd) pO); Auto; Simpl; Auto.
Qed.

Theorem reducetop_sp:
  (Q:(list poly))
  (a, b:Term)
  (nZb:~ (zeroP b)) (p, q:(list Term)) (inPolySet (pX b q) Q) -> (divP a b) ->
  (reduce Q (pX a p) (spminusf a b nZb p q)).
Intros Q a b nZb p q H' H'0.
Apply reducetop with b := b nZb := nZb q := q; Auto.
Qed.
Hints Resolve reducetop_sp.

Theorem fconfl_top:
  (Q:(list poly)) (q, r:(list Term)) (Spoly_1 Q q r) ->
  (p:(list Term))
  (Reducep Q p q r) ->
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  ((q0:(list Term)) (canonical q0) -> (ltP q0 p) ->(ReduStarConfluent Q q0)) ->
  (Confluent Q p q r).
Intros Q q r Spoly1; Elim Spoly1; Clear Spoly1 q r.
Intros q r Cq Cr red0 p H'1;
 (Generalize Cq Cr red0; Inversion H'1; Clear red0 Cq Cr H'1).
Intros Cq Cr H' H'0 H'1 H'2 H'3.
Apply Confluent0 with nZb := nZb nZc := nZc; Auto.
Intros r1 s1 H'5 H'6.
Apply eqp_sym.
Apply (confl_top_simpl Q a c b nZc nZb p0 r0 q0 Cq Cr); Auto.
Intros b c d q r H' H'0 t H'1 H'2 H'3 H'4 H'5 H'6 p H'7 H'8 H'9 H'10 H'11.
Generalize H'11 H'8; Clear H'11 H'8; Inversion H'7.
Simpl in H2; Simpl in H1; Simpl in H0; Simpl in H.
Intros H'8 H'11.
Cut (divP a d); [Intros div0 | Idtac].
2:Apply divP_trans with y := (ppc c b); Auto.
2:Rewrite <- H; Rewrite <- H1; Auto.
2:LApply (pcc_is_ppcm b0 c0);
   [Intros H'15; LApply H'15; [Intros H'16; Elim H'16; Clear H'15 | Clear H'15] |
     Idtac]; Auto.
Cut ~ (zeroP d); [Intros nZd | Inversion div0]; Auto.
Cut (Reducep Q (pX a p0) (pX d t) (pX b q)); [Intros Red0 | Idtac].
2:Rewrite <- H1; Rewrite <- H2; Auto.
2:Apply Reducep0 with nZb := nZd nZc := nZc; Auto.
Cut (Reducep Q (pX a p0) (pX c r) (pX d t)); [Intros Red1 | Idtac].
2:Rewrite <- H; Rewrite <- H0; Auto.
2:Apply Reducep0 with nZb := nZb nZc := nZd; Auto.
Apply Conf_trans with d := d t := t; Auto.
Apply H'2; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Apply H'4; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Qed.

Theorem confl_top:
  (Q:(list poly)) (SpolyQ Q) ->
  (a, b, c:Term)
  (nZb:~ (zeroP b))
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical (pX a p)) ->
  (divP a b) ->
  (inPolySet (pX b q) Q) ->
  (divP a c) ->
  (inPolySet (pX c r) Q) ->
  ((q:(list Term)) (canonical q) -> (ltP q (pX a p)) ->(ReduStarConfluent Q q)) ->
  (reduce Q (pX a p) (spminusf a b nZb p q)) ->
  (reduce Q (pX a p) (spminusf a c nZc p r)) ->
  (r1, s1:(list Term))
  (reducestar Q (spminusf a b nZb p q) r1) ->
  (reducestar Q (spminusf a c nZc p r) s1) ->(eqP r1 s1).
Intros Q H' a b c nZb nZc p q r H'0 H'1 H'2 H'3 H'4 H'5 H'6 H'7 r1 s1 H'8 H'9.
Cut (canonical (pX b q)); [Intros Op0 | Idtac].
Cut (canonical (pX c r)); [Intros Op0bis | Idtac].
Cut (Confluent Q (pX a p) (pX b q) (pX c r)); Auto.
2:Apply fconfl_top; Auto.
2:Inversion H'; Auto.
2:Apply Reducep0 with nZb := nZb nZc := nZc; Auto.
2:Apply inPolySet_imp_canonical with L := Q; Auto.
2:Apply inPolySet_imp_canonical with L := Q; Auto.
Intros H'10.
Apply (Conf_inv1 Q a b c nZb nZc p q r r1 s1); Auto.
Qed.

Theorem rew_spminusf:
  (a, b:Term) (nZb:~ (zeroP b)) (p, q:(list Term))
  (spminusf a b nZb p q)=(minuspf p (multm_lm (divTerm a b nZb) q)).
Simpl; Auto.
Qed.

Theorem confl_mix:
  (Q:(list poly))
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical (pX a p)) ->
  (divP a b) ->
  (inPolySet (pX b r) Q) ->
  ((q:(list Term)) (canonical q) -> (ltP q (pX a p)) ->(ReduStarConfluent Q q)) ->
  (reduce Q (pX a p) (pX a q)) -> (reduce Q (pX a p) (spminusf a b nZb p r)) ->
  (r1, s1:(list Term))
  (reducestar Q (pX a q) r1) -> (reducestar Q (spminusf a b nZb p r) s1) ->
  (eqP r1 s1).
Intros Q a b nZb p q r Op1 dviP_ab in_br Rec red1 red2 r1 s1 red3 red4.
Cut (canonical (pX b r));
 [Intros Op2 | Apply inPolySet_imp_canonical with L := Q; Auto].
LApply (reduce_inv Q a a p q);
 [Intro H'4; LApply H'4;
   [Intro H'5; LApply H'5; [Intro H'6; Clear H'5 H'4 | Clear H'5 H'4] |
     Clear H'4] | Idtac]; Auto.
Cut (canonical p); [Intros Op3 | EAuto].
Cut (canonical r); [Intros Op4 | EAuto].
Cut (canonical q);
 [Intros Op5 | Apply canonical_reduce with Q := Q p := p; Auto].
LApply (one_minus_reduceplus Q p q);
 [Intro H'2; Elim (H'2 (multm_lm (divTerm a b nZb) r));
   [Intros s E; Elim E; Intros H'7 H'8; Clear E | Idtac | Idtac] | Idtac]; Auto.
Cut (canonical s);
 [Intros Op6 |
   Apply canonical_reduceplus
        with Q := Q p := (minuspf p (multm_lm (divTerm a b nZb) r)); Auto].
Elim (reduce0_reducestar Q s); [Intros t E | EAuto]; Auto.
Apply eqp_trans with y := t; Auto.
LApply (Rec (pX a q));
 [Intros H'0; LApply H'0; [Intros H'1; Elim H'1; Clear H'0 | Clear H'0] | Idtac];
 Auto.
Intros PH'0; Apply PH'0; Auto.
Apply reducestar_trans with y := s; Auto.
Apply canonical_reduce with Q := Q p := (pX a p); Auto.
Apply Rstar_n with y := (minuspf q (multm_lm (divTerm a b nZb) r)); Auto.
Apply reducetop with b := b nZb := nZb q := r; Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply canonical_reduce with Q := Q p := (pX a p); Auto.
LApply (Rec (spminusf a b nZb p r));
 [Intros H'0; LApply H'0; [Intros H'1; Elim H'1; Clear H'0 | Clear H'0] | Idtac];
 Auto.
Intros PH'0; Apply PH'0; Auto.
Apply reducestar_trans with y := s; Auto.
Apply ltP_reduce with Q := Q; Auto.
Qed.
Hints Resolve pO_irreducible.

Theorem reducestar_reduceplus:
  (Q:(list poly)) (p, q:(list Term)) (reducestar Q p q) ->(reduceplus Q p q).
Intros Q p q H'; Inversion H'; Auto.
Qed.

Theorem reducestar_irreducible:
  (Q:(list poly)) (p, q:(list Term)) (reducestar Q p q) ->(irreducible Q q).
Intros Q p q H'; Inversion H'; Auto.
Qed.
Hints Resolve reducestar_reduceplus.

Theorem reducestar_inv:
  (Q:(list poly)) (p, q:(list Term)) (reducestar Q p q) -> (canonical p) ->
  (eqP p q) /\ (irreducible Q p) \/
  (Ex [r:(list Term)] (reduce Q p r) /\ (reducestar Q r q)).
Intros Q p q H'; Elim H'.
Intros p0 q0 H'0; Inversion H'0.
Intros H'1 H'2; Left; Split; Auto.
Apply irreducible_eqp_com with p := q0; Auto.
Apply eqp_imp_canonical with p := p0; Auto.
Apply eqp_sym; Auto.
Intros H'1 H'2; Right; Exists y; Split; Auto.
Apply reducestar0; Auto.
Qed.

Theorem confl_restar:
  (Q:(list poly)) (SpolyQ Q) ->(p:poly)(ReduStarConfluent Q (s2p p)).
Intros Q H'.
Intros p; Pattern p; Apply (well_founded_ind poly sltP sltp_wf); Auto.
Intros x; Elim x; Clear x.
Intros x Op_x Rec; Simpl.
Apply ReduStarConfluent0; Auto.
Intros r s H'1 H'2.
Elim (reducestar_inv Q x s);
 [Intros H'8; Elim H'8; Intros r0 E |
   Intros H'8; Elim H'8; Intros r0 E; Elim E; Intros H'9 H'10; Try Exact H'9;
    Clear E H'8 | Idtac | Idtac]; Auto.
Elim (reducestar_inv Q x r);
 [Intros H'9; Elim H'9; Intros r1 E0 |
   Intros H'9; Elim H'9; Intros r1 E0; Elim E0; Intros H'10 H'11;
    Try Exact H'11; Clear E0 H'9 | Idtac | Idtac]; Auto.
Apply eqp_trans with y := x; Auto; Apply eqp_sym; Auto.
Absurd (reduce Q x r1); Auto.
Elim (reducestar_inv Q x r);
 [Intros H'8; Elim H'8; Intros r1 E |
   Intros H'8; Elim H'8; Intros r1 E; Elim E; Intros H'11 H'12; Clear E H'8 |
   Idtac | Idtac]; Auto.
Absurd (reduce Q x r0); Auto.
Generalize H'12 H'9 H'10 Op_x Rec; Clear H'12 H'9 H'10 Rec Op_x; Inversion H'11.
Intros H'3 H'4 H'5 Op_x H'6.
Generalize Op_x H'11 H'6 H'5 H'3; Clear H'11 H'5 H'3 H'6 Op_x.
Rewrite <- H2.
Inversion H'4.
Intros H'3 H'5 H'6 H'7 H'8.
Apply confl_top
     with Q := Q a := a b := b c := b0 nZb := nZb nZc := nZb0 p := p0 q := q
          r := q0; Auto.
Intros q1 H'9 H'10.
Generalize (H'6 (mks ? ? T2M zeroP zeroP_dec ltM q1 H'9)); Simpl; Auto.
Apply reducestar_eqp_com with p := r1 q := r; Auto.
Apply canonical_reduce with Q := Q p := (pX a p0); Auto.
Apply eqp_sym; Auto.
Apply reducestar_eqp_com with p := r0 q := s; Auto.
Apply canonical_reduce with Q := Q p := (pX a p0); Auto.
Apply eqp_sym; Auto.
Intros H'3 H'5 H'6 H'7 H'8.
Apply eqp_sym.
Apply confl_mix with Q := Q a := a b := b nZb := nZb p := p0 q := q0 r := q;
 Auto.
Intros q1 H'9 H'10.
Generalize (H'6 (mks ? ? T2M zeroP zeroP_dec ltM q1 H'9)); Simpl; Auto.
Apply reducestar_eqp_com with p := (pX b0 q0) q := s; Auto.
Apply eqp_imp_canonical with p := (pX a q0); Auto.
Apply canonical_reduce with Q := Q p := (pX a p0); Auto.
Apply reducestar_eqp_com with p := r1 q := r; Auto.
Apply canonical_reduce with Q := Q p := (pX a p0); Auto.
Apply eqp_sym; Auto.
Intros H'3 H'4 H'5 Op_x H'6.
Generalize H'11 Op_x H'6 H'5 H'3; Clear H'11 H'6 Op_x H'5 H'3.
Inversion H'4.
Intros H'3 Op_x H'5 H'6 H'7.
Apply (confl_mix Q a b0 nZb p0 q q0); Auto.
Intros q1 H'9 H'10.
Generalize (H'5 (mks ? ? T2M zeroP zeroP_dec ltM q1 H'9)); Simpl; Auto.
Apply reducestar_eqp_com with p := (pX b q) q := r; Auto.
Apply canonical_reduce with Q := Q p := (pX a p0); Auto.
Apply reducestar_eqp_com with p := r0 q := s; Auto.
Apply canonical_reduce with Q := Q p := (pX a p0); Auto.
Apply eqp_sym; Auto.
Intros H'3 Op_x H'5 H'6 H'7.
Apply confl_under with Q := Q a := a p := p0 r := q s := q0; Auto.
Intros q1 H'9 H'10.
Generalize (H'5 (mks ? ? T2M zeroP zeroP_dec ltM q1 H'9)); Simpl; Auto.
Apply reducestar_eqp_com with p := (pX b q) q := r; Auto.
Apply canonical_reduce with Q := Q p := (pX a p0); Auto.
Apply reducestar_eqp_com with p := (pX b0 q0) q := s; Auto.
Apply canonical_reduce with Q := Q p := (pX a p0); Auto.
Qed.

Theorem confl_reducestar:
  (Q:(list poly)) (SpolyQ Q) ->
  (p:(list Term)) (canonical p) ->(ReduStarConfluent Q p).
Intros Q H' p H'0.
Generalize (confl_restar Q H' (mks ? ? T2M zeroP zeroP_dec ltM p H'0)); Simpl;
 Auto.
Qed.

Mutual Inductive CombLinear[Q:(list poly)]: (list Term) ->Prop :=
     CombLinear0: (CombLinear Q pO)
    | CombLinear1:
        (a:Term)
        (p, q, s:(list Term))
        ~ (zeroP a) ->
        (inPolySet q Q) ->
        (CombLinear Q p) -> (eqP s (pluspf (multm_lm a q) p)) ->(CombLinear Q s) .
Hints Resolve CombLinear0 CombLinear1.

Theorem CombLinear_canonical:
  (Q:(list poly)) (p:(list Term)) (CombLinear Q p) ->(canonical p).
Intros Q p H'; Elim H'; Auto.
Intros a p0 q s H'0 H'1 H'2 H'3 H'4.
Apply eqp_imp_canonical with p := (pluspf (multm_lm a q) p0); Auto.
Apply eqp_sym; Auto.
Apply canonical_pluspf; Auto.
Apply canonical_multlm_lm; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Qed.

Theorem CombLinear_comp:
  (Q:(list poly)) (p:(list Term)) (CombLinear Q p) ->
  (q:(list Term)) (canonical q) -> (eqP p q) ->(CombLinear Q q).
Intros Q p H'; Elim H'; Auto.
Intros q H'0 H'1; Inversion H'1; Auto.
Intros a p0 q s H'0 H'1 H'2 H'3 H'4 q0 H'5 H'6.
Cut (canonical q); [Intros C0 | Idtac].
Cut (canonical s); [Intros C1 | Idtac].
Apply CombLinear1 with a := a p := p0 q := q; Auto.
Apply eqp_trans with y := s; Auto; Apply eqp_sym; Auto.
Apply eqp_imp_canonical with p := q0; Auto.
Apply eqp_sym; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Qed.
Hints Resolve CombLinear_canonical.

Theorem CombLinear_pluspf:
  (Q:(list poly)) (p:(list Term)) (CombLinear Q p) ->
  (q:(list Term)) (CombLinear Q q) ->(CombLinear Q (pluspf p q)).
Intros Q p H'; Elim H'; Auto.
Intros q H'0.
Cut (canonical q); [Intros Op1 | Apply CombLinear_canonical with Q := Q; Auto].
Apply CombLinear_comp with p := q; Auto.
Apply eqp_trans with y := (pluspf pO q); Auto; Apply eqp_sym; Auto.
Intros a p0 q s H'0 H'1 H'2 H'3 H'4 q0 H'5.
Cut (canonical q); [Intros C0 | Apply inPolySet_imp_canonical with L := Q; Auto].
Cut (canonical q0); [Intros Op1 | Apply CombLinear_canonical with Q := Q; Auto].
Cut (canonical p0); [Intros Op2 | Apply CombLinear_canonical with Q := Q; Auto].
Apply CombLinear1 with a := a p := (pluspf p0 q0) q := q; Auto.
Apply eqp_trans with y := (pluspf (pluspf (multm_lm a q) p0) q0); Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_imp_canonical with p := (pluspf (multm_lm a q) p0); Auto.
Apply eqp_sym; Auto.
Qed.
Hints Resolve CombLinear_pluspf.

Theorem CombLinear_multm_lm:
  (Q:(list poly)) (a:Term) (p:(list Term)) ~ (zeroP a) -> (CombLinear Q p) ->
  (CombLinear Q (multm_lm a p)).
Intros Q a p H' H'0; Elim H'0; Auto.
Intros a0 p0 q s H'1 H'2 H'3 H'4 H'5.
Cut (canonical q); [Intros C0 | Apply inPolySet_imp_canonical with L := Q; Auto].
Cut (canonical p0); [Intros Op1 | Apply CombLinear_canonical with Q := Q; Auto].
Apply CombLinear1 with a := (multTerm a a0) p := (multm_lm a p0) q := q; Auto.
Apply eqp_trans with y := (multm_lm a (pluspf (multm_lm a0 q) p0)); Auto.
Apply eqp_trans with y := (pluspf (multm_lm a (multm_lm a0 q)) (multm_lm a p0));
 Auto.
Apply eqp_sym; Auto.
Qed.
Hints Resolve CombLinear_multm_lm.

Theorem CombLinear_minuspf:
  (Q:(list poly)) (p:(list Term)) (CombLinear Q p) ->
  (q:(list Term)) (CombLinear Q q) ->(CombLinear Q (minuspf p q)).
Intros Q p H' q H'0; Try Assumption.
Cut (canonical p); [Intros Op1 | Apply CombLinear_canonical with Q := Q; Auto].
Cut (canonical q); [Intros Op2 | Apply CombLinear_canonical with Q := Q; Auto].
Apply CombLinear_comp with p := (pluspf p (multm_lm (invTerm T1) q)); Auto.
Apply eqp_sym; Auto.
Qed.
Hints Resolve CombLinear_minuspf.

Theorem CombLinear_id:
  (Q:(list poly)) (p:(list Term)) (inPolySet p Q) ->(CombLinear Q p).
Intros Q p H'.
Cut (canonical p); [Intros C0 | Apply inPolySet_imp_canonical with L := Q; Auto].
Apply CombLinear_comp with p := (pluspf (multm_lm T1 p) pO); Auto.
Apply CombLinear1 with a := T1 p := pO q := p; Auto.
Apply eqp_trans with y := (multm_lm T1 p); Auto.
Qed.
Hints Resolve CombLinear_id.

Theorem CombLinear_spoly:
  (Q:(list poly))
  (p, q:(list Term))
  (Cp:(canonical p)) (Cq:(canonical q)) (inPolySet p Q) -> (inPolySet q Q) ->
  (CombLinear Q (spolyf p q Cp Cq)).
Intros Q p; Case p.
Simpl; Auto.
Intros a l q; Case q.
Simpl; Auto.
Intros a0 l0 Cp Cq H' H'0.
Cut (canonical l0); [Intros Op1 | Apply canonical_imp_canonical with a := a0];
 Auto.
Cut (canonical l); [Intros Op2 | Apply canonical_imp_canonical with a := a];
 Auto.
Cut ~ (zeroP a); [Intros nZa | Apply canonical_nzeroP with p := l; Auto].
Cut ~ (zeroP a0); [Intros nZa0 | Apply canonical_nzeroP with p := l0; Auto].
Apply CombLinear_comp
     with p :=
          (minuspf
             (multm_lm (divTerm (ppc a a0) a nZa) (pX a l))
             (multm_lm (divTerm (ppc a a0) a0 nZa0) (pX a0 l0))); Auto.
Apply eqp_sym; Auto.
Apply spoly_is_minus; Auto.
Qed.

Theorem plusTerm_is_pX:
  (a:Term) (p:(list Term)) (canonical (pX a p)) ->
  (eqP (pluspf (pX a pO) p) (pX a p)).
Intros a p; Case p; Auto.
Intros a0 l H'.
Apply eqp_trans with y := (pX a (pluspf pO (cons a0 l))); Auto.
Apply eqp_sym; Apply pluspf_inv1; Auto.
Apply canonical_imp_order with l := l; Auto.
Qed.
Hints Resolve plusTerm_is_pX.

Theorem reduce_inv2:
  (Q:(list poly)) (p, q:(list Term)) (reduce Q p q) -> (canonical p) ->
  (Ex [x:(list Term)]
  (Ex [a:Term]
  ~ (zeroP a) /\
  ((inPolySet x Q) /\ ((canonical x) /\ (eqP q (minuspf p (multm_lm a x))))))).
Intros Q p q H'; Elim H'; Auto.
Intros a b nZb p0 q0 r H'0 H'1 H'2 H'3; Cut (canonical (pX b q0));
 [Intros C1 | Apply inPolySet_imp_canonical with L := Q; Auto];
 Exists (pX b q0); Exists (divTerm a b nZb); Split;
 [Idtac | Split; [Idtac | Split; [Idtac | Idtac]]]; Auto.
Apply eqp_trans with y := (spminusf a b nZb p0 q0); Auto.
Apply eqp_sym; Auto.
Apply eqp_trans with y := (spminusf a b nZb (pX a p0) (pX b q0)); Auto.
Intros a b p0 q0 H'0 H'1 H'2 H'3.
(Cut (canonical p0); [Intros Op1 | Apply canonical_imp_canonical with a := a]);
 Auto.
Cut (canonical (pX a q0));
 [Intros Op2 | Apply canonical_reduce with Q := Q p := (pX a p0)]; Auto.
Cut (canonical q0); [Intros Op3 | Apply canonical_imp_canonical with a := a];
 Auto.
Elim H'1;
 [Intros x E; Elim E; Intros a0 E0; Elim E0; Intros H'4 H'5; Elim H'5;
   Intros H'6 H'7; Elim H'7; Intros H'8 H'9; Clear H'7 H'5 E0 E H'1 | Clear H'1];
 Auto.
Exists x; Exists a0; Split; [Idtac | Split; [Idtac | Split]]; Auto.
Apply eqp_trans with y := (pX a q0); [Auto | Apply eqp_sym].
Apply eqp_trans
     with y := (pluspf (pX a p0) (multm_lm (invTerm T1) (multm_lm a0 x)));
 [Auto | Idtac].
Cut ~ (zeroP a); [Intros Z0 | Idtac].
Apply eqp_trans
     with y :=
          (pluspf (pluspf (pX a pO) p0) (multm_lm (invTerm T1) (multm_lm a0 x))).
Apply eqp_sym; Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf (pX a pO) (pluspf p0 (multm_lm (invTerm T1) (multm_lm a0 x))));
 [Auto | Idtac].
Apply eqp_trans with y := (pluspf (pX a pO) (minuspf p0 (multm_lm a0 x))).
Apply eqp_sym; Auto.
Apply eqp_trans with y := (pluspf (pX a pO) q0); Auto.
Apply eqp_sym; Auto.
Apply canonical_nzeroP with p := q0; Auto.
Qed.

Theorem CombLinear_reduce:
  (Q:(list poly))
  (p, q:(list Term)) (reduce Q p q) -> (canonical p) -> (CombLinear Q q) ->
  (CombLinear Q p).
Intros Q p q H'0 H'1 H'2.
Elim (reduce_inv2 Q p q);
 [Intros x E; Elim E; Intros a E0; Elim E0; Intros H'7 H'8; Elim H'8;
   Intros H'9 H'10; Elim H'10; Intros H'11 H'12; Clear H'10 H'8 E0 E | Idtac |
   Idtac]; Auto.
Apply CombLinear1 with a := a p := q q := x; Auto.
Apply eqp_sym.
Cut (canonical q); [Intros Op1 | Apply canonical_reduce with Q := Q p := p];
 Auto.
Apply eqp_trans with y := (pluspf (multm_lm a x) (minuspf p (multm_lm a x)));
 Auto.
Apply eqp_trans with y := (pluspf (minuspf p (multm_lm a x)) (multm_lm a x));
 Auto.
Qed.

Theorem CombLinear_reduceplus:
  (Q:(list poly))
  (p, q:(list Term)) (reduceplus Q p q) -> (canonical p) -> (CombLinear Q q) ->
  (CombLinear Q p).
Intros Q p q H'0; Elim H'0; Auto.
Intros x y H' H'1 H'2.
Apply CombLinear_comp with p := y; Auto.
Apply eqp_sym; Auto.
Intros x y z H' H'1 H'2 H'3 H'4.
Apply CombLinear_reduce with q := y; Auto.
Apply H'2; Auto.
Apply canonical_reduce with Q := Q p := x; Auto.
Qed.

Theorem CombLinear_reducestar:
  (Q:(list poly))
  (p, q:(list Term)) (reducestar Q p q) -> (canonical p) -> (CombLinear Q q) ->
  (CombLinear Q p).
Intros Q p q H'; Elim H'; Auto.
Intros p0 q0 H'0 H'1 H'2 H'3.
Apply CombLinear_reduceplus with q := q0; Auto.
Qed.

Theorem Reducestar_pO_imp_CombLinear:
  (Q:(list poly))
  (p, q:(list Term)) (reducestar Q p q) -> (canonical p) -> (eqP q pO) ->
  (CombLinear Q p).
Intros Q p q H' H'0 H'1; Inversion H'1; Auto.
Apply CombLinear_reducestar with q := q; Auto.
Rewrite <- H.
Apply CombLinear0; Auto.
Qed.

Mutual Inductive Grobner[Q:(list poly)]: Prop :=
     Grobner0:
       ((p, q:(list Term)) (CombLinear Q p) -> (reducestar Q p q) ->(eqP q pO)) ->
       (Grobner Q) .

Theorem Grobner_imp_SpolyQ: (Q:(list poly)) (Grobner Q) ->(SpolyQ Q).
Intros Q H'; Elim H'.
Intros H'1.
Apply SpolyQ0; Auto.
Intros p q H'0 H'2 H'3 H'4.
Elim (reduce0_reducestar Q (spolyf p q H'2 H'4)); [Intros t E | Idtac]; Auto.
Apply Spoly_10 with Cp := H'2 Cq := H'4; Auto.
Apply reducestar_eqp_com with p := (spolyf p q H'2 H'4) q := t; Auto.
Apply H'1 with p := (spolyf p q H'2 H'4); Auto.
Apply CombLinear_spoly; Auto.
Qed.

Mutual Inductive ConfluentReduce[Q:(list poly)]: Prop :=
     ConfluentReduce0:
       ((p:(list Term)) (canonical p) ->(ReduStarConfluent Q p)) ->
       (ConfluentReduce Q) .

Theorem SpolyQ_imp_ConfluentReduce:
  (Q:(list poly)) (SpolyQ Q) ->(ConfluentReduce Q).
Intros Q H'0.
Apply ConfluentReduce0.
Intros p H'1.
LApply (confl_restar Q);
 [Intros H'2;
   Specialize 1 H'2 with p := (mks ? ? T2M zeroP zeroP_dec ltM p H'1);
   Intros H'3 | Idtac]; Auto.
Qed.

Lemma pO_reduce: (Q:(list poly)) (p:(list Term))~ (reduce Q pO p).
Intros Q p; Red; Intros H'; Inversion H'; Auto.
Qed.

Lemma pO_reduceplus:
  (Q:(list poly)) (p, q:(list Term)) (reduceplus Q pO p) ->p=pO.
Intros Q p q H'; Inversion H'; Auto.
Inversion H; Auto.
Specialize 2 pO_reduce with Q := Q p := y; Intros H'1; Elim H'1; Auto.
Qed.

Lemma pO_reducestar: (Q:(list poly)) (p:(list Term)) (reducestar Q pO p) ->p=pO.
Intros Q p H'.
Rewrite (pO_reduceplus Q p pO); Auto.
Qed.

Theorem reduce_in_pO:
  (Q:(list poly)) (p:(list Term)) (inPolySet p Q) ->(reduce Q p pO).
Intros Q p; Case p; Simpl; Auto.
Intros H'; Elim (inPolySet_is_pX pO Q);
 [Intros a E; Elim E; Intros q E0; Inversion E0 | Idtac]; Auto.
Intros a l H'.
Cut (canonical (pX a l)); [Intros Op0 | Idtac].
Cut (canonical l); [Intros Op1 | Apply canonical_imp_canonical with a := a];
 Auto.
Cut ~ (zeroP a); [Intros Nza | Idtac].
Cut (spminusf a a Nza l l)=pO.
Intros H'0; Rewrite <- H'0.
Apply reducetop with b := a nZb := Nza q := l; Auto.
Cut (eqP (spminusf a a Nza l l) pO).
Intros H'0; Inversion H'0; Auto.
Apply eqp_trans with y := (minuspf l (multm_lm (divTerm a a Nza) l)); Auto.
Apply eqp_trans with y := (minuspf l (multm_lm T1 l)); Auto.
Apply eqp_trans with y := (minuspf l l); Auto.
Apply canonical_nzeroP with p := l; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Qed.

Theorem reducestar_pO_is_pO:
  (Q:(list poly)) (p, q:(list Term)) (canonical p) ->(reducestar Q pO pO).
Intros Q p H' H'0; Auto.
Apply reducestar0; Auto.
Qed.

Theorem reducestar_in_pO:
  (Q:(list poly)) (a:Term) (p:(list Term)) (inPolySet p Q) -> ~ (zeroP a) ->
  (reducestar Q (multm_lm a p) pO).
Intros Q a p H' H'0.
Cut (canonical p); [Intros Op0 | Idtac].
Apply reducestar_eqp_com with p := (multm_lm a p) q := (multm_lm a pO); Auto.
Apply reducestar0; Auto.
Apply reduceplus_multm; Auto.
Apply reduce_imp_reduceplus; Auto.
Apply reduce_in_pO; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Qed.

Theorem ConfluentReduce_imp_Grobner:
  (Q:(list poly)) (ConfluentReduce Q) ->(Grobner Q).
Intros Q H'; Elim H'.
Intros H'0.
Apply Grobner0; Auto.
Intros p q H'1; Generalize q; Clear q; Elim H'1.
Intros q H'2.
Rewrite (pO_reducestar Q q); Auto.
Intros a p0 q s H'2 H'3 H'4 H'5 H'6 q0 H'7.
Cut (canonical q); [Intros Op0 | Apply inPolySet_imp_canonical with L := Q];
 Auto.
Cut (canonical p0); [Intros Op2 | Apply CombLinear_canonical with Q := Q]; Auto.
Cut (canonical s); [Intros Op1 | Idtac].
Cut (canonical q0); [Intros Op2b | Idtac]; Auto.
LApply (reducestar_in_pO Q a q);
 [Intros H'11; LApply H'11; [Intros H'12; Clear H'11 | Clear H'11] | Idtac];
 Auto.
Elim (red_minus_zero_reduce Q s p0);
 [Intros r1 E; Elim E; Intros H'15 H'16; Clear E | Idtac | Idtac | Idtac]; Auto.
Elim (reduce0_reducestar Q r1); [Intros t E | Idtac]; Auto.
LApply (H'0 s); [Intros H'11; Inversion H'11 | Idtac]; Auto.
Apply H; Auto.
Apply reducestar_eqp_com with p := s q := t; Auto.
Apply reducestar_trans with y := r1; Auto.
Apply H'5; Auto.
Apply (reducestar_trans Q) with y := r1; Auto.
Apply canonical_reduceplus with Q := Q p := s; Auto.
Apply reduceplus_eqp_com with p := (multm_lm a q) q := pO; Auto.
Apply eqp_sym;
 Apply eqp_trans with y := (minuspf (pluspf (multm_lm a q) p0) p0); Auto.
Apply eqp_trans
     with y := (pluspf (pluspf (multm_lm a q) p0) (multm_lm (invTerm T1) p0));
 Auto.
Apply eqp_trans
     with y := (pluspf (multm_lm a q) (pluspf p0 (multm_lm (invTerm T1) p0)));
 Auto.
Apply eqp_trans with y := (pluspf (multm_lm a q) pO); Auto.
Apply canonical_reduceplus with Q := Q p := s; Auto.
Apply eqp_imp_canonical with p := (pluspf (multm_lm a q) p0); Auto.
Apply eqp_sym; Auto.
Qed.
End def_spoly.
