
(****************************************************************************
                                                                           
          Buchberger : final reduction algorithm                           
                                                                           
          Laurent Thery April 98 (revised Mai 98)                          
                                                                           
****************************************************************************)
Section Reduce.
Variable poly:Set.
Require Import PolyList.
Variable cb:(list poly) -> poly ->Prop.
Variable divp:poly -> poly ->Prop.
Variable reduce:(list poly) -> poly -> poly ->Prop.
Variable nf:(list poly) -> poly ->poly.
Variable stable:(list poly) -> (list poly) ->Prop.
Variable grobner:(list poly) ->Prop.
Variable zero:poly.
Variable zerop:poly ->Prop.
Variable zerop_dec:(p:poly){(zerop p)}+{~ (zerop p)}.
Require Import LetP.
Hypothesis cb_id:(L:(list poly)) (p:poly) (In p L) ->(cb L p).
Hypothesis cb_zerop:(L:(list poly)) (p:poly) (zerop p) ->(cb L p).
Hypothesis cb_incl:
           (L1, L2:(list poly)) (p:poly) (incl L1 L2) -> (cb L1 p) ->(cb L2 p).
Hypothesis nf_cb:(p:poly) (L:(list poly))(cb (cons p L) (nf L p)).
Hypothesis cb_trans:
           (L:(list poly)) (p, q:poly) (cb (cons p L) q) -> (cb L p) ->(cb L q).
Hypothesis cb_comp:
           (L1, L2:(list poly)) ((p:poly) (In p L1) ->(cb L2 p)) ->
           (q:poly) (cb L1 q) ->(cb L2 q).
Hypothesis cb_nf:(p:poly) (L:(list poly))(cb (cons (nf L p) L) p).
Hypothesis cb_compo:
           (p:poly) (L1:(list poly)) (cb L1 p) ->
           (L2:(list poly)) ((q:poly) (In q L1) ->(cb L2 q)) ->(cb L2 p).
Hypothesis zerop_elim_cb:
           (L:(list poly)) (p, q:poly) (zerop p) -> (cb (cons p L) q) ->(cb L q).
Hypothesis grobner_def:
           (L:(list poly)) (grobner L) ->
           (p:poly) (cb L p) ->(zerop p) \/ (Ex [q:poly] (reduce L p q)).
Hypothesis def_grobner:
           (L:(list poly))
           ((p:poly) (cb L p) ->(zerop p) \/ (Ex [q:poly] (reduce L p q))) ->
           (grobner L).
Hypothesis nf_div:
           (p:poly) (L:(list poly)) ~ (zerop p) -> ~ (zerop (nf L p)) ->
           (Ex [q:poly] (In q (cons (nf L p) L)) /\ ((divp p q) /\ ~ (zerop q))).
Hypothesis div_reduce:
           (p:poly)
           (L1, L2:(list poly))
           ((r1:poly) (In r1 L1) -> ~ (zerop r1) ->
            (Ex [r2:poly] (In r2 L2) /\ (divp r1 r2))) ->
           (q:poly) (reduce L1 p q) ->(Ex [r:poly] (reduce L2 p r)).
Hypothesis divp_id:(p:poly)(divp p p).
Require Import Relation_Definitions.
Hypothesis divp_trans:(transitive poly divp).
Hypothesis nf_div_zero:
           (p:poly) (L:(list poly)) ~ (zerop p) -> (zerop (nf L p)) ->
           (Ex [q:poly] (In q L) /\ ((divp p q) /\ ~ (zerop q))).

Theorem zerop_nf_cb: (L:(list poly)) (p:poly) (zerop (nf L p)) ->(cb L p).
Intros L p H'.
Apply zerop_elim_cb with p := (nf L p); Auto.
Qed.
Hints Resolve zerop_nf_cb.

Definition redacc: (list poly) -> (list poly) ->(list poly).
Intros H'; Elim H'.
Intros L; Exact (nil poly).
Intros a p Rec Acc.
Apply LetP with A := poly h := (nf (app p Acc) a).
Intros u H'0; Case (zerop_dec u); Intros Z.
Exact (Rec Acc).
Exact (cons u (Rec (cons u Acc))).
Defined.

Definition red: (list poly) ->(list poly) := [L:(list poly)](redacc L (nil ?)).
Hints Resolve incl_refl incl_tl incl_appr incl_appl incl_cons incl_app in_or_app
.

Theorem redacc_cb:
  (L1, L2:(list poly)) (p:poly) (In p (redacc L1 L2)) ->(cb (app L1 L2) p).
Intros L1; Elim L1; Auto.
Simpl; Auto.
Simpl; Unfold LetP; Intros a l H' L2 p.
Case (zerop_dec (nf (app l L2) a)).
Intros H'0 H'1.
Apply cb_incl with L1 := (app l L2); Auto.
Simpl.
Intros H'0 H'1; Case H'1;
 [Intros H'2; Rewrite <- H'2; Clear H'1 | Intros H'2; Clear H'1]; Auto.
Apply cb_trans with p := (nf (app l L2) a); Auto.
Apply cb_incl with L1 := (app l (cons (nf (app l L2) a) L2)); Auto.
Apply incl_app; Auto with datatypes.
Qed.

Theorem red_cb: (L:(list poly)) (p:poly) (In p (red L)) ->(cb L p).
Unfold red.
Intros L p H'.
Generalize (redacc_cb L (nil poly)); Simpl; Auto.
Rewrite <- app_nil_end; Auto.
Qed.

Theorem cb_redacc:
  (L1, L2:(list poly)) (p:poly) (In p L1) ->(cb (app (redacc L1 L2) L2) p).
Intros L1; Elim L1; Simpl; Auto.
Intros L2 p H'; Elim H'; Auto.
Unfold LetP.
Intros a l H' L2 p H'0; Case H'0;
 [Intros H'1; Rewrite H'1; Clear H'0 | Intros H'1; Clear H'0]; Auto.
Case (zerop_dec (nf (app l L2) p)); Auto.
Intros H'0.
Apply cb_comp with L1 := (app l L2); Auto.
Intros p0 H'2.
LApply (!in_app_or ? l L2 p0); Auto.
Intros H'3; Case H'3; Auto.
Intros H'0.
2:Case (zerop_dec (nf (app l L2) a)); Auto.
2:Intros H'0.
2:Apply cb_incl
       with L1 :=
            (app
               (redacc l (cons (nf (app l L2) a) L2))
               (cons (nf (app l L2) a) L2)); Auto with datatypes.
Apply cb_compo with L1 := (cons (nf (app l L2) p) (app l L2)); Simpl; Auto.
Intros q H'2; Case H'2;
 [Intros H'3; Rewrite <- H'3; Clear H'2 | Intros H'3; Clear H'2];
 Auto with datatypes.
Case (!in_app_or ? l L2 q H'3); Auto with datatypes.
Intros H'2.
Apply cb_incl
     with L1 :=
          (app
             (redacc l (cons (nf (app l L2) p) L2)) (cons (nf (app l L2) p) L2));
 Auto with datatypes.
Qed.

Theorem cb_red: (L:(list poly)) (p:poly) (In p L) ->(cb (red L) p).
Intros L p H'.
LApply (cb_redacc L (nil ?) p); [Intros H'3; Generalize H'3 | Idtac]; Simpl;
 Auto.
Rewrite <- app_nil_end; Auto.
Qed.

Theorem cb_red_cb1: (p:poly) (L:(list poly)) (cb L p) ->(cb (red L) p).
Intros p L H'.
Apply cb_compo with L1 := L; Auto.
Intros q H'0.
Apply cb_red; Auto.
Qed.

Theorem cb_red_cb2: (p:poly) (L:(list poly)) (cb (red L) p) ->(cb L p).
Intros p L H'.
Apply cb_compo with L1 := (red L); Auto.
Intros q H'0.
Apply red_cb; Auto.
Qed.
Hypothesis nf_div_zero1:
           (p:poly) (L:(list poly)) ~ (zerop p) -> (zerop (nf L p)) ->
           (ex poly [q:poly](In q L) /\ ((divp p q) /\ ~ (zerop q))).

Theorem redacc_divp:
  (L1, L2:(list poly)) (p:poly) ~ (zerop p) -> (In p (app L1 L2)) ->
  (Ex [q:poly] (In q (app (redacc L1 L2) L2)) /\ ((divp p q) /\ ~ (zerop q))).
Intros L1; Elim L1; Simpl; Auto.
Intros L2 p H' H'0; Exists p; Split; Auto.
Unfold LetP.
Intros a l H' L2 p H'0 H'1; Case H'1;
 [Intros H'2; Rewrite <- H'2; Clear H'1 | Intros H'2; Clear H'1]; Auto.
Case (zerop_dec (nf (app l L2) a)); Simpl; Auto.
Intros H'0.
LApply (nf_div_zero1 a (app l L2));
 [Intros H'5; LApply H'5; [Intros H'6; Clear H'5 | Clear H'5] | Idtac]; Auto.
Case H'6; Intros q E; Case E; Intros H'3 H'4; Case H'4; Intros H'5 H'7;
 Clear H'4 E H'6.
LApply (H' L2 q);
 [Intros H'8; LApply H'8; [Intros H'9; Clear H'8 | Clear H'8] | Idtac]; Auto.
Case H'9; Intros q0 E; Case E; Intros H'4 H'6; Case H'6; Intros H'8 H'10;
 Clear H'6 E H'9.
Exists q0; Split; [Idtac | Split]; Auto.
Apply divp_trans with y := q; Auto.
Rewrite H'2; Auto.
Intros H'0.
LApply (nf_div a (app l L2));
 [Intros H'5; LApply H'5; [Intros H'6; Clear H'5 | Clear H'5] | Idtac]; Auto.
Case H'6; Intros q E; Case E; Intros H'3 H'4; Case H'4; Intros H'5 H'7;
 Clear H'4 E H'6.
Simpl in H'3.
Case H'3; [Intros H'4; Clear H'3 | Intros H'4; Clear H'3].
Exists q; Split; [Idtac | Split]; Auto.
LApply (H' (cons (nf (app l L2) a) L2) q);
 [Intros H'8; LApply H'8; [Intros H'9; Clear H'8 | Clear H'8] | Idtac]; Auto.
Case H'9; Intros q0 E; Case E; Intros H'3 H'6; Case H'6; Intros H'8 H'10;
 Clear H'6 E H'9.
Exists q0; Split; [Idtac | Split]; Auto.
Case (!in_app_or
        ? (redacc l (cons (nf (app l L2) a) L2)) (cons (nf (app l L2) a) L2) q0
        H'3); Auto.
Simpl; Auto.
Intros H'6; Case H'6; [Intros H'9; Clear H'6 | Intros H'9; Clear H'6]; Auto.
Apply divp_trans with y := q; Auto.
Case (!in_app_or ? l L2 q H'4); Auto with datatypes.
Rewrite H'2; Auto.
Case (zerop_dec (nf (app l L2) a)); Simpl; Auto.
Intros H'0.
Case (!in_app_or ? l L2 p H'2); Auto.
Intros H'3.
LApply (H' (cons (nf (app l L2) a) L2) p);
 [Intros H'6; LApply H'6; [Intros H'7; Clear H'6 | Clear H'6] | Idtac]; Auto.
Case H'7; Intros q E; Case E; Intros H'4 H'5; Case H'5; Intros H'6 H'8;
 Clear H'5 E H'7.
Exists q; Split; Auto.
Case (!in_app_or
        ? (redacc l (cons (nf (app l L2) a) L2)) (cons (nf (app l L2) a) L2) q
        H'4); Auto.
Simpl; Auto.
Intros H'5; Case H'5;
 [Intros H'6; Try Exact H'6; Clear H'5 | Intros H'6; Clear H'5]; Auto.
Intros H'2; Exists p; Split; [Right | Idtac]; Auto.
Qed.

Theorem red_divp:
  (L:(list poly)) (p:poly) (In p L) -> ~ (zerop p) ->
  (Ex [q:poly] (In q (red L)) /\ ((divp p q) /\ ~ (zerop q))).
Intros L p H' H'0.
LApply (redacc_divp L (nil ?) p); Auto.
Simpl; Auto.
Rewrite <- app_nil_end; Auto.
Rewrite <- app_nil_end; Auto.
Qed.

Theorem red_grobner: (L:(list poly)) (grobner L) ->(grobner (red L)).
Intros L H'.
Apply def_grobner; Auto.
Intros p H'0.
LApply (grobner_def L);
 [Intros H'2; LApply (H'2 p); [Intros H'4 | Idtac] | Idtac]; Auto.
Case H'4; Auto.
Intros H'1; Case H'1; Intros q E; Clear H'1.
Right.
Apply div_reduce with L1 := L q := q; Auto.
Intros r1 H'1 H'3.
LApply (red_divp L r1);
 [Intros H'7; LApply H'7; [Intros H'8; Clear H'7 | Clear H'7] | Idtac]; Auto.
Case H'8; Intros q0 E0; Case E0; Intros H'5 H'6; Case H'6; Intros H'7 H'9;
 Clear H'6 E0 H'8.
Exists q0; Split; Auto.
Apply cb_red_cb2; Auto.
Qed.
End Reduce.

