
(****************************************************************************
                                                                           
          Buchberger : instanciation                                       
                                                                           
          Laurent Thery April 98 (revised Mai 98)                          
                                                                           
****************************************************************************)
Require Export Monomials.
Section Instan.
Variable A:Set.
Variable eqA:A -> A ->Prop.
Require Export Relation_Definitions.
Hypothesis eqA_ref:(reflexive A eqA).
Hypothesis eqA_sym:(symmetric A eqA).
Hypothesis eqA_trans:(transitive A eqA).
Hypothesis eqA_dec:(a, b:A){(eqA a b)}+{~ (eqA a b)}.
Variable A0:A.
Variable n:nat.
Require Export pair.

Local eqA_A0: (a, b:A) (eqA a A0) -> (eqA a b) ->(eqA b A0) :=
   (eqA_A0 A A0 eqA eqA_sym eqA_trans).
Variable plusA:A -> A ->A.
Hypothesis plusA_assoc:
           (a, b, c:A)(eqA (plusA a (plusA b c)) (plusA (plusA a b) c)).
Hypothesis plusA_com:(a, b:A)(eqA (plusA a b) (plusA b a)).
Hypothesis plusA_eqA_comp:
           (a, b, c, d:A) (eqA a c) -> (eqA b d) ->(eqA (plusA a b) (plusA c d)).

Local plusA_eqA_comp_l: (a, b, c:A) (eqA a b) ->(eqA (plusA a c) (plusA b c)) :=
   (plusA_eqA_comp_l A eqA eqA_ref plusA plusA_eqA_comp).

Local plusA_eqA_comp_r: (a, b, c:A) (eqA a b) ->(eqA (plusA c a) (plusA c b)) :=
   (plusA_eqA_comp_r A eqA eqA_ref plusA plusA_eqA_comp).
Hypothesis plusA_A0:(a:A)(eqA a (plusA a A0)).

Local eqA0_left: (a, b:A) (eqA a A0) ->(eqA b (plusA a b)) :=
   (eqA0_left
      A A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_com plusA_eqA_comp plusA_A0).

Local eqA0_right: (a, b:A) (eqA b A0) ->(eqA a (plusA a b)) :=
   (eqA0_right A A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_eqA_comp plusA_A0).
Variable invA:A ->A.
Hypothesis invA_plusA:(a:A)(eqA A0 (plusA a (invA a))).

Local invA0: (eqA A0 (invA A0)) :=
   (invA0 A A0 eqA eqA_sym eqA_trans plusA plusA_com plusA_A0 invA invA_plusA).
Hypothesis invA_eqA_comp:(a, b:A) (eqA a b) ->(eqA (invA a) (invA b)).

Local invA0_comp: (a:A) (eqA a A0) ->(eqA a (invA a)) :=
   (invA0_comp
      A A0 eqA eqA_sym eqA_trans plusA plusA_com plusA_A0 invA invA_plusA
      invA_eqA_comp).
Hypothesis minusA:A -> A ->A.
Hypothesis minusA_def:(a, b:A)(eqA (minusA a b) (plusA a (invA b))).
Hypothesis multA:A -> A ->A.
Hypothesis multA_eqA_comp:
           (a, b, c, d:A) (eqA a c) -> (eqA b d) ->(eqA (multA a b) (multA c d)).

Local multA_eqA_comp_l: (a, b, c:A) (eqA a b) ->(eqA (multA a c) (multA b c)) :=
   (multA_eqA_comp_l A eqA eqA_ref multA multA_eqA_comp).

Local multA_eqA_comp_r: (a, b, c:A) (eqA a b) ->(eqA (multA c a) (multA c b)) :=
   (multA_eqA_comp_r A eqA eqA_ref multA multA_eqA_comp).
Hypothesis multA_assoc:
           (a, b, c:A)(eqA (multA a (multA b c)) (multA (multA a b) c)).
Hypothesis multA_com:(a, b:A)(eqA (multA a b) (multA b a)).
Hypothesis multA_dist_l:
           (a, b, c:A)
           (eqA (plusA (multA c a) (multA c b)) (multA c (plusA a b))).

Local multA_dist_r:
  (a, b, c:A)(eqA (plusA (multA a c) (multA b c)) (multA (plusA a b) c)) :=
   (multA_dist_r
      A eqA eqA_sym eqA_trans plusA plusA_eqA_comp multA multA_com multA_dist_l).

Definition M1 := (zero_mon n).

Local Term := A * (mon n).

Local zeroP: Term ->Prop := (zeroP ? ? A0 eqA).

Local eqTerm: Term -> Term ->Prop := (eqTerm ? ? eqA).

Local eqTerm_ref: (reflexive Term eqTerm) := (eqTerm_refl ? ? eqA eqA_ref).

Local eqTerm_sym: (symmetric Term eqTerm) := (eqTerm_sym ? ? eqA eqA_sym).

Local eqTerm_trans: (transitive Term eqTerm) :=
   (eqTerm_trans ? ? eqA eqA_trans).

Local T2M: Term ->(mon n) := (T2M ? ?).

Local eqTerm_imp_eq: (a, b:Term) (eqTerm a b) ->(T2M a)=(T2M b) :=
   (eqTerm_imp_eq ? ? eqA).

Local eqMoef_dec: (x, y:(mon n)){x=y}+{~ x=y} := (eqmon_dec n).

Local eqTerm_dec: (x, y:Term){(eqTerm x y)}+{~ (eqTerm x y)} :=
   (eqTerm_dec ? ? eqA eqA_dec eqMoef_dec).

Local eq_zerop_is_eqpA:
  (a, b:Term) (T2M a)=(T2M b) -> (zeroP a) -> (zeroP b) ->(eqTerm a b) :=
   (eq_zerop_is_eqTerm ? ? A0 eqA eqA_sym eqA_trans).

Local zeroP_dec: (x:Term){(zeroP x)}+{~ (zeroP x)} :=
   (zeroP_dec ? ? A0 eqA eqA_dec).

Local zeroP_comp_eqTerm: (a, b:Term) (zeroP a) -> (eqTerm a b) ->(zeroP b) :=
   (zeroP_comp_eqTerm ? ? A0 eqA eqA_sym eqA_trans).

Local plusTerm: Term -> Term ->Term := (plusTerm ? ? plusA).

Local zeroP_plusTermr:
  (a, b:Term) (T2M a)=(T2M b) -> (zeroP b) ->(eqTerm a (plusTerm a b)) :=
   (zeroP_plusTermr
      ? ? A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_eqA_comp plusA_A0).

Local zeroP_plusTerml:
  (a, b:Term) (T2M a)=(T2M b) -> (zeroP a) ->(eqTerm b (plusTerm a b)) :=
   (zeroP_plusTerml
      ? ? A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_com plusA_eqA_comp
      plusA_A0).

Local plusTerm_comp_l:
  (a, b, c:Term) (T2M a)=(T2M c) -> (T2M b)=(T2M c) -> (eqTerm a b) ->
  (eqTerm (plusTerm a c) (plusTerm b c)) :=
   (plusTerm_comp_l A (mon n) eqA eqA_ref eqA_sym plusA plusA_eqA_comp).

Local plusTerm_comp_r:
  (a, b, c:Term) (T2M c)=(T2M a) -> (T2M c)=(T2M b) -> (eqTerm a b) ->
  (eqTerm (plusTerm c a) (plusTerm c b)) :=
   (plusTerm_comp_r A (mon n) eqA eqA_ref eqA_sym plusA plusA_eqA_comp).

Local plusTerm_com:
  (x, y:Term) (T2M x)=(T2M y) ->(eqTerm (plusTerm x y) (plusTerm y x)) :=
   (plusTerm_com A (mon n) eqA plusA plusA_com).

Local plusTerm_eq1:
  (m1, m2:Term) (T2M m1)=(T2M m2) ->(T2M (plusTerm m1 m2))=(T2M m1) :=
   (plusTerm_eq1 ? ? plusA).

Local plusTerm_eq2:
  (m1, m2:Term) (T2M m1)=(T2M m2) ->(T2M (plusTerm m1 m2))=(T2M m2) :=
   (plusTerm_eq2 ? ? plusA).

Local plusTerm_assoc:
  (a, a0, a1:Term) (T2M a1)=(T2M a0) -> (T2M a)=(T2M a0) ->
  (eqTerm (plusTerm (plusTerm a1 a0) a) (plusTerm a1 (plusTerm a0 a))) :=
   (plusTerm_assoc ? ? eqA eqA_sym plusA plusA_assoc).

Local eqTerm_plusTerm_comp:
  (a, b, c, d:Term)
  (T2M a)=(T2M c) -> (T2M b)=(T2M d) -> (eqTerm a b) -> (eqTerm c d) ->
  (eqTerm (plusTerm a c) (plusTerm b d)) :=
   (eqTerm_plusTerm_com ? ? eqA eqA_sym plusA plusA_eqA_comp).
Variable ltM:(mon n) -> (mon n) ->Prop.
Hypothesis ltM_nonrefl:(x:(mon n))~ (ltM x x).
Hypothesis ltM_trans:(transitive (mon n) ltM).
Hypothesis ltM_dec:(x, y:(mon n)){(ltM x y)}+{(ltM y x)}+{x=y}.
Hypothesis ltM_wf:(well_founded (mon n) ltM).
Hypothesis ltM_plusM_r:
           (a, b, c:(mon n)) (ltM a b) ->(ltM (mult_mon n a c) (mult_mon n b c)).
Hypothesis ltM_plusM_l:
           (a, b, c:(mon n)) (ltM a b) ->(ltM (mult_mon n c a) (mult_mon n c b)).

Local multTerm: Term -> Term ->Term := (multTerm ? ? multA (mult_mon n)).

Local multTerm_ltM_l:
  (m1, m2, m3:Term) (ltM (T2M m1) (T2M m2)) ->
  (ltM (T2M (multTerm m3 m1)) (T2M (multTerm m3 m2))) :=
   (multTerm_ltM_l ? ? multA (mult_mon n) ltM ltM_plusM_l).

Local multTerm_ltM_r:
  (m1, m2, m3:Term) (ltM (T2M m1) (T2M m2)) ->
  (ltM (T2M (multTerm m1 m3)) (T2M (multTerm m2 m3))) :=
   (multTerm_ltM_r ? ? multA (mult_mon n) ltM ltM_plusM_r).
Hypothesis multA_A0_l:(a:A)(eqA (multA A0 a) A0).

Local zeroP_multTerm_l: (a, b:Term) (zeroP a) ->(zeroP (multTerm a b)) :=
   (zeroP_multTerm_l
      ? ? A0 eqA eqA_ref eqA_sym eqA_trans multA multA_eqA_comp (mult_mon n)
      multA_A0_l).

Local zeroP_multTerm_r: (a, b:Term) (zeroP a) ->(zeroP (multTerm b a)) :=
   (zeroP_multTerm_r
      ? (mon n) A0 eqA eqA_ref eqA_sym eqA_trans multA multA_eqA_comp multA_com
      (mult_mon n) multA_A0_l).

Local multTerm_plusTerm_dist_l:
  (a, b, c:Term)
  (eqTerm (plusTerm (multTerm a c) (multTerm b c)) (multTerm (plusTerm a b) c)) :=
   (multTerm_plusTerm_dist_l
      ? ? eqA eqA_sym eqA_trans plusA plusA_eqA_comp multA multA_com
      multA_dist_l (mult_mon n)).

Local multTerm_plusTerm_dist_r:
  (a, b, c:Term)
  (eqTerm (plusTerm (multTerm c a) (multTerm c b)) (multTerm c (plusTerm a b))) :=
   (multTerm_plusTerm_dist_r ? ? eqA plusA multA multA_dist_l (mult_mon n)).

Local multTerm_eq:
  (a, b, c, d:Term) (T2M a)=(T2M b) -> (T2M c)=(T2M d) ->
  (T2M (multTerm a c))=(T2M (multTerm b d)) :=
   (multTerm_eq ? ? multA (mult_mon n)).

Local plusM_assoc:
  (a, b, c:(mon n))
  (mult_mon n a (mult_mon n b c))=(mult_mon n (mult_mon n a b) c) :=
   (mult_mon_assoc n).

Local multTerm_assoc:
  (a, b, c:Term)(eqTerm (multTerm a (multTerm b c)) (multTerm (multTerm a b) c)) :=
   (multTerm_assoc ? ? eqA multA multA_assoc (mult_mon n) plusM_assoc).

Local plusM_com: (a, b:(mon n))(mult_mon n a b)=(mult_mon n b a) :=
   (mult_mon_com n).

Local multTerm_com: (a, b:Term)(eqTerm (multTerm a b) (multTerm b a)) :=
   (multTerm_com ? ? eqA multA multA_com (mult_mon n) plusM_com).

Local eqTerm_multTerm_comp:
  (a, b, c, d:Term) (eqTerm a b) -> (eqTerm c d) ->
  (eqTerm (multTerm a c) (multTerm b d)) :=
   (eqTerm_multTerm_comp ? ? eqA eqA_sym multA multA_eqA_comp (mult_mon n)).

Local invTerm: Term ->Term := (invTerm ? ? invA).

Local plusTerm_invTerm_zeroP: (a:Term)(zeroP (plusTerm a (invTerm a))) :=
   (plusTerm_invTerm_zeroP ? ? A0 eqA eqA_sym plusA invA invA_plusA).

Local zeroP_invTerm_zeroP: (a:Term) (zeroP a) ->(zeroP (invTerm a)) :=
   (zeroP_invTerm_zeroP
      ? ? A0 eqA eqA_sym eqA_trans plusA plusA_com plusA_A0 invA invA_plusA
      invA_eqA_comp).

Local invTerm_invol: (a:Term)(eqTerm a (invTerm (invTerm a))) :=
   (invTerm_invol
      ? (mon n) A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_assoc plusA_com
      plusA_eqA_comp plusA_A0 invA invA_plusA).
Variable A1:A.
Hypothesis multA_A1_l:(a:A)(eqA (multA A1 a) a).

Local T1: Term := (T1 ? ? M1 A1).

Local plusM_M1_l: (a:(mon n))(mult_mon n M1 a)=a := (mult_mon_zero_l n).

Local plusM_M1_r: (a:(mon n))(mult_mon n a M1)=a := (mult_mon_zero_r n).
Hypothesis A1_diff_A0:~ (eqA A1 A0).

Local T1_multTerm_l: (a, b:Term) (eqTerm a T1) ->(eqTerm b (multTerm a b)) :=
   (T1_multTerm_l
      ? ? eqA eqA_ref eqA_sym eqA_trans multA multA_eqA_comp M1 (mult_mon n) A1
      multA_A1_l plusM_M1_l).

Local T1_multTerm_r: (a, b:Term) (eqTerm a T1) ->(eqTerm b (multTerm b a)) :=
   (T1_multTerm_r
      ? ? eqA eqA_ref eqA_sym eqA_trans multA multA_eqA_comp multA_com M1
      (mult_mon n) A1 multA_A1_l plusM_M1_r).

Local mult_invTerm_com:
  (a, b:Term)(eqTerm (multTerm (invTerm a) b) (invTerm (multTerm a b))) :=
   (mult_invTerm_com
      ? ? A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_assoc plusA_com
      plusA_eqA_comp plusA_A0 invA invA_plusA multA multA_eqA_comp multA_com
      multA_dist_l (mult_mon n) multA_A0_l).

Local mult_invTerm_com_r:
  (a, b:Term)(eqTerm (multTerm a (invTerm b)) (invTerm (multTerm a b))) :=
   (mult_invTerm_com_r
      ? ? A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_assoc plusA_com
      plusA_eqA_comp plusA_A0 invA invA_plusA invA_eqA_comp multA multA_eqA_comp
      multA_com multA_dist_l (mult_mon n) multA_A0_l).

Local eqTerm_invTerm_comp:
  (a, b:Term) (eqTerm a b) ->(eqTerm (invTerm a) (invTerm b)) :=
   (eqTerm_invTerm_comp ? ? eqA eqA_sym invA invA_eqA_comp).

Local invTerm_eq: (a:Term)(T2M a)=(T2M (invTerm a)) := (invTerm_eq ? ? invA).
Hypothesis M1_min:(a:(mon n))~ (ltM a M1).

Local T1_is_min_ltM: (a:Term)~ (ltM (T2M a) (T2M T1)) :=
   (T1_is_min_ltM ? ? M1 ltM A1 M1_min).

Local T1_eq: (a, b:Term) (eqTerm a T1) ->(T2M b)=(T2M (multTerm a b)) :=
   (T1_eq ? ? eqA multA M1 (mult_mon n) A1 plusM_M1_l).

Local eqTerm_T1_eq: (a:Term) (eqTerm a T1) ->(T2M a)=(T2M T1) :=
   (eqTerm_T1_eq ? ? eqA M1 A1).

Local minusTerm: Term -> Term ->Term := (minusTerm ? ? minusA).

Local eqTerm_minusTerm_plusTerm_invTerm:
  (a, b:Term)(eqTerm (minusTerm a b) (plusTerm a (invTerm b))) :=
   (eqTerm_minusTerm_plusTerm_invTerm ? ? eqA plusA invA minusA minusA_def).

Local minusTerm_ltM_l:
  (a, b, c:Term) (T2M a)=(T2M b) -> (ltM (T2M a) (T2M c)) ->
  (ltM (T2M (minusTerm a b)) (T2M c)) := (minusTerm_ltM_l ? ? minusA ltM).

Local invTerm_ltM_l:
  (a, c:Term) (ltM (T2M a) (T2M c)) ->(ltM (T2M (invTerm a)) (T2M c)) :=
   (invTerm_ltM_l ? ? invA ltM).

Local minusTerm_eq:
  (m1, m2:Term) (T2M m1)=(T2M m2) ->(T2M (minusTerm m1 m2))=(T2M m1) :=
   (minusTerm_eq ? ? minusA).
Variable divA:A ->(b:A) (nZa:~ (eqA b A0))A.
Hypothesis divA_is_multA:
           (a, b:A) (nZb:~ (eqA b A0))(eqA a (multA (divA a b nZb) b)).

Local divTerm: Term ->(b:Term) (nZb:~ (zeroP b))Term :=
   (divTerm ? ? A0 eqA divA (div_mon n)).
Hypothesis divA_eqA_comp:
           (a, b, c, d:A)
           (nZb:~ (eqA b A0)) (nZd:~ (eqA d A0)) (eqA a c) -> (eqA b d) ->
           (eqA (divA a b nZb) (divA c d nZd)).
Hypothesis divA_A0_l:(a:A) (nZa:~ (eqA a A0))(eqA (divA A0 a nZa) A0).
Hypothesis divA_multA_comp_r:
           (a, b, c:A) (nZc:~ (eqA c A0))
           (eqA (divA (multA a b) c nZc) (multA (divA a c nZc) b)).

Local eq_divTerm_plusTerm:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (T2M a)=(T2M b) ->
  (eqTerm a (multTerm (divTerm a c nZc) c)) ->
  (eqTerm b (multTerm (divTerm b c nZc) c)) ->
  (eqTerm
     (divTerm (plusTerm a b) c nZc)
     (plusTerm (divTerm a c nZc) (divTerm b c nZc))) :=
   (eq_divTerm_plusTerm
      A (mon n) A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_eqA_comp multA
      multA_eqA_comp multA_com multA_dist_l (mult_mon n) A1 multA_A1_l divA
      divA_is_multA (div_mon n) divA_eqA_comp divA_multA_comp_r).

Local divTerm_invTerm_l:
  (a, b:Term) (nZa:~ (zeroP a)) (eqTerm b (multTerm (divTerm b a nZa) a)) ->
  (eqTerm (divTerm (invTerm b) a nZa) (invTerm (divTerm b a nZa))) :=
   (divTerm_invTerm_l
      A (mon n) A0 eqA eqA_ref eqA_sym eqA_trans plusA plusA_assoc plusA_com
      plusA_eqA_comp plusA_A0 invA invA_plusA invA_eqA_comp multA multA_eqA_comp
      multA_com multA_dist_l (mult_mon n) multA_A0_l A1 multA_A1_l divA
      divA_is_multA (div_mon n) divA_eqA_comp divA_multA_comp_r).
Hypothesis divA_invA_r:
           (a, b:A) (nZb:~ (eqA b A0)) (nZib:~ (eqA (invA b) A0))
           (eqA (divA a (invA b) nZib) (invA (divA a b nZb))).

Local divTerm_invTerm_r:
  (a, b:Term)
  (nZa:~ (zeroP a))
  (nZia:~ (zeroP (invTerm a))) (eqTerm b (multTerm (divTerm b a nZa) a)) ->
  (eqTerm (divTerm b (invTerm a) nZia) (invTerm (divTerm b a nZa))) :=
   (divTerm_invTerm_r
      A (mon n) A0 eqA invA multA (mult_mon n) divA (div_mon n) divA_invA_r).

Local eq_divTerm:
  (a, b, c, d:Term)
  (nZc:~ (zeroP c)) (nZd:~ (zeroP d)) (T2M a)=(T2M b) -> (T2M c)=(T2M d) ->
  (T2M (divTerm a c nZc))=(T2M (divTerm b d nZd)) :=
   (eq_divTerm A (mon n) A0 eqA divA (div_mon n)).

Local plusM_minusM_com: (a, b:(mon n))(div_mon n (mult_mon n a b) b)=a :=
   (mult_div_com n).

Local divTerm_multTerm_l:
  (a, b, c:Term) (nZa:~ (zeroP a)) (eqTerm b (multTerm (divTerm b a nZa) a)) ->
  (eqTerm (divTerm (multTerm c b) a nZa) (multTerm c (divTerm b a nZa))) :=
   (divTerm_multTerm_l
      A (mon n) A0 eqA eqA_ref eqA_sym eqA_trans multA multA_com (mult_mon n)
      plusM_assoc divA (div_mon n) divA_eqA_comp divA_multA_comp_r
      plusM_minusM_com).

Local divTerm_multTerm_r:
  (a, b, c:Term) (nZa:~ (zeroP a)) (eqTerm b (multTerm (divTerm b a nZa) a)) ->
  (eqTerm (divTerm (multTerm b c) a nZa) (multTerm (divTerm b a nZa) c)) :=
   (divTerm_multTerm_r
      A (mon n) A0 eqA multA (mult_mon n) plusM_assoc plusM_com divA (div_mon n)
      divA_multA_comp_r plusM_minusM_com).

Local multTerm_zeroP_div:
  (a, b:Term) (zeroP (multTerm a b)) ->(zeroP a) \/ (zeroP b) :=
   (multTerm_zeroP_div
      A (mon n) A0 eqA eqA_ref eqA_sym eqA_trans eqA_dec multA multA_com
      (mult_mon n) multA_A0_l divA divA_is_multA divA_eqA_comp divA_multA_comp_r).

Local T1_nz: ~ (zeroP T1) := (T1_nz A (mon n) A0 eqA M1 A1 A1_diff_A0).
Require Export spminus.

Local zeroP_divTerm:
  (a, b:Term) (zeroP a) ->(nZb:~ (zeroP b))(zeroP (divTerm a b nZb)) :=
   (zeroP_divTerm
      A (mon n) A0 eqA eqA_ref eqA_sym eqA_trans multA multA_com multA_A0_l divA
      (div_mon n) divA_eqA_comp divA_multA_comp_r).

Local divTerm_on_eqM:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b)) (T2M a)=(T2M b) ->(eqTerm a (multTerm (divTerm a b nZb) b)) :=
   (divTerm_on_eqM
      A (mon n) A0 eqA multA M1 (mult_mon n) plusM_M1_l divA divA_is_multA
      (div_mon n) (mult_div_id n)).

Local divTerm_on_eqM_eqM:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b)) (T2M a)=(T2M b) ->(T2M (divTerm a b nZb))=(T2M T1) :=
   (divTerm_on_eqM_eqM A (mon n) A0 eqA M1 A1 divA (div_mon n) (mult_div_id n)).
Require Export def_spoly.

Local ppcm := (ppcm Term zeroP eqTerm multTerm divTerm).

Local ppcm0 := (ppcm0 Term zeroP eqTerm multTerm divTerm).

Local ppc: Term -> Term ->Term := (ppc A (mon n) A1 (ppcm_mon n)).

Local pcc_com: (a, b:Term)(eqTerm (ppc a b) (ppc b a)) :=
   (pcc_com A (mon n) eqA eqA_ref A1 (ppcm_mon n) (ppcm_com n)).

Local divTerm_ppc:
  (a, b, c:Term)
  (nZa:~ (zeroP a))
  (nZb:~ (zeroP b))
  (nZppab:~ (zeroP (ppc a b)))
  (eqTerm c (multTerm (divTerm c a nZa) a)) ->
  (eqTerm c (multTerm (divTerm c b nZb) b)) ->
  (eqTerm c (multTerm (divTerm c (ppc a b) nZppab) (ppc a b))) :=
   (divTerm_ppc
      A (mon n) A0 eqA multA (mult_mon n) A1 divA divA_is_multA (div_mon n)
      (ppcm_mon n) (ppcm_mom_is_ppcm n)).
Require Export Buch.
Require Export porder.

Definition op := (canonical ? ltM ? T2M zeroP zeroP_dec).
Require Export PolyList.

Definition pol := (list Term).

Definition spol := {l:pol | (op l)}.

Definition lspol := (list spol).
Require Export main.

Local divTerm_dec:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b))
  {(eqTerm a (multTerm (divTerm a b nZb) b))}+
  {~ (eqTerm a (multTerm (divTerm a b nZb) b))} :=
   (divTerm_dec
      A (mon n) A0 eqA multA (mult_mon n) divA divA_is_multA (div_mon n)
      (div_mon_clean n) (div_clean_dec1 n) (div_clean_dec2 n)).
Hypothesis divA_rew:
           (a, b:A) (nZ1, nZ2:~ (eqA b A0))(divA a b nZ1)=(divA a b nZ2).

Local divTerm_rew:
  (a, b:Term) (nZ1, nZ2:~ (zeroP b))(divTerm a b nZ1)=(divTerm a b nZ2) :=
   (divTerm_rew A (mon n) A0 eqA divA (div_mon n) divA_rew).

Local eqTerm_divTerm_com:
  (a, b, c, d:Term)
  (nZc:~ (zeroP c)) (nZd:~ (zeroP d)) (eqTerm a b) -> (eqTerm c d) ->
  (eqTerm (divTerm a c nZc) (divTerm b d nZd)) :=
   (eqTerm_divTerm_com A (mon n) A0 eqA eqA_sym divA (div_mon n) divA_eqA_comp).

Definition Red :=
   (Reduce1a
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec ltM_wf eqTerm_imp_eq plusTerm
      plusTerm_comp_r plusTerm_eq1 plusTerm_eq2 eqTerm_plusTerm_comp multTerm
      multTerm_ltM_l zeroP_multTerm_l multTerm_eq eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l mult_invTerm_com
      eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div T1_nz minusTerm
      eqTerm_minusTerm_plusTerm_invTerm minusTerm_ltM_l invTerm_ltM_l
      minusTerm_eq divTerm eqTerm_divTerm_com divTerm_dec).
Require Export wf_incl.

Definition get_mon: spol ->(mon n).
Intros sp; Case sp.
Intros x H'; Case x.
Exact M1.
Intros a H'1; Exact (T2M a).
Defined.
Require Export sreduce.

Local s2p := (s2p ? ? T2M zeroP zeroP_dec ltM).

Definition unit: spol ->Term.
Intros p; Case p.
Intros x; Case x.
Intros H'; Exact T1.
Intros a l; Case a.
Intros co m H'; Cut ~ (eqA co A0).
Intros H'0; Exact ((divA A1 co H'0), M1).
Inversion H'; Auto.
Simpl in H0.
Generalize H0; Case (zeroP_dec <A, (mon n)> (co, m)); Auto.
Defined.

Theorem unit_T1: (p:spol)(T2M (unit p))=(T2M T1).
Intros p; Case p.
Intros x; Case x; Simpl; Auto.
Intros a l; Case a; Auto.
Qed.

Theorem divA_nZ:
  (a:A) (b:A) ~ (eqA b A0) ->(nZa:~ (eqA a A0))~ (eqA (divA b a nZa) A0).
Intros a b H' nZa; Red; Intros H'1; Auto.
Case H'; Apply eqA_trans with y := (multA (divA b a nZa) a); Auto.
Apply eqA_trans with y := (multA A0 a); Auto.
Qed.
Hints Resolve divA_nZ.

Theorem unit_nZ: (p:spol)~ (zeroP (unit p)).
Intros p; Case p.
Intros x; Case x; Simpl; Auto.
Intros a l; Case a; Simpl; Auto.
Qed.

Local nf :=
   (nf
      ? ? T2M zeroP zeroP_dec eqTerm ltM ltM_dec multTerm multTerm_ltM_l invTerm
      T1 multTerm_zeroP_div minusTerm divTerm Red unit unit_nZ).

Local zerop := (zerop ? ? T2M zeroP zeroP_dec ltM).

Local irreducible :=
   (irreducible
      ? ? T2M zeroP zeroP_dec eqTerm ltM ltM_dec multTerm invTerm T1 minusTerm
      divTerm).

Local spminusf :=
   (spminusf
      ? ? T2M zeroP zeroP_dec ltM ltM_dec multTerm invTerm T1 minusTerm divTerm).

Local reducetop_sp :=
   (reducetop_sp
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref ltM ltM_dec multTerm invTerm T1
      minusTerm divTerm).

Local in_inPolySet := (in_inPolySet ? ? T2M zeroP zeroP_dec eqTerm ltM).

Local divTerm_def := (divTerm_def ? zeroP eqTerm multTerm divTerm).
Require Export crit.

Local canonical_imp_in_nzero :=
   (canonical_imp_in_nzero ? ? T2M zeroP zeroP_dec ltM).

Local canonical_nzeroP := (canonical_nzeroP ? ltM ? T2M zeroP zeroP_dec).

Local divTerm_on_plusM_minusM :=
   (divTerm_on_plusM_minusM
      A (mon n) A0 eqA multA (mult_mon n) divA divA_is_multA (div_mon n)).

Theorem eq_nzero_divTerm:
  (a, b:Term) (nZb:~ (zeroP b)) (T2M a)=(T2M (multTerm (divTerm a b nZb) b)) ->
  (eqTerm a (multTerm (divTerm a b nZb) b)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Local div_mon_id: (a:(mon n))<(mon n)> (div_mon n a a)=M1.
Intros a; Rewrite mult_div_id.
Simpl; Auto.
Defined.

Local div_is_T1 :=
   (div_is_T1
      A (mon n) A0 eqA eqA_ref eqA_sym eqA_trans multA M1 A1 multA_A1_l divA
      divA_is_multA (div_mon n) divA_eqA_comp divA_multA_comp_r div_mon_id).

Local nf_irreducible: (p:spol) (aP:lspol)(irreducible aP (s2p (nf p aP))).
Exact (nf_irreducible
         Term (mon n) T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
         zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
         eqTerm_imp_eq plusTerm plusTerm_comp_r eqTerm_plusTerm_comp multTerm
         multTerm_ltM_l zeroP_multTerm_l zeroP_multTerm_r
         multTerm_plusTerm_dist_r multTerm_eq multTerm_assoc multTerm_com
         eqTerm_multTerm_comp invTerm zeroP_invTerm_zeroP invTerm_invol T1
         T1_multTerm_l T1_multTerm_r mult_invTerm_com eqTerm_invTerm_comp
         invTerm_eq multTerm_zeroP_div A1_diff_A0 minusTerm
         eqTerm_minusTerm_plusTerm_invTerm divTerm eq_divTerm eqTerm_divTerm_com
         divTerm_multTerm_l div_is_T1 divTerm_on_eqM eq_nzero_divTerm Red unit
         unit_T1 unit_nZ).
Qed.

Theorem divTerm_ppcl:
  (a, b:Term) (nZa:~ (zeroP a)) ~ (zeroP b) ->
  (eqTerm (ppc a b) (multTerm (divTerm (ppc a b) a nZa) a)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros a0 m a1 m0 nZa H'; Split; Auto.
Apply ppcm_prop_l; Auto.
Qed.

Theorem divTerm_ppcr:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b))(eqTerm (ppc a b) (multTerm (divTerm (ppc a b) b nZb) b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros d0 m d1 m0 H' H'0; Split; Auto.
Apply ppcm_prop_r; Auto.
Qed.

Local Grobner :=
   (Grobner
      Term ? T2M zeroP zeroP_dec eqTerm ltM ltM_dec plusTerm multTerm invTerm T1
      minusTerm divTerm).

Theorem divTerm_on_eq:
  (a, b:Term) (nZb:~ (zeroP b)) (T2M a)=(T2M b) ->
  (eqTerm a (multTerm (divTerm a b nZb) b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros a0 m a1 m0 nZb H'; Split; Auto.
Rewrite <- H'; Auto.
Rewrite mult_div_id.
Simpl; Auto.
Rewrite (mult_mon_zero_l n m0); Auto.
Qed.

Theorem divTerm_eq:
  (a, b, c:Term) (T2M a)=(T2M b) -> ~ (zeroP a) -> ~ (zeroP b) ->
  (nZc:~ (zeroP c)) (eqTerm a (multTerm (divTerm a c nZc) c)) ->
  (eqTerm b (multTerm (divTerm b c nZc) c)).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intros a0 m a1 m0 a2 m1 H' H'0 H'1 nZc H'2; Elim H'2; Intros H'3 H'4; Clear H'2;
 Split; Auto.
Rewrite <- H'; Auto.
Qed.

Theorem multTerm_eq_com: (a, b:Term)(T2M (multTerm a b))=(T2M (multTerm b a)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros H' m H'0 m0; Apply mult_mon_com; Auto.
Qed.

Theorem divTerm_multTerm_eq_com:
  (a, b:Term) (nZa:~ (zeroP a))(T2M (divTerm (multTerm a b) a nZa))=(T2M b).
Intros a b; Case a; Case b; Simpl; Auto.
Intros H' m a0 m0 H'0.
Rewrite mult_mon_com.
Apply mult_div_com; Auto.
Qed.

Theorem get_is_correct:
  (a, b:spol) (P:lspol) ~ (zerop (nf a P)) -> (In b P) -> ~ (zerop b) ->
  ~ (mdiv n (get_mon b) (get_mon (nf a P))).
Intros a b P H' H'0 H'1.
Cut (irreducible P (s2p (nf a P))).
2:Try Exact (nf_irreducible a P); Auto.
Generalize H'; Case (nf a P); Simpl; Auto.
Intros x; Case x; Simpl.
Intros H'2 H'3; Elim H'3; Auto.
Intros a0 l H'2 H'3 H'4; Red; Intros H'5; Red in H'4.
Red in H'4.
Generalize H'0 H'1 H'5; Case b.
Intros x0; Case x0; Simpl.
Intros o H'6 H'7; Elim H'7; Auto.
Intros a1 l0 o H'6 H'7 H'8.
Cut ~ (zeroP a1); [Intros nZa1 | Idtac].
Specialize 1 H'4 with q := (spminusf a0 a1 nZa1 l l0); Intros H'9; Apply H'9;
 Clear H'9; Auto.
Apply reducetop_sp with b := a1 nZb := nZa1 q := l0; Auto.
Apply (in_inPolySet P (exist pol [l:pol](op l) (cons a1 l0) o)); Auto.
Simpl; Auto.
Red; Intros H'9; Inversion H'9.
Apply divTerm_def with nZb := nZa1; Auto.
Apply canonical_nzeroP with p := l; Auto.
Apply divTerm_on_plusM_minusM; Auto.
Apply canonical_nzeroP with p := l; Auto.
Simpl; Auto.
Apply sym_equal; Auto.
Apply mdiv_div; Auto.
Apply canonical_nzeroP with p := l0; Auto.
Qed.

Local RO := (RO ? ? T2M zeroP zeroP_dec ltM nf zerop).
Require Export splus.
Require Export seq.

Definition lpO: spol.
Exists (nil Term); Auto.
Apply (canonicalpO ? ? T2M zeroP zeroP_dec ltM); Auto.
Defined.

Local wf_RO :=
   (wf_incl ? ? T2M zeroP zeroP_dec ltM nf zerop lpO n get_mon get_is_correct).

Local RO0 := (RO0 ? ? T2M zeroP zeroP_dec ltM nf zerop).

Theorem ppc_nz: (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->~ (zeroP (ppc a b)).
Intros a b H'; Case a; Case b; Simpl; Auto.
Qed.

Local ppcp := (ppcp ? ? T2M zeroP zeroP_dec ltM ppc ppc_nz).

Local divp := (divp ? ? T2M zeroP zeroP_dec eqTerm ltM multTerm divTerm).

Theorem eqTerm_nZ_dec:
  (a, b:Term) (nZa:~ (zeroP a)) (nZb:~ (zeroP b))
  {(eqTerm a (multTerm (divTerm a b nZb) b))}+
  {~ (eqTerm a (multTerm (divTerm a b nZb) b))}.
Intros a b; Case a; Case b; Simpl; Auto.
Intros d0 m d1 m0 H' H'0.
Generalize (div_clean_dec1 n m0 m); Generalize (div_clean_dec2 n m0 m); Auto.
Case (div_mon_clean n m0 m).
Intros m1 b2; Case b2; Auto.
Intros H'1 H'2; Left; Split; Auto.
Apply sym_equal; Auto.
Elim H'2; [Intros H'4 H'5; Clear H'2 | Clear H'2]; Auto.
Intros H'1 H'2; Right; Red; Intros H'3; Elim H'3; Intros H'4 H'5; Clear H'3;
 Auto.
LApply H'1; [Intros H'3; Apply H'3; Apply sym_equal; Clear H'1 | Clear H'1];
 Auto.
Qed.

Local divp_dec :=
   (divp_dec
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref eqTerm_sym ltM
      multTerm eqTerm_multTerm_comp divTerm eqTerm_divTerm_com eqTerm_nZ_dec).

Theorem divTerm_nZ:
  (a, b:Term) ~ (zeroP a) ->(nZb:~ (zeroP b))~ (zeroP (divTerm a b nZb)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem nZero_T1: ~ (zeroP T1).
Simpl; Auto.
Qed.

Local spolyp :=
   (spolyp
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref zeroP_comp_eqTerm eqTerm_sym ltM
      ltM_trans ltM_dec eqTerm_imp_eq multTerm multTerm_ltM_l multTerm_eq
      invTerm zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm minusTerm_ltM_l invTerm_ltM_l
      minusTerm_eq divTerm ppc ppc_nz divTerm_nZ).

Local zerop_dec := (zerop_dec ? ? T2M zeroP zeroP_dec ltM).

Definition tbuch: lspol ->lspol :=
   (redbuchL
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym eqMoef_dec ltM ltM_trans ltM_dec eqTerm_imp_eq multTerm
      multTerm_ltM_l multTerm_eq eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm minusTerm_ltM_l invTerm_ltM_l
      minusTerm_eq divTerm eqTerm_divTerm_com eqTerm_nZ_dec ppc ppc_nz
      divTerm_nZ Red unit unit_nZ RO wf_RO RO0).

Local Cb :=
   (Cb Term ? T2M zeroP zeroP_dec eqTerm ltM ltM_dec plusTerm multTerm).
Require Export OBuch.

Local stable := (stable ? Cb).

Theorem tbuch_stable: (P:lspol)(stable P (tbuch P)).
Exact (redbuchL_stable
         Term (mon n) T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
         zeroP_comp_eqTerm eqTerm_sym eqMoef_dec ltM ltM_nonrefl ltM_trans
         ltM_dec eqTerm_imp_eq plusTerm zeroP_plusTermr zeroP_plusTerml
         plusTerm_comp_l plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2
         plusTerm_assoc eqTerm_plusTerm_comp multTerm multTerm_ltM_l
         zeroP_multTerm_l zeroP_multTerm_r multTerm_plusTerm_dist_r multTerm_eq
         multTerm_assoc eqTerm_multTerm_comp invTerm plusTerm_invTerm_zeroP
         zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l mult_invTerm_com
         eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div nZero_T1 minusTerm
         eqTerm_minusTerm_plusTerm_invTerm minusTerm_ltM_l invTerm_ltM_l
         minusTerm_eq divTerm eqTerm_divTerm_com eqTerm_nZ_dec ppc
         divTerm_on_eqM ppc_nz divTerm_ppc divTerm_ppcl divTerm_ppcr divTerm_nZ
         Red unit unit_T1 unit_nZ RO wf_RO RO0).
Qed.

Local tselectpoly :=
   (tselectpoly
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref eqTerm_sym ltM
      ltM_dec multTerm eqTerm_multTerm_comp invTerm T1 minusTerm divTerm
      eqTerm_divTerm_com eqTerm_nZ_dec).

Theorem buch_Grobner: (P:lspol)(Grobner (tbuch P)).
Exact (redbuchL_Grobner
         Term (mon n) T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
         eqTerm_dec zeroP_comp_eqTerm eqTerm_sym eqMoef_dec ltM ltM_nonrefl
         ltM_trans ltM_dec ltM_wf eqTerm_imp_eq plusTerm zeroP_plusTermr
         zeroP_plusTerml plusTerm_comp_l plusTerm_comp_r plusTerm_com
         plusTerm_eq1 plusTerm_eq2 plusTerm_assoc eqTerm_plusTerm_comp multTerm
         multTerm_ltM_l multTerm_ltM_r zeroP_multTerm_l zeroP_multTerm_r
         multTerm_plusTerm_dist_l multTerm_plusTerm_dist_r multTerm_eq
         multTerm_assoc multTerm_com eqTerm_multTerm_comp invTerm
         plusTerm_invTerm_zeroP zeroP_invTerm_zeroP invTerm_invol T1
         T1_multTerm_l T1_multTerm_r mult_invTerm_com eqTerm_invTerm_comp
         invTerm_eq [a:Term](M1_min (T2M a)) multTerm_zeroP_div nZero_T1
         minusTerm eqTerm_minusTerm_plusTerm_invTerm minusTerm_ltM_l
         invTerm_ltM_l minusTerm_eq divTerm eq_divTerm_plusTerm
         divTerm_invTerm_l eq_divTerm eqTerm_divTerm_com divTerm_multTerm_l
         divTerm_multTerm_r div_is_T1 eqTerm_nZ_dec divTerm_eq ppc divTerm_rew
         divTerm_on_eqM divTerm_on_eqM_eqM ppc_nz divTerm_ppc divTerm_ppcl
         divTerm_ppcr divTerm_nZ eq_nzero_divTerm multTerm_eq_com
         divTerm_multTerm_eq_com pcc_com tselectpoly Red unit unit_T1 unit_nZ RO
         wf_RO RO0).
Qed.

Definition lplus :=
   [n:nat]
   (splus
      ? ? T2M zeroP zeroP_dec ltM ltM_trans ltM_dec plusTerm plusTerm_eq1
      plusTerm_eq2).
Require Export smult.

Definition lmult :=
   [n:nat]
   (smult
      ? ? T2M zeroP zeroP_dec ltM ltM_trans ltM_dec plusTerm plusTerm_eq1
      plusTerm_eq2 multTerm multTerm_ltM_l).
Require Export smultm_lm.
Require Export seq.

Definition lscal: A -> spol ->spol.
Intros H' H'0; Case H'0.
Intros x H'1; Exists (tmultm_lm ? zeroP zeroP_dec multTerm (H', M1) x); Auto.
Unfold tmultm_lm; Case (zeroP_dec <A, (mon n)> (H', M1)); Simpl; Auto.
Intros H'2; Apply (canonicalpO ? ? T2M zeroP zeroP_dec ltM).
Intros H'2;
 Apply (canonical_multlm_lm ? ? T2M zeroP zeroP_dec ltM multTerm multTerm_ltM_l);
 Auto.
Try Exact multTerm_zeroP_div.
Qed.

Definition lp1: spol.
Exists (cons (A1, M1) (nil Term)); Auto.
Apply (canonical1 ? ? T2M zeroP zeroP_dec ltM); Auto.
Defined.

Definition lgen: nat ->spol.
Intros m; Exists (cons (A1, (gen_mon n m)) (nil Term)).
Apply (canonical1 ? ? T2M zeroP zeroP_dec ltM); Auto.
Defined.
End Instan.
