(*****************************************************************************)
(*                                                                           *)
(*          Buchberger : ordering: lexi and total                            *)
(*                                                                           *)
(*          Laurent Thery 	                                             *)
(*                                                                           *)
(*****************************************************************************)

Require Import Eqdep.
Section lexi_order.
Require Import Monomials.

Inductive orderc: (n:nat) (mon n) -> (mon n) ->Prop :=
   lo1:
     (n:nat) (a, b:nat) (p:(mon n)) (lt b a) ->
     (orderc (S n) (c_n n a p) (c_n n b p))
  | lo2:
      (n:nat) (a, b:nat) (p, q:(mon n)) (orderc n p q) ->
      (orderc (S n) (c_n n a p) (c_n n b q)).
Hints Resolve lo1 lo2.
Require Import Arith.
Require Import Compare_dec.

Theorem orderc_dec:
  (n:nat) (a, b:(mon n)){(orderc n a b)}+{(orderc n b a)}+{a =b}.
Intros n a; Elim a; Auto.
Intro b.
Rewrite <- (mon_0 b); Auto.
Intros d n0 m H' b; Try Assumption.
Rewrite <- (proj_ok d b).
Case (H' (pmon2 (S d) b)).
Intro H'0; Case H'0.
Intro H'1.
Left; Left; Auto.
Intro H'1; Left; Right; Auto.
Intro H'0.
Elim (lt_eq_lt_dec n0 (pmon1 (S d) b)); [Intro H'1; Elim H'1 | Idtac];
 Intro H'2; Auto.
Left; Right; Auto.
Rewrite H'0; Auto.
Right; Rewrite H'0; Rewrite H'2; Auto.
Left; Left; Rewrite H'0; Auto.
Qed.

Definition degc: (n:nat) (mon n) ->nat.
Intros n H'; Elim H'.
Exact O.
Intros d n1 M n2; Exact (plus n1 n2).
Defined.

Inductive total_orderc: (n:nat) (mon n) -> (mon n) ->Prop :=
   total_orderc0:
     (n:nat) (p, q:(mon n)) (lt (degc n p) (degc n q)) ->(total_orderc n p q)
  | total_orderc1:
      (n:nat) (p, q:(mon n)) (degc n p) =(degc n q) -> (orderc n p q) ->
      (total_orderc n p q).
Hints Resolve total_orderc0 total_orderc1.
Require Import LetP.

Theorem total_orderc_dec:
  (n:nat) (a, b:(mon n)){(total_orderc n a b)}+{(total_orderc n b a)}+{a =b}.
Intros n a b.
Apply LetP with A := nat h := (degc n a).
Intros u H'; Apply LetP with A := nat h := (degc n b).
Intros u0 H'0.
Case (le_lt_dec u u0); Auto.
Intro H'1; Case (le_lt_eq_dec u u0); Auto.
Rewrite H'0; Rewrite H'; Auto.
Rewrite H'0; Rewrite H'; Intro H'2; Case (orderc_dec n a b); Auto.
Intro H'3; Case H'3; Auto.
Rewrite H'0; Rewrite H'; Auto.
Qed.
End lexi_order.
