
(****************************************************************************
                                                                           
          Buchberger : module for porder                                   
                                                                           
          Laurent Thery 	                                             
                                                                           
****************************************************************************)
Require Import porder.
Require Import Lexicographic_Exponentiation.
Require Import Relation_Definitions.
Require Import Inverse_Image.
Require Import Inclusion.
Require Import PolyList.
Require Import Relation_Operators.

Local eqM := (eqB ? ? T2M).

Local eqM_refl: (reflexive Term eqM) := (eqB_refl ? ? T2M).
Hints Resolve eqM_refl.

Local eqM_sym: (symmetric Term eqM) := (eqB_sym ? ? T2M).

Local eqM_trans: (transitive Term eqM) := (eqB_trans ? ? T2M).

Local eqM_dec: (x, y:Term){(eqM x y)}+{~ (eqM x y)} :=
   (eqB_dec ? eqMonome_dec ? T2M).

Local pO := (nil Term).

Local pX := (!cons Term).

Local order := (ltB ? ltM ? T2M).

Local order_trans: (transitive Term order) := (ltB_trans ? ltM ltM_trans ? T2M).

Local order_dec: (m1, m2:Term){(order m1 m2)}+{(order m2 m1)}+{(eqM m1 m2)} :=
   (ltB_dec ? ltM ltM_dec ? T2M).

Local order_not_eqM: (m, n:Term) (eqM m n) ->~ (order m n) :=
   (ltB_not_eqB ? ltM ltM_nonrefl ? T2M).

Local eqM_not_order: (m, n:Term) (order m n) ->~ (eqM m n) :=
   (eqB_not_ltB ? ltM ltM_nonrefl ? T2M).

Local order_not_refl: (x:Term)~ (order x x) :=
   (ltB_not_refl ? ltM ltM_nonrefl ? T2M).
Hints Resolve order_not_refl.

Local order_not_order: (m, n:Term) (order m n) ->~ (order n m) :=
   (ltB_not_ltB ? ltM ltM_trans ltM_nonrefl ? T2M).
Hints Resolve order_not_eqM eqM_not_order order_not_order.

Local canonical := (canonical ? ltM ? T2M zeroP zeroP_dec).
Hints Unfold  eqM.

Local canonical0:
  (a, b:Term) (order b a) -> ~ (zeroP a) -> ~ (zeroP b) ->
  (canonical (pX a (pX b pO))) := (canonical0 ? ltM ? T2M zeroP zeroP_dec).

Local canonical_imp_order:
  (l:(list Term)) (a, b:Term) (canonical (pX a (pX b l))) ->(order b a) :=
   (canonical_pX_order ? ltM ? T2M zeroP zeroP_dec).

Local canonical_skip_fst:
  (l:(list Term)) (a, b:Term) (canonical (pX a (pX b l))) ->(canonical (pX a l)) :=
   (ocanonical_skip_fst ? ltM ltM_trans ? T2M zeroP zeroP_dec).

Local canonical_imp_canonical:
  (l:(list Term)) (a:Term) (canonical (pX a l)) ->(canonical l) :=
   (canonical_imp_canonical ? ltM ? T2M zeroP zeroP_dec).

Local canonical_cons:
  (l:(list Term))
  (a, b:Term) (order b a) -> ~ (zeroP a) -> (canonical (pX b l)) ->
  (canonical (pX a (pX b l))) := (canonical_cons ? ltM ? T2M zeroP zeroP_dec).
Hints Resolve canonical_cons.

Local canonical_nzeroP:
  (a:Term) (p:(list Term)) (canonical (pX a p)) ->~ (zeroP a) :=
   (canonical_nzeroP ? ltM ? T2M zeroP zeroP_dec).

Local order_eqM_r: (a, b, c:Term) (eqM a b) -> (order a c) ->(order b c) :=
   (ltB_eqBr ? ltM ? T2M).

Local order_eqM_l: (a, b, c:Term) (eqM a b) -> (order c a) ->(order c b) :=
   (ltB_eqBl ? ltM ? T2M).

Local canonical_pX_eqM:
  (a, b:Term)
  (p:(list Term)) (canonical (pX a p)) -> (eqM a b) -> ~ (zeroP b) ->
  (canonical (pX b p)) := (canonical_pX_eqC ? ltM ? T2M zeroP zeroP_dec).
Hints Resolve canonical_imp_canonical.
Hints Resolve canonical_pX_eqM.

Local ltP: (list Term) -> (list Term) ->Prop := (ltP ? ltM ? T2M).

Local ltPO: (x:Term) (p:(list Term))(ltP pO (pX x p)) := (ltPO ? ltM ? T2M).

Local ltP_hd:
  (x, y:Term) (p, q:(list Term)) (order x y) ->(ltP (pX x p) (pX y q)) :=
   (ltP_hd ? ltM ? T2M).

Local ltP_tl:
  (x, y:Term) (p, q:(list Term)) (eqM x y) -> (ltP p q) ->
  (ltP (pX x p) (pX y q)) := (ltP_tl ? ltM ? T2M).
Hints Resolve ltPO ltP_hd ltP_tl.

Local ltP_not_refl: (x:(list Term))~ (ltP x x) :=
   (ltp_not_refl ? ltM ltM_nonrefl ? T2M).

Local ltP_trans: (x, y, z:(list Term)) (ltP x y) -> (ltP y z) ->(ltP x z) :=
   (ltP_trans ? ltM ltM_trans ? T2M).

Local canonical_pX_ltP:
  (a:Term) (p:(list Term)) (canonical (pX a p)) ->(ltP p (pX a pO)) :=
   (ocanonical_pX_ltP ? ltM ? T2M zeroP zeroP_dec).
Hints Resolve canonical_pX_ltP.

Local ltP_pX_canonical:
  (a:Term) (p:(list Term)) (canonical p) -> ~ (zeroP a) -> (ltP p (pX a pO)) ->
  (canonical (pX a p)) := (ltP_pX_canonical ? ltM ? T2M zeroP zeroP_dec).

Local poly := {l:(list Term) | (canonical l)}.

Local sltP: poly -> poly ->Prop := (sltP ? ltM ? T2M zeroP zeroP_dec).

Local sltp_wf: (well_founded poly sltP) :=
   (sltp_wf ? ltM ltM_wf ? T2M zeroP zeroP_dec).

