
(****************************************************************************
                                                                           
          Buchberger : module for splus                                    
                                                                           
          Laurent Thery 	                                             
                                                                           
****************************************************************************)
Require Import splus.
Load "mod_seq".

Local plusTerm_eqM1: (m1, m2:Term) (eqM m1 m2) ->(eqM (plusTerm m1 m2) m1) :=
   (plusTerm_eqM1 ? ? T2M plusTerm plusTerm_eq1).
Hints Resolve plusTerm_eqM1.

Local plusTerm_eqM2: (m1, m2:Term) (eqM m1 m2) ->(eqM (plusTerm m1 m2) m2) :=
   (plusTerm_eqM2 ? ? T2M plusTerm plusTerm_eq2).
Hints Resolve plusTerm_eqM2.

Local pluspf: (list Term) -> (list Term) ->(list Term) :=
   (pluspf ? ? T2M zeroP zeroP_dec ltM ltM_dec plusTerm).

Local canonical_pluspf:
  (l1, l2:(list Term)) (canonical l1) -> (canonical l2) ->
  (canonical (pluspf l1 l2)) :=
   (canonical_pluslmf
      ? ? T2M zeroP zeroP_dec ltM ltM_trans ltM_dec plusTerm plusTerm_eqM1
      plusTerm_eqM2).
Hints Resolve canonical_pluspf.

Local p0_pluspf_l: (p:(list Term))(eqP (pluspf pO p) p) :=
   (p0_pluspf_l
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref ltM ltM_nonrefl ltM_trans
      ltM_dec plusTerm).

Local p0_pluspf_r: (p:(list Term))(eqP (pluspf p pO) p) :=
   (p0_pluspf_r
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref ltM ltM_nonrefl ltM_trans
      ltM_dec plusTerm).
Hints Resolve p0_pluspf_l p0_pluspf_r.

Local pluspf_inv1:
  (a, b:Term) (p, q:(list Term)) (order b a) ->
  (eqP (pX a (pluspf p (pX b q))) (pluspf (pX a p) (pX b q))) :=
   (pluspf_inv1
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref ltM ltM_nonrefl ltM_trans
      ltM_dec plusTerm).

Local pluspf_inv2:
  (a, b:Term) (p, q:(list Term)) (order a b) ->
  (eqP (pX b (pluspf (pX a p) q)) (pluspf (pX a p) (pX b q))) :=
   (pluspf_inv2
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref ltM ltM_nonrefl ltM_trans
      ltM_dec plusTerm).

Local pluspf_inv3a:
  (a, b:Term) (p, q:(list Term)) (eqM a b) -> (zeroP (plusTerm a b)) ->
  (eqP (pluspf p q) (pluspf (pX a p) (pX b q))) :=
   (pluspf_inv3a
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref ltM ltM_nonrefl ltM_trans
      ltM_dec plusTerm).

Local pluspf_inv3b:
  (a, b:Term) (p, q:(list Term)) (eqM a b) -> ~ (zeroP (plusTerm a b)) ->
  (eqP (pX (plusTerm a b) (pluspf p q)) (pluspf (pX a p) (pX b q))) :=
   (pluspf_inv3b
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref ltM ltM_nonrefl ltM_trans
      ltM_dec plusTerm).
Hints Resolve pluspf_inv1 pluspf_inv2 pluspf_inv3a pluspf_inv3b.

Local plusTerm_orderr:
  (a, b, c:Term) (eqM a b) -> (order c a) ->(order c (plusTerm a b)) :=
   (plusTerm_orderr ? ? T2M ltM plusTerm plusTerm_eqM1).

Local plusTerm_orderl:
  (a, b, c:Term) (eqM a b) -> (order a c) ->(order (plusTerm a b) c) :=
   (plusTerm_orderl ? ? T2M ltM plusTerm plusTerm_eqM1).
Hints Resolve plusP_zero_pOl plusP_zero_pOr.
Hints Resolve eqp_trans.

Local pluspf_assoc:
  (p, q, r:(list Term)) (canonical p) -> (canonical q) -> (canonical r) ->
  (eqP (pluspf (pluspf p q) r) (pluspf p (pluspf q r))) :=
   (pluspf_assoc
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec plusTerm zeroP_plusTermr
      zeroP_plusTerml plusTerm_comp_l plusTerm_comp_r plusTerm_com plusTerm_eqM1
      plusTerm_eqM2 plusTerm_assoc).
Hints Resolve pluspf_assoc.

Local pluspf_com: (p, q:(list Term))(eqP (pluspf p q) (pluspf q p)) :=
   (pluspf_com
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_ref zeroP_comp_eqTerm eqTerm_sym ltM
      ltM_nonrefl ltM_trans ltM_dec plusTerm plusTerm_com).

Local eqp_pluspf_com:
  (p, q, r, t:(list Term))
  (canonical p) ->
  (canonical q) -> (canonical r) -> (canonical t) -> (eqP p r) -> (eqP q t) ->
  (eqP (pluspf p q) (pluspf r t)) :=
   (eqp_pluspf_com
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec eqTerm_imp_eq plusTerm
      eqTerm_plusTerm_comp).
Hints Resolve pluspf_com eqp_pluspf_com.

Local pluspf_inv1_eqa:
  (a:Term) (p, q:(list Term)) (canonical (pX a p)) -> (canonical (pX a q)) ->
  <(list Term)> (pX a (pluspf p q))=(pluspf (pX a p) q) :=
   (pluspf_inv1_eqa
      ? ? T2M zeroP zeroP_dec ltM ltM_nonrefl ltM_trans ltM_dec plusTerm).



Local pluspf_inv2_eq:
  (a, b:Term) (p, q:(list Term)) (order a b) ->
  (pX b (pluspf (pX a p) q))=(pluspf (pX a p) (pX b q)) :=
   (pluspf_inv2_eq
      ? ? T2M zeroP zeroP_dec ltM ltM_nonrefl ltM_trans ltM_dec
      plusTerm).

Local pluspf_inv2_eqa:
  (a:Term)
  (p, q:(list Term)) (canonical (pX a p)) -> (canonical (pX a q)) ->
  <(list Term)> (pX a (pluspf p q))=(pluspf p (pX a q)) :=
   (pluspf_inv2_eqa
      ? ? T2M zeroP zeroP_dec ltM ltM_nonrefl ltM_trans ltM_dec
      plusTerm).

Local pluspf_inv1_eq:
  (a, b:Term) (p, q:(list Term)) (order b a) ->
  (pX a (pluspf p (pX b q)))=(pluspf (pX a p) (pX b q)) :=
   (pluspf_inv1_eq
      ? ? T2M zeroP zeroP_dec ltM ltM_nonrefl ltM_trans ltM_dec
      plusTerm).

Local pluspf_inv3a_eq:
  (a, b:Term)
  (p, q:(list Term)) (eqM a b) -> (zeroP (plusTerm a b)) ->
  (pluspf p q)=(pluspf (pX a p) (pX b q)) :=
   (pluspf_inv3a_eq
      ? ? T2M zeroP zeroP_dec ltM ltM_nonrefl ltM_trans ltM_dec
      plusTerm).

Local pluspf_inv3b_eq:
  (a, b:Term)
  (p, q:(list Term)) (eqM a b) -> ~ (zeroP (plusTerm a b)) ->
  (pX (plusTerm a b) (pluspf p q))=(pluspf (pX a p) (pX b q)) :=
   (pluspf_inv3b_eq
      ? ? T2M zeroP zeroP_dec ltM ltM_nonrefl ltM_trans ltM_dec
      plusTerm).
