(****************************************************************************
                                                                           
          Buchberger : module for spminus                                  
                                                                           
          Laurent Thery 	                                             
                                                                           
****************************************************************************)
Load "mod_sminus".
Require Import spminus.

Local divP: Term -> Term ->Prop := (divP ? zeroP eqTerm multTerm divTerm).

Local divTerm_def:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b)) (eqTerm a (multTerm (divTerm a b nZb) b)) ->(divP a b) :=
   (divTerm_def ? zeroP eqTerm multTerm divTerm).

Local divP_inv1: (a, b:Term) (divP a b) ->~ (zeroP a) :=
   (divP_inv1 ? zeroP eqTerm multTerm divTerm).

Local divP_inv2: (a, b:Term) (divP a b) ->~ (zeroP b) :=
   (divP_inv2 ? zeroP eqTerm multTerm divTerm).

Local divP_inv3:
  (a, b:Term) (nZb:~ (zeroP b)) (divP a b) ->
  (eqTerm a (multTerm (divTerm a b nZb) b)) :=
   (divP_inv3
      ? zeroP eqTerm eqTerm_trans eqTerm_ref eqTerm_sym multTerm
      eqTerm_multTerm_comp divTerm eqTerm_divTerm_com).
Hints Resolve divP_inv1 divP_inv2 divP_inv3.

Local divP_plusTerm:
  (a, b, c:Term)
  (divP a c) -> (divP b c) -> (eqM a b) -> ~ (zeroP (plusTerm a b)) ->
  (divP (plusTerm a b) c) :=
   (divP_plusTerm
      ? ? T2M zeroP eqTerm eqTerm_trans eqTerm_ref eqTerm_sym eqTerm_imp_eq
      plusTerm eqTerm_plusTerm_comp multTerm multTerm_plusTerm_dist_l
      multTerm_eq eqTerm_multTerm_comp divTerm eq_divTerm_plusTerm eq_divTerm
      eqTerm_divTerm_com).
Hints Resolve divP_plusTerm.

Local divP_invTerm_l: (a, b:Term) (divP a b) ->(divP (invTerm a) b) :=
   (divP_invTerm_l
      ? zeroP eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm eqTerm_sym
      multTerm eqTerm_multTerm_comp invTerm zeroP_invTerm_zeroP invTerm_invol
      mult_invTerm_com eqTerm_invTerm_comp divTerm divTerm_invTerm_l).
Hints Resolve divP_invTerm_l.

Local divP_invTerm_r: (a, b:Term) (divP a b) ->(divP a (invTerm b)) :=
   (divP_invTerm_r
      ? zeroP eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm eqTerm_sym
      multTerm eqTerm_multTerm_comp invTerm zeroP_invTerm_zeroP invTerm_invol
      mult_invTerm_com mult_invTerm_com_r eqTerm_invTerm_comp divTerm
      divTerm_invTerm_r).
Hints Resolve divP_invTerm_r.

Local divP_is_not_order: (a, b:Term) (divP a b) ->~ (order a b) :=
   (divP_is_not_order
      ? ? T2M zeroP eqTerm eqTerm_ref eqTerm_sym ltM ltM_nonrefl ltM_trans
      ltM_dec eqTerm_imp_eq multTerm multTerm_ltM_r multTerm_eq T1 T1_multTerm_l
      T1_is_min divTerm).
Hints Resolve divP_is_not_order.

Local divTerm_multTerml:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->(divP (multTerm a b) a) :=
   (divTerm_multTerml
      ? zeroP eqTerm eqTerm_trans eqTerm_ref eqTerm_sym multTerm multTerm_com
      eqTerm_multTerm_comp T1 T1_multTerm_l multTerm_zeroP_div divTerm
      divTerm_multTerm_r div_is_T1).
Hints Resolve divTerm_multTerml.

Local divTerm_multTermr:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->(divP (multTerm a b) b) :=
   (divTerm_multTermr
      ? zeroP eqTerm eqTerm_trans eqTerm_ref eqTerm_sym multTerm
      eqTerm_multTerm_comp T1 T1_multTerm_l T1_multTerm_r multTerm_zeroP_div
      divTerm divTerm_multTerm_l div_is_T1).
Hints Resolve divTerm_multTermr.

Local divP_trans: (transitive Term divP) :=
   (divP_trans
      ? zeroP eqTerm eqTerm_trans eqTerm_ref eqTerm_sym multTerm multTerm_assoc
      eqTerm_multTerm_comp divTerm eqTerm_divTerm_com divTerm_multTerm_l).

Local spminusf:
  Term ->(b:Term) ~ (zeroP b) -> (list Term) -> (list Term) ->(list Term) :=
   (spminusf
      ? ? T2M zeroP zeroP_dec ltM ltM_dec multTerm invTerm T1 minusTerm divTerm).

Local divP_nZero:
  (a, b:Term) (nZb:~ (zeroP b)) (divP a b) ->~ (zeroP (divTerm a b nZb)) :=
   (divP_nZero
      ? zeroP eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm eqTerm_sym
      multTerm zeroP_multTerm_l eqTerm_multTerm_comp divTerm eqTerm_divTerm_com).
Hints Resolve divP_nZero.

Local canonical_spminusf:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term)) (canonical p) -> (canonical q) -> (divP a b) ->
  (canonical (spminusf a b nZb p q)) :=
   (canonical_spminusf
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym ltM ltM_trans ltM_dec eqTerm_imp_eq multTerm multTerm_ltM_l
      zeroP_multTerm_l multTerm_eq eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm minusTerm_ltM_l invTerm_ltM_l
      minusTerm_eq divTerm eqTerm_divTerm_com).
Hints Resolve canonical_spminusf.

Local spminusf_extend:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term))
  (divP a b) -> (canonical (pX a p)) -> (canonical (pX b q)) ->
  (eqP (spminusf a b nZb p q) (spminusf a b nZb (pX a p) (pX b q))) :=
   (spminusf_extend
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec eqTerm_imp_eq plusTerm
      plusTerm_comp_r eqTerm_plusTerm_comp multTerm multTerm_ltM_l
      zeroP_multTerm_l multTerm_eq eqTerm_multTerm_comp invTerm
      plusTerm_invTerm_zeroP zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l
      mult_invTerm_com eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div
      nZero_T1 minusTerm eqTerm_minusTerm_plusTerm_invTerm divTerm
      eqTerm_divTerm_com).
Hints Resolve spminusf_extend.

Local canonical_spminusf_full:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term))
  (canonical (pX a p)) -> (canonical (pX b q)) -> (divP a b) ->
  (canonical (spminusf a b nZb p q)) :=
   (canonical_spminusf_full
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym ltM ltM_trans ltM_dec eqTerm_imp_eq multTerm multTerm_ltM_l
      zeroP_multTerm_l multTerm_eq eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm minusTerm_ltM_l invTerm_ltM_l
      minusTerm_eq divTerm eqTerm_divTerm_com).
Hints Resolve canonical_spminusf_full.

Local canonical_spminusf_full_t:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term))
  (canonical (pX a p)) -> (canonical (pX b q)) -> (divP a b) ->
  (canonical (pX a (spminusf a b nZb p q))) :=
   (canonical_spminus_full_t
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec eqTerm_imp_eqM plusTerm
      plusTerm_comp_r plusTerm_eq1 plusTerm_eq2 eqTerm_plusTerm_comp multTerm
      multTerm_ltM_l zeroP_multTerm_l multTerm_eqM eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l mult_invTerm_com
      eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div nZero_T1 minusTerm
      eqTerm_minusTerm_plusTerm_invTerm divTerm eqTerm_divTerm_com).
Hints Resolve canonical_spminusf_full_t.

Local spminusf_pluspf:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (pluspf p q) r) (pluspf (spminusf a b nZb p r) q)) :=
   (spminusf_pluspf
      ? ? T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec eqTerm_imp_eqM plusTerm
      zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l plusTerm_comp_r
      plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc eqTerm_plusTerm_comp
      multTerm multTerm_ltM_l zeroP_multTerm_l multTerm_eqM eqTerm_multTerm_comp
      invTerm zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l
      mult_invTerm_com eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div
      nZero_T1 minusTerm eqTerm_minusTerm_plusTerm_invTerm minusTerm_ltM_l
      invTerm_ltM_l minusTerm_eq divTerm eqTerm_divTerm_com).
Hints Resolve spminusf_pluspf.

Local eqp_spminusf_com:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) -> (canonical r) -> (eqP p q) -> (eqTerm a b) -> (divP a c) ->
  (eqP (spminusf a c nZc p r) (spminusf b c nZc q r)) :=
   (eqp_spminusf_com
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm plusTerm_comp_r eqTerm_plusTerm_comp multTerm
      multTerm_ltM_l zeroP_multTerm_l multTerm_eq eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l mult_invTerm_com
      eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div nZero_T1 minusTerm
      eqTerm_minusTerm_plusTerm_invTerm divTerm eqTerm_divTerm_com).
Hints Resolve eqp_spminusf_com.

Local eqTerm_spminusf_com:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q:(list Term))
  (canonical p) -> (canonical q) -> (eqTerm a b) -> (divP a c) ->
  (eqP (spminusf a c nZc p q) (spminusf b c nZc p q)) :=
   (eqTerm_spminusf_com
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm plusTerm_comp_r eqTerm_plusTerm_comp multTerm
      multTerm_ltM_l zeroP_multTerm_l multTerm_eq eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l mult_invTerm_com
      eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div nZero_T1 minusTerm
      eqTerm_minusTerm_plusTerm_invTerm divTerm eqTerm_divTerm_com).
Hints Resolve eqTerm_spminusf_com.

Local spminusf_minuspf:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (minuspf p q) r) (minuspf (spminusf a b nZb p r) q)) :=
   (spminusf_minuspf
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l
      plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc
      eqTerm_plusTerm_comp multTerm multTerm_ltM_l zeroP_multTerm_l multTerm_eq
      eqTerm_multTerm_comp invTerm zeroP_invTerm_zeroP invTerm_invol T1
      T1_multTerm_l mult_invTerm_com eqTerm_invTerm_comp invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm eqTerm_minusTerm_plusTerm_invTerm
      minusTerm_ltM_l invTerm_ltM_l minusTerm_eq divTerm eqTerm_divTerm_com).
Hints Resolve spminusf_minuspf.

Local spminusf_plusTerm:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  (divP a c) -> (divP b c) -> (eqM a b) -> ~ (zeroP (plusTerm a b)) ->
  (eqP
     (spminusf (plusTerm a b) c nZc (pluspf p q) r)
     (pluspf (spminusf a c nZc p r) (spminusf b c nZc q r))) :=
   (spminusf_plusTerm
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l
      plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc
      eqTerm_plusTerm_comp multTerm multTerm_ltM_l zeroP_multTerm_l
      zeroP_multTerm_r multTerm_plusTerm_dist_l multTerm_plusTerm_dist_r
      multTerm_eq eqTerm_multTerm_comp invTerm zeroP_invTerm_zeroP invTerm_invol
      T1 T1_multTerm_l mult_invTerm_com eqTerm_invTerm_comp invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm eqTerm_minusTerm_plusTerm_invTerm
      minusTerm_ltM_l invTerm_ltM_l minusTerm_eq divTerm eq_divTerm_plusTerm
      eq_divTerm eqTerm_divTerm_com).
Hints Resolve spminusf_plusTerm.

Local eqM_divTerm:
  (a, b, c, d:Term)
  (nZc:~ (zeroP c)) (nZd:~ (zeroP d)) (eqM a b) -> (eqM c d) ->
  (eqM (divTerm a c nZc) (divTerm b d nZd)) := eq_divTerm.
Hints Resolve eqM_divTerm.

Local spminusf_multTerm:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q:(list Term))
  (canonical p) -> (canonical q) -> (divP b c) -> ~ (zeroP a) ->
  (eqP
     (spminusf (multTerm a b) c nZc (multm_lm a p) q)
     (multm_lm a (spminusf b c nZc p q))) :=
   (spminusf_multTerm
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm plusTerm_comp_r eqTerm_plusTerm_comp multTerm
      multTerm_ltM_l zeroP_multTerm_l zeroP_multTerm_r multTerm_plusTerm_dist_r
      multTerm_eq multTerm_assoc multTerm_com eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l mult_invTerm_com
      eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div nZero_T1 minusTerm
      eqTerm_minusTerm_plusTerm_invTerm divTerm eqTerm_divTerm_com
      divTerm_multTerm_l).
Hints Resolve spminusf_multTerm.

Local spminusf_minusTerm_l:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (minuspf p q) r) (minuspf (spminusf a b nZb p r) q)) :=
   (spminusf_minusTerm_l
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l
      plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc
      eqTerm_plusTerm_comp multTerm multTerm_ltM_l zeroP_multTerm_l multTerm_eq
      eqTerm_multTerm_comp invTerm zeroP_invTerm_zeroP invTerm_invol T1
      T1_multTerm_l mult_invTerm_com eqTerm_invTerm_comp invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm eqTerm_minusTerm_plusTerm_invTerm
      minusTerm_ltM_l invTerm_ltM_l minusTerm_eq divTerm eqTerm_divTerm_com).
Hints Resolve spminusf_minusTerm_l.

Local spminusf_plusTerm_l:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (pluspf p q) r) (pluspf (spminusf a b nZb p r) q)) :=
   (spminusf_plusTerm_l
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l
      plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc
      eqTerm_plusTerm_comp multTerm multTerm_ltM_l zeroP_multTerm_l multTerm_eq
      eqTerm_multTerm_comp invTerm zeroP_invTerm_zeroP invTerm_invol T1
      T1_multTerm_l mult_invTerm_com eqTerm_invTerm_comp invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm eqTerm_minusTerm_plusTerm_invTerm
      minusTerm_ltM_l invTerm_ltM_l minusTerm_eq divTerm eqTerm_divTerm_com).
Hints Resolve spminusf_plusTerm_l.

Local spminusf_plusTerm_r:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (pluspf p q) r) (pluspf p (spminusf a b nZb q r))) :=
   (spminusf_plusTerm_r
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l
      plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc
      eqTerm_plusTerm_comp multTerm multTerm_ltM_l zeroP_multTerm_l multTerm_eq
      eqTerm_multTerm_comp invTerm zeroP_invTerm_zeroP invTerm_invol T1
      T1_multTerm_l mult_invTerm_com eqTerm_invTerm_comp invTerm_eq
      multTerm_zeroP_div nZero_T1 minusTerm eqTerm_minusTerm_plusTerm_invTerm
      minusTerm_ltM_l invTerm_ltM_l minusTerm_eq divTerm eqTerm_divTerm_com).
Hints Resolve spminusf_plusTerm_r.

Local divP_eqTerm_comp:
  (a, b, c:Term) (divP a c) -> (eqTerm a b) ->(divP b c) :=
   (divP_eqTerm_comp
      Term zeroP eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm eqTerm_sym
      multTerm eqTerm_multTerm_comp divTerm eqTerm_divTerm_com).
Hints Resolve divP_eqTerm_comp.

Local divP_minusTerm_comp:
  (a, b, c:Term)
  (divP a c) -> (divP b c) -> (eqM a b) -> ~ (zeroP (minusTerm a b)) ->
  (divP (minusTerm a b) c) :=
   (divP_minusTerm_comp
      Term Monome T2M zeroP eqTerm eqTerm_trans eqTerm_ref zeroP_comp_eqTerm
      eqTerm_sym eqTerm_imp_eq plusTerm eqTerm_plusTerm_comp multTerm
      multTerm_plusTerm_dist_l multTerm_eq eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol mult_invTerm_com eqTerm_invTerm_comp
      invTerm_eq minusTerm eqTerm_minusTerm_plusTerm_invTerm divTerm
      eq_divTerm_plusTerm divTerm_invTerm_l eq_divTerm eqTerm_divTerm_com).
Hints Resolve divP_minusTerm_comp.

Local spminusf_minusTerm:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  (divP a c) -> (divP b c) -> (eqM a b) -> ~ (zeroP (minusTerm a b)) ->
  (eqP
     (spminusf (minusTerm a b) c nZc (minuspf p q) r)
     (minuspf (spminusf a c nZc p r) (spminusf b c nZc q r))) :=
   (spminusf_minusTerm
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eqM plusTerm zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l
      plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc
      eqTerm_plusTerm_comp multTerm multTerm_ltM_l zeroP_multTerm_l
      zeroP_multTerm_r multTerm_plusTerm_dist_l multTerm_plusTerm_dist_r
      multTerm_eq multTerm_assoc multTerm_com eqTerm_multTerm_comp invTerm
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l T1_multTerm_r
      mult_invTerm_com eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div
      nZero_T1 minusTerm eqTerm_minusTerm_plusTerm_invTerm minusTerm_ltM_l
      invTerm_ltM_l minusTerm_eq divTerm eq_divTerm_plusTerm divTerm_invTerm_l
      eq_divTerm eqTerm_divTerm_com divTerm_multTerm_l div_is_T1).
Hints Resolve spminusf_minusTerm.

Local spminusf_minusTerm_z:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  (divP a c) -> (divP b c) -> (eqM a b) -> (zeroP (minusTerm a b)) ->
  (eqP (minuspf p q) (minuspf (spminusf a c nZc p r) (spminusf b c nZc q r))) :=
   (spminusf_minusTerm_z
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eq plusTerm zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l
      plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc
      eqTerm_plusTerm_comp multTerm multTerm_ltM_l zeroP_multTerm_l
      zeroP_multTerm_r multTerm_plusTerm_dist_l multTerm_plusTerm_dist_r
      multTerm_eq eqTerm_multTerm_comp invTerm plusTerm_invTerm_zeroP
      zeroP_invTerm_zeroP invTerm_invol T1 T1_multTerm_l mult_invTerm_com
      eqTerm_invTerm_comp invTerm_eq multTerm_zeroP_div nZero_T1 minusTerm
      eqTerm_minusTerm_plusTerm_invTerm minusTerm_ltM_l invTerm_ltM_l
      minusTerm_eq divTerm eqTerm_divTerm_com).

Local spminusf_plusTerm_z:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  (divP a c) -> (divP b c) -> (eqM a b) -> (zeroP (plusTerm a b)) ->
  (eqP (pluspf p q) (pluspf (spminusf a c nZc p r) (spminusf b c nZc q r))) :=
   (spminusf_plusTerm_z
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_trans eqTerm_ref
      zeroP_comp_eqTerm eqTerm_sym ltM ltM_nonrefl ltM_trans ltM_dec
      eqTerm_imp_eq plusTerm zeroP_plusTermr zeroP_plusTerml plusTerm_comp_l
      plusTerm_comp_r plusTerm_com plusTerm_eq1 plusTerm_eq2 plusTerm_assoc
      eqTerm_plusTerm_comp multTerm multTerm_ltM_l zeroP_multTerm_l
      zeroP_multTerm_r multTerm_plusTerm_dist_r multTerm_eq multTerm_assoc
      eqTerm_multTerm_comp invTerm plusTerm_invTerm_zeroP zeroP_invTerm_zeroP
      invTerm_invol T1 T1_multTerm_l mult_invTerm_com eqTerm_invTerm_comp
      invTerm_eq multTerm_zeroP_div nZero_T1 minusTerm
      eqTerm_minusTerm_plusTerm_invTerm minusTerm_ltM_l invTerm_ltM_l
      minusTerm_eq divTerm divTerm_invTerm_l eqTerm_divTerm_com).

Local divP_order_comp:
  (a, b:Term) (p:(list Term)) (canonical (pX b p)) -> (divP a b) ->
  (canonical (pX a p)) :=
   (divP_order_comp
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_ref eqTerm_sym ltM
      ltM_nonrefl ltM_trans ltM_dec eqTerm_imp_eqM multTerm multTerm_ltM_r
      multTerm_eq T1 T1_multTerm_l T1_is_min_ltM divTerm).
Hints Resolve divP_order_comp.

Local ltP_divP_pX:
  (a, b:Term)
  (p, q:(list Term))
  (canonical (pX a p)) -> (canonical (pX b q)) -> (divP a b) ->(ltP q (pX a p)) :=
   (ltP_divP_pX
      Term Monome T2M zeroP zeroP_dec eqTerm eqTerm_ref eqTerm_sym ltM
      ltM_nonrefl ltM_trans ltM_dec eqTerm_imp_eqM multTerm multTerm_ltM_r
      multTerm_eq T1 T1_multTerm_l T1_is_min_ltM divTerm).
Hints Resolve ltP_divP_pX.

