Section Pair.
Variable A:Set.
Variable Monome:Set.
Variable A0:A.
Variable eqA:A -> A ->Prop.
Require Import Relation_Definitions.
Hypothesis eqA_ref:(reflexive A eqA).
Hypothesis eqA_sym:(symmetric A eqA).
Hypothesis eqA_trans:(transitive A eqA).
Hypothesis eqA_dec:(a, b:A){(eqA a b)}+{~ (eqA a b)}.

Theorem eqA_A0: (a, b:A) (eqA a A0) -> (eqA a b) ->(eqA b A0).
Intros a b H H1; Apply eqA_trans with y := a; Auto.
Qed.
Variable plusA:A -> A ->A.
Hypothesis plusA_assoc:
           (a, b, c:A)(eqA (plusA a (plusA b c)) (plusA (plusA a b) c)).
Hypothesis plusA_com:(a, b:A)(eqA (plusA a b) (plusA b a)).
Hypothesis plusA_eqA_comp:
           (a, b, c, d:A) (eqA a c) -> (eqA b d) ->(eqA (plusA a b) (plusA c d)).

Theorem plusA_eqA_comp_l: (a, b, c:A) (eqA a b) ->(eqA (plusA a c) (plusA b c)).
Intros a b c H'; Apply (plusA_eqA_comp a c b c); Auto.
Qed.

Theorem plusA_eqA_comp_r: (a, b, c:A) (eqA a b) ->(eqA (plusA c a) (plusA c b)).
Intros a b c H'; Apply (plusA_eqA_comp c a c b); Auto.
Qed.
Hypothesis plusA_A0:(a:A)(eqA a (plusA a A0)).

Theorem eqA0_left: (a, b:A) (eqA a A0) ->(eqA b (plusA a b)).
Intros a b H'; Apply eqA_trans with y := (plusA A0 b); Auto.
Apply eqA_trans with y := (plusA b A0); Auto.
Qed.

Theorem eqA0_right: (a, b:A) (eqA b A0) ->(eqA a (plusA a b)).
Intros a b H'; Apply eqA_trans with y := (plusA a A0); Auto.
Qed.
Variable invA:A ->A.
Hypothesis invA_plusA:(a:A)(eqA A0 (plusA a (invA a))).

Theorem invA0: (eqA A0 (invA A0)).
Apply eqA_trans with y := (plusA A0 (invA A0)); Auto.
Apply eqA_trans with y := (plusA (invA A0) A0); Auto.
Qed.
Hypothesis invA_eqA_comp:(a, b:A) (eqA a b) ->(eqA (invA a) (invA b)).

Theorem invA0_comp: (a:A) (eqA a A0) ->(eqA a (invA a)).
Intros a H'; Apply eqA_sym; Apply eqA_trans with y := A0; Auto.
Apply eqA_trans with y := (invA A0); Auto.
Apply eqA_sym; Apply invA0; Auto.
Qed.

Theorem invA0_inv: (a:A)(eqA a (invA (invA a))).
Intros a.
Apply eqA_trans with y := (plusA a A0); Auto.
Apply eqA_trans with y := (plusA A0 a); Auto.
Apply eqA_trans with y := (plusA (plusA (invA a) (invA (invA a))) a); Auto.
Apply eqA_trans with y := (plusA (plusA (invA (invA a)) (invA a)) a); Auto.
Apply eqA_trans with y := (plusA (invA (invA a)) (plusA (invA a) a)); Auto.
Apply eqA_trans with y := (plusA (invA (invA a)) (plusA a (invA a))); Auto.
Apply eqA_trans with y := (plusA (invA (invA a)) A0); Auto.
Qed.
Hypothesis minusA:A -> A ->A.
Hypothesis minusA_def:(a, b:A)(eqA (minusA a b) (plusA a (invA b))).
Hypothesis multA:A -> A ->A.
Hypothesis multA_eqA_comp:
           (a, b, c, d:A) (eqA a c) -> (eqA b d) ->(eqA (multA a b) (multA c d)).

Theorem multA_eqA_comp_l: (a, b, c:A) (eqA a b) ->(eqA (multA a c) (multA b c)).
Intros a b c H'; Apply (multA_eqA_comp a c b c); Auto.
Qed.

Theorem multA_eqA_comp_r: (a, b, c:A) (eqA a b) ->(eqA (multA c a) (multA c b)).
Intros a b c H'; Apply (multA_eqA_comp c a c b); Auto.
Qed.
Hypothesis multA_assoc:
           (a, b, c:A)(eqA (multA a (multA b c)) (multA (multA a b) c)).
Hypothesis multA_com:(a, b:A)(eqA (multA a b) (multA b a)).
Hypothesis multA_dist_l:
           (a, b, c:A)
           (eqA (plusA (multA c a) (multA c b)) (multA c (plusA a b))).

Theorem multA_dist_r:
  (a, b, c:A)(eqA (plusA (multA a c) (multA b c)) (multA (plusA a b) c)).
Intros a b c; Apply eqA_trans with y := (plusA (multA c a) (multA c b)); Auto.
Apply eqA_trans with y := (multA c (plusA a b)); Auto.
Qed.
Variable M1:Monome.

Local Term := A * Monome.

Definition zeroP: Term ->Prop.
Intros H'; Case H'.
Intros a H'1; Exact (eqA a A0).
Defined.

Definition eqTerm: Term -> Term ->Prop.
Intros H'; Case H'.
Intros a a' H'2; Case H'2.
Intros b b'; Exact (eqA a b) /\ a'=b'.
Defined.

Theorem eqTerm_refl: (reflexive Term eqTerm).
Red.
Intros x; Case x; Simpl; Auto.
Qed.

Theorem eqTerm_sym: (symmetric Term eqTerm).
Red.
Intros x y; Case x; Case y; Simpl; Intuition.
Qed.

Theorem eqTerm_trans: (transitive Term eqTerm).
Red.
Intros x y z; Case x; Case y; Case z; Simpl; Intuition.
Apply eqA_trans with y := a0; Auto.
Rewrite H; Auto.
Qed.

Definition T2M: Term ->Monome.
Intros t; Case t; Intros a m; Exact m.
Defined.

Theorem eqTerm_imp_eq: (a, b:Term) (eqTerm a b) ->(T2M a)=(T2M b).
Intros a b; Case a; Case b; Simpl; Intuition.
Qed.
Hypothesis eqM_dec:(x, y:Monome){x=y}+{~ x=y}.

Theorem eqTerm_dec: (x, y:Term){(eqTerm x y)}+{~ (eqTerm x y)}.
Intros x y; Case x; Case y; Simpl.
Intros b2 c2 b3 c3.
Case (eqA_dec b3 b2); Intros eqAZ; Auto.
Case (eqM_dec c3 c2); Intros eqAZ1; Auto.
Right; Red; Intros H'; Elim H'; Intros H'0 H'1; Clear H'; Auto.
Right; Red; Intros H'; Elim H'; Intros H'0 H'1; Clear H'; Auto.
Qed.

Theorem eq_zerop_is_eqTerm:
  (a, b:Term) (T2M a)=(T2M b) -> (zeroP a) -> (zeroP b) ->(eqTerm a b).
Intros a b; Case a; Case b; Simpl; Auto.
Intuition.
Apply eqA_trans with y := A0; Auto.
Qed.

Theorem zeroP_dec: (x:Term){(zeroP x)}+{~ (zeroP x)}.
Intros x; Case x; Simpl.
Intros b H'.
Apply eqA_dec; Auto.
Qed.

Theorem zeroP_comp_eqTerm: (a, b:Term) (zeroP a) -> (eqTerm a b) ->(zeroP b).
Intros a b; Case a; Case b; Simpl; Auto.
Intuition.
Apply eqA_trans with y := a0; Auto; Apply eqp_sym; Auto.
Qed.

Definition plusTerm: Term -> Term ->Term.
Intros x; Case x; Intros b2 c2; Intros y; Case y; Intros b3 c3;
 Exact ((plusA b2 b3), c2).
Defined.

Theorem zeroP_plusTermr:
  (a, b:Term) (T2M a)=(T2M b) -> (zeroP b) ->(eqTerm a (plusTerm a b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intuition.
Apply eqA_trans with y := (plusA a0 A0); Auto.
Qed.

Theorem zeroP_plusTerml:
  (a, b:Term) (T2M a)=(T2M b) -> (zeroP a) ->(eqTerm b (plusTerm a b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intuition.
Apply eqA_trans with y := (plusA A0 a); Auto.
Apply eqA_trans with y := (plusA a A0); Auto.
Qed.

Theorem plusTerm_comp_l:
  (a, b, c:Term) (T2M a)=(T2M c) -> (T2M b)=(T2M c) -> (eqTerm a b) ->
  (eqTerm (plusTerm a c) (plusTerm b c)).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intuition.
Qed.

Theorem plusTerm_comp_r:
  (a, b, c:Term) (T2M c)=(T2M a) -> (T2M c)=(T2M b) -> (eqTerm a b) ->
  (eqTerm (plusTerm c a) (plusTerm c b)).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intuition.
Qed.

Theorem plusTerm_com:
  (x, y:Term) (T2M x)=(T2M y) ->(eqTerm (plusTerm x y) (plusTerm y x)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem plusTerm_eq1:
  (m1, m2:Term) (T2M m1)=(T2M m2) ->(T2M (plusTerm m1 m2))=(T2M m1).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem plusTerm_eq2:
  (m1, m2:Term) (T2M m1)=(T2M m2) ->(T2M (plusTerm m1 m2))=(T2M m2).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem plusTerm_assoc:
  (a, a0, A1:Term) (T2M A1)=(T2M a0) -> (T2M a)=(T2M a0) ->
  (eqTerm (plusTerm (plusTerm A1 a0) a) (plusTerm A1 (plusTerm a0 a))).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Qed.

Theorem eqTerm_plusTerm_com:
  (a, b, c, d:Term)
  (T2M a)=(T2M c) -> (T2M b)=(T2M d) -> (eqTerm a b) -> (eqTerm c d) ->
  (eqTerm (plusTerm a c) (plusTerm b d)).
Intros a b c d; Case a; Case b; Case c; Case d; Simpl; Auto.
Intuition.
Qed.
Hypothesis plusM:Monome -> Monome ->Monome.
Variable ltM:Monome -> Monome ->Prop.
Hypothesis ltM_nonrefl:(x:Monome)~ (ltM x x).
Hypothesis ltM_trans:(transitive Monome ltM).
Hypothesis ltM_dec:(x, y:Monome){(ltM x y)}+{(ltM y x)}+{x=y}.
Hypothesis ltM_wf:(well_founded Monome ltM).
Hypothesis ltM_plusM_r:
           (a, b, c:Monome) (ltM a b) ->(ltM (plusM a c) (plusM b c)).
Hypothesis ltM_plusM_l:
           (a, b, c:Monome) (ltM a b) ->(ltM (plusM c a) (plusM c b)).

Definition multTerm: Term -> Term ->Term.
Intros H'; Case H'; Intros b2 c2 H1'; Case H1'; Intros b3 c3;
 Exact ((multA b2 b3), (plusM c2 c3)).
Defined.

Theorem multTerm_ltM_l:
  (m1, m2, m3:Term) (ltM (T2M m1) (T2M m2)) ->
  (ltM (T2M (multTerm m3 m1)) (T2M (multTerm m3 m2))).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Qed.

Theorem multTerm_ltM_r:
  (m1, m2, m3:Term) (ltM (T2M m1) (T2M m2)) ->
  (ltM (T2M (multTerm m1 m3)) (T2M (multTerm m2 m3))).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Qed.
Hypothesis multA_A0_l:(a:A)(eqA (multA A0 a) A0).

Theorem multA_A0_r: (a:A)(eqA (multA a A0) A0).
Intros a; Apply eqA_trans with y := (multA A0 a); Auto.
Qed.

Theorem zeroP_multTerm_l: (a, b:Term) (zeroP a) ->(zeroP (multTerm a b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros A1 H' b2 H'0 H'1; Apply eqA_trans with y := (multA A0 A1); Auto.
Qed.
Hints Resolve multA_A0_r.

Theorem zeroP_multTerm_r: (a, b:Term) (zeroP a) ->(zeroP (multTerm b a)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros A1 H' b2 H'0 H'1; Apply eqA_trans with y := (multA A1 A0); Auto.
Qed.

Theorem multTerm_plusTerm_dist_l:
  (a, b, c:Term)
  (eqTerm (plusTerm (multTerm a c) (multTerm b c)) (multTerm (plusTerm a b) c)).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intuition; Auto.
Apply multA_dist_r; Auto.
Qed.

Theorem multTerm_plusTerm_dist_r:
  (a, b, c:Term)
  (eqTerm (plusTerm (multTerm c a) (multTerm c b)) (multTerm c (plusTerm a b))).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Qed.

Theorem multTerm_eq:
  (a, b, c, d:Term) (T2M a)=(T2M b) -> (T2M c)=(T2M d) ->
  (T2M (multTerm a c))=(T2M (multTerm b d)).
Intros a b c d; Case a; Case b; Case c; Case d; Simpl; Auto.
Intros H' c0 H'0 c2 H'1 c3 H'2 c4 H'3 H'4; Rewrite H'3; Rewrite H'4; Auto.
Qed.
Hypothesis plusM_assoc:
           (a, b, c:Monome)(plusM a (plusM b c))=(plusM (plusM a b) c).

Theorem multTerm_assoc:
  (a, b, c:Term)(eqTerm (multTerm a (multTerm b c)) (multTerm (multTerm a b) c)).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Qed.
Hypothesis plusM_com:(a, b:Monome)(plusM a b)=(plusM b a).

Theorem multTerm_com: (a, b:Term)(eqTerm (multTerm a b) (multTerm b a)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem eqTerm_multTerm_comp:
  (a, b, c, d:Term) (eqTerm a b) -> (eqTerm c d) ->
  (eqTerm (multTerm a c) (multTerm b d)).
Intros a b c d; Case a; Case b; Case c; Case d; Simpl; Auto.
Intuition; Auto.
Rewrite H4; Rewrite H3; Auto.
Qed.

Definition invTerm: Term ->Term.
Intros H; Case H; Intros b2 c2; Exact ((invA b2), c2).
Defined.

Theorem plusTerm_invTerm_zeroP: (a:Term)(zeroP (plusTerm a (invTerm a))).
Intros a; Case a; Simpl; Auto.
Qed.

Theorem zeroP_invTerm_zeroP: (a:Term) (zeroP a) ->(zeroP (invTerm a)).
Intros a; Case a; Simpl; Auto.
Intros b H' H'0; Apply eqA_trans with y := b; Auto.
Apply eqA_sym; Auto.
Apply invA0_comp; Auto.
Qed.
Hints Resolve invA0_inv.

Theorem invTerm_invol: (a:Term)(eqTerm a (invTerm (invTerm a))).
Intros a; Case a; Simpl; Auto.
Qed.
Variable A1:A.
Hypothesis multA_A1_l:(a:A)(eqA (multA A1 a) a).

Theorem multA_A1_r: (a:A)(eqA (multA a A1) a).
Intros a; Apply eqA_trans with y := (multA A1 a); Auto.
Qed.
Hints Resolve multA_A1_r.

Definition T1: Term.
Exact (A1, M1).
Defined.
Hypothesis plusM_M1_l:(a:Monome)(plusM M1 a)=a.
Hypothesis plusM_M1_r:(a:Monome)(plusM a M1)=a.
Hypothesis A1_diff_A0:~ (eqA A1 A0).

Theorem T1_multTerm_l: (a, b:Term) (eqTerm a T1) ->(eqTerm b (multTerm a b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros a0 m a1 m0 H'; Elim H'; Intros H'0 H'1; Rewrite H'1; Clear H'; Auto.
Split; Auto.
Apply eqA_trans with y := (multA A1 a0); Auto.
Qed.

Theorem T1_multTerm_r: (a, b:Term) (eqTerm a T1) ->(eqTerm b (multTerm b a)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros a0 m a1 m0 H'; Elim H'; Intros H'0 H'1; Rewrite H'1; Clear H'.
Split; Auto.
Apply eqA_trans with y := (multA a0 A1); Auto.
Qed.

Theorem multA_invA_com_l: (a, b:A)(eqA (multA (invA a) b) (invA (multA a b))).
Intros a b; Apply eqA_trans with y := (plusA (invA (multA a b)) A0); Auto.
Apply eqA_trans with y := (plusA A0 (invA (multA a b))); Auto.
Apply eqA_trans with y := (plusA (multA A0 b) (invA (multA a b))); Auto.
Apply eqA_trans
     with y := (plusA (multA (plusA a (invA a)) b) (invA (multA a b))); Auto.
Apply eqA_trans
     with y := (plusA (plusA (multA a b) (multA (invA a) b)) (invA (multA a b)));
 Auto.
Apply eqA_trans
     with y := (plusA (plusA (multA (invA a) b) (multA a b)) (invA (multA a b)));
 Auto.
Apply eqA_trans
     with y := (plusA (multA (invA a) b) (plusA (multA a b) (invA (multA a b))));
 Auto.
Apply eqA_trans with y := (plusA (multA (invA a) b) A0); Auto.
Apply plusA_eqA_comp; Auto.
Apply multA_dist_r.
Qed.
Hints Resolve multA_invA_com_l.

Theorem mult_invTerm_com:
  (a, b:Term)(eqTerm (multTerm (invTerm a) b) (invTerm (multTerm a b))).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem multA_invA_com_r: (a, b:A)(eqA (multA a (invA b)) (invA (multA a b))).
Intros a b; Apply eqA_trans with y := (multA (invA b) a); Auto.
Apply eqA_trans with y := (invA (multA b a)); Auto.
Qed.
Hints Resolve multA_invA_com_r.

Theorem mult_invTerm_com_r:
  (a, b:Term)(eqTerm (multTerm a (invTerm b)) (invTerm (multTerm a b))).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem eqTerm_invTerm_comp:
  (a, b:Term) (eqTerm a b) ->(eqTerm (invTerm a) (invTerm b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intuition.
Qed.

Theorem invTerm_eq: (a:Term)(T2M a)=(T2M (invTerm a)).
Intros a; Case a; Simpl; Auto.
Qed.
Hypothesis M1_min:(a:Monome)~ (ltM a M1).

Theorem T1_is_min_ltM: (a:Term)~ (ltM (T2M a) (T2M T1)).
Intros a; Case a; Simpl; Auto.
Qed.

Theorem T1_eq: (a, b:Term) (eqTerm a T1) ->(T2M b)=(T2M (multTerm a b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intuition.
Rewrite H1; Auto.
Qed.

Theorem eqTerm_T1_eq: (a:Term) (eqTerm a T1) ->(T2M a)=(T2M T1).
Intros a; Case a; Simpl; Auto.
Intuition.
Qed.

Definition minusTerm: Term -> Term ->Term.
Intros H; Case H; Intros b2 c2 H'; Case H'; Intros b3 c3;
 Exact ((minusA b2 b3), c2).
Defined.

Theorem eqTerm_minusTerm_plusTerm_invTerm:
  (a, b:Term)(eqTerm (minusTerm a b) (plusTerm a (invTerm b))).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem minusTerm_ltM_l:
  (a, b, c:Term) (T2M a)=(T2M b) -> (ltM (T2M a) (T2M c)) ->
  (ltM (T2M (minusTerm a b)) (T2M c)).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Qed.

Theorem invTerm_ltM_l:
  (a, c:Term) (ltM (T2M a) (T2M c)) ->(ltM (T2M (invTerm a)) (T2M c)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem minusTerm_eq:
  (m1, m2:Term) (T2M m1)=(T2M m2) ->(T2M (minusTerm m1 m2))=(T2M m1).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.
Variable divA:A ->(b:A) (nZa:~ (eqA b A0))A.
Hypothesis divA_is_multA:
           (a, b:A) (nZb:~ (eqA b A0))(eqA a (multA (divA a b nZb) b)).
Variable minusM:Monome -> Monome ->Monome.

Definition divTerm: Term ->(b:Term) (nZb:~ (zeroP b))Term.
Intros H; Case H; Intros b2 c2 H'; Case H'; Intros b3 c3; Simpl.
Intros nZb3; Exact ((divA b2 b3 nZb3), (minusM c2 c3)).
Defined.
Hypothesis divA_eqA_comp:
           (a, b, c, d:A)
           (nZb:~ (eqA b A0)) (nZd:~ (eqA d A0)) (eqA a c) -> (eqA b d) ->
           (eqA (divA a b nZb) (divA c d nZd)).
Hypothesis divA_multA_comp_r:
           (a, b, c:A) (nZc:~ (eqA c A0))
           (eqA (divA (multA a b) c nZc) (multA (divA a c nZc) b)).

Theorem divA_multA_comp_l:
  (a, b, c:A) (nZc:~ (eqA c A0))
  (eqA (divA (multA a b) c nZc) (multA a (divA b c nZc))).
Intros a b c nZc; Apply eqA_trans with y := (divA (multA b a) c nZc); Auto.
Apply eqA_trans with y := (multA (divA b c nZc) a); Auto.
Qed.
Hints Resolve divA_multA_comp_l.

Theorem invA_is_invA1: (a:A)(eqA (invA a) (multA (invA A1) a)).
Intros a; Apply eqA_trans with y := (plusA (invA a) A0); Auto.
Apply eqA_trans with y := (plusA (invA a) (multA A0 a)); Auto.
Apply eqA_trans with y := (plusA (invA a) (multA (plusA A1 (invA A1)) a)); Auto.
Apply eqA_trans
     with y := (plusA (invA a) (plusA (multA A1 a) (multA (invA A1) a))); Auto.
Apply plusA_eqA_comp; Auto.
Apply eqA_sym; Auto.
Apply multA_dist_r.
Apply eqA_trans with y := (plusA (invA a) (plusA a (multA (invA A1) a))); Auto.
Apply eqA_trans with y := (plusA (plusA (invA a) a) (multA (invA A1) a)); Auto.
Apply eqA_trans with y := (plusA (plusA a (invA a)) (multA (invA A1) a)); Auto.
Apply eqA_trans with y := (plusA A0 (multA (invA A1) a)); Auto.
Apply eqA_trans with y := (plusA (multA (invA A1) a) A0); Auto.
Qed.

Theorem divA_A0_l: (a:A) (nZa:~ (eqA a A0))(eqA (divA A0 a nZa) A0).
Intros a nZa; Apply eqA_trans with y := (divA (multA A0 a) a nZa).
Apply divA_eqA_comp; Auto.
Apply eqA_trans with y := (multA A0 (divA a a nZa)); Auto.
Qed.
Hints Resolve divA_A0_l.

Theorem A_sep: (a, b:A) (eqA (multA a b) A0) ->(eqA a A0) \/ (eqA b A0).
Intros a b H'; Case (eqA_dec a A0); Auto; Case (eqA_dec b A0); Auto.
Intros nZb nZa; Case nZa.
Apply eqA_trans with y := (multA (divA a b nZb) b); Auto.
Apply eqA_trans with y := (divA (multA a b) b nZb); Auto.
Apply eqA_trans with y := (divA A0 b nZb); Auto.
Qed.
Hints Resolve A_sep.

Theorem divA_A1: (a:A) (nZa:~ (eqA a A0))(eqA (divA a a nZa) A1).
Intros a nZa; Apply eqA_trans with y := (divA (multA A1 a) a nZa); Auto.
Apply eqA_trans with y := (multA (divA A1 a nZa) a); Auto.
Qed.
Hints Resolve divA_A1.

Theorem eq_divTerm_plusTerm:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (T2M a)=(T2M b) ->
  (eqTerm a (multTerm (divTerm a c nZc) c)) ->
  (eqTerm b (multTerm (divTerm b c nZc) c)) ->
  (eqTerm
     (divTerm (plusTerm a b) c nZc)
     (plusTerm (divTerm a c nZc) (divTerm b c nZc))).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intros a0 m a1 m0 a2 m1 nZc H' H'0 H'1; Split; Auto.
Case H'1; Intros H'3 H'4; Clear H'1.
Case H'0; Intros H'2 H'5; Clear H'0.
Apply eqA_trans
     with y :=
          (divA
             (plusA (multA (divA a2 a0 nZc) a0) (multA (divA a1 a0 nZc) a0)) a0
             nZc); Auto.
Apply eqA_trans
     with y :=
          (divA (multA (plusA (divA a2 a0 nZc) (divA a1 a0 nZc)) a0) a0 nZc);
 Auto.
Apply divA_eqA_comp; Auto.
Apply multA_dist_r; Auto.
Apply eqA_trans
     with y :=
          (multA (plusA (divA a2 a0 nZc) (divA a1 a0 nZc)) (divA a0 a0 nZc));
 Auto.
Apply eqA_trans with y := (multA (plusA (divA a2 a0 nZc) (divA a1 a0 nZc)) A1);
 Auto.
Qed.

Theorem divTerm_invTerm_l:
  (a, b:Term) (nZa:~ (zeroP a)) (eqTerm b (multTerm (divTerm b a nZa) a)) ->
  (eqTerm (divTerm (invTerm b) a nZa) (invTerm (divTerm b a nZa))).
Intros a b; Case a; Case b; Simpl; Auto.
Intros d c A0 c0 nZA2 H'0; Split; Auto.
Case H'0; Intros H'1 H'2; Clear H'0.
Apply eqA_trans with y := (divA (invA (multA (divA d A2 nZA2) A2)) A2 nZA2);
 Auto.
Apply eqA_trans with y := (divA (multA (invA (divA d A2 nZA2)) A2) A2 nZA2);
 Auto.
Apply eqA_trans with y := (multA (invA (divA d A2 nZA2)) (divA A2 A2 nZA2));
 Auto.
Apply eqA_trans with y := (multA (invA (divA d A2 nZA2)) A1); Auto.
Qed.
Hypothesis divA_invA_r:
           (a, b:A) (nZb:~ (eqA b A0)) (nZib:~ (eqA (invA b) A0))
           (eqA (divA a (invA b) nZib) (invA (divA a b nZb))).

Theorem divTerm_invTerm_r:
  (a, b:Term)
  (nZa:~ (zeroP a))
  (nZia:~ (zeroP (invTerm a))) (eqTerm b (multTerm (divTerm b a nZa) a)) ->
  (eqTerm (divTerm b (invTerm a) nZia) (invTerm (divTerm b a nZa))).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem eq_divTerm:
  (a, b, c, d:Term)
  (nZc:~ (zeroP c)) (nZd:~ (zeroP d)) (T2M a)=(T2M b) -> (T2M c)=(T2M d) ->
  (T2M (divTerm a c nZc))=(T2M (divTerm b d nZd)).
Intros a b c d; Case a; Case b; Case c; Case d; Simpl; Auto.
Intros A0 c0 d2 c2 H' c3 H'0 c4 H'1 H'2 H'3 H'4; Rewrite H'3; Rewrite H'4; Auto.
Qed.

Theorem eqTerm_divTerm_com:
  (a, b, c, d:Term)
  (nZc:~ (zeroP c)) (nZd:~ (zeroP d)) (eqTerm a b) -> (eqTerm c d) ->
  (eqTerm (divTerm a c nZc) (divTerm b d nZd)).
Intros a b c d; Case a; Case b; Case c; Case d; Simpl; Auto.
Intros A0 c0 d2 c2 d3 c3 d4 c4 nZd2 H'0 H'1 H'2.
Case H'2; Intros H'3 H'4; Clear H'2.
Case H'1; Intros H'2 H'5; Clear H'1.
Split; Auto.
Rewrite H'5; Rewrite H'4; Auto.
Qed.
Hypothesis plusM_minusM_com:(a, b:Monome)(minusM (plusM a b) b)=a.

Theorem divTerm_multTerm_l:
  (a, b, c:Term) (nZa:~ (zeroP a)) (eqTerm b (multTerm (divTerm b a nZa) a)) ->
  (eqTerm (divTerm (multTerm c b) a nZa) (multTerm c (divTerm b a nZa))).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intros d c0 A0 c2 d2 c3 nZd2 H'0; Case H'0; Intros H'1 H'2; Auto.
Split; Auto.
Rewrite H'2.
Rewrite plusM_minusM_com.
Rewrite plusM_assoc.
Rewrite plusM_minusM_com; Auto.
Qed.

Theorem divTerm_multTerm_r:
  (a, b, c:Term) (nZa:~ (zeroP a)) (eqTerm b (multTerm (divTerm b a nZa) a)) ->
  (eqTerm (divTerm (multTerm b c) a nZa) (multTerm (divTerm b a nZa) c)).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intros a0 m a1 m0 a2 m1 nZa H'; Split; Auto.
Elim H'; Intros H'0 H'1; Rewrite H'1; Clear H'.
Rewrite plusM_minusM_com; Auto.
Rewrite <- plusM_assoc.
Rewrite (plusM_com m1 m).
Rewrite plusM_assoc.
Rewrite plusM_minusM_com; Auto.
Qed.

Theorem multTerm_zeroP_div:
  (a, b:Term) (zeroP (multTerm a b)) ->(zeroP a) \/ (zeroP b).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.
Hypothesis minusM_id:(a:Monome)(minusM a a)=M1.

Theorem div_is_T1: (a:Term) (nZa:~ (zeroP a))(eqTerm (divTerm a a nZa) T1).
Intros a; Case a; Simpl; Auto.
Qed.

Theorem T1_nz: ~ (zeroP T1).
Simpl; Auto.
Qed.

Theorem divTerm_eq:
  (a, b, c:Term) (T2M a)=(T2M b) -> ~ (zeroP a) -> ~ (zeroP b) ->
  (nZc:~ (zeroP c)) (eqTerm a (multTerm (divTerm a c nZc) c)) ->
  (eqTerm b (multTerm (divTerm b c nZc) c)).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intuition.
Rewrite <- H8; Auto.
Qed.
Variable minusM_clean:Monome -> Monome ->Monome * bool.

Local gb: Monome * bool ->bool.
Intros H'; Case H'; Auto.
Defined.

Local gm: Monome * bool ->Monome.
Intros H'; Case H'; Auto.
Defined.
Hypothesis minusM_dec1:
           (a, b:Monome) (gb (minusM_clean a b))=true ->
           (gm (minusM_clean a b))=(minusM a b) /\ (plusM (minusM a b) b)=a.
Hypothesis minusM_dec2:
           (a, b:Monome) (gb (minusM_clean a b))=false ->
           ~ (plusM (minusM a b) b)=a.

Definition mk_clean: (a, b:Monome){c:Monome * bool | c=(minusM_clean a b)}.
Intros a b; Exists (minusM_clean a b); Auto.
Qed.

Theorem divTerm_dec:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b))
  {(eqTerm a (multTerm (divTerm a b nZb) b))}+
  {~ (eqTerm a (multTerm (divTerm a b nZb) b))}.
Intros a b; Case a; Case b; Simpl; Auto.
Intros b2 c2 b3 c3.
Intros Zp1 Zp2.
Case (mk_clean c3 c2).
Intros x; Case x.
Intros c b4; Case b4.
Intros H0; Left; Simpl; Auto.
Generalize (minusM_dec1 c3 c2); Rewrite <- H0; Simpl; Auto.
Intuition.
Elim H1; [Intros H'0 H'1; Rewrite H'1; Clear H1 | Clear H1]; Auto.
Intros H0; Right; Red; Intros dviP_H; Inversion dviP_H.
Generalize (minusM_dec2 c3 c2); Simpl; Auto.
Intros H'; LApply H'; [Intros H'0; Apply H'0; Clear H' | Clear H'].
Rewrite <- H1; Auto.
Rewrite <- H0; Simpl; Auto.
Qed.

Theorem zeroP_divTerm:
  (a, b:Term) (zeroP a) ->(nZb:~ (zeroP b))(zeroP (divTerm a b nZb)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros d H' A0 H'0 H'1 nZd; Auto.
Apply eqA_trans with y := (divA A0 d nZd); Auto.
Qed.

Theorem divTerm_on_eqM:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b)) (T2M a)=(T2M b) ->(eqTerm a (multTerm (divTerm a b nZb) b)).
Intros a b; Case a; Case b; Simpl; Auto.
Intros d c A0 c0 H' H'0 H'1; Rewrite <- H'1.
Rewrite minusM_id; Auto.
Qed.

Theorem divTerm_on_eqM_eqM:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b)) (T2M a)=(T2M b) ->(T2M (divTerm a b nZb))=(T2M T1).
Intros a b; Case a; Case b; Simpl; Auto.
Intros b2 c b3 c0 H' H'0 H'1; Rewrite H'1; Auto.
Qed.

Theorem divTerm_on_plusM_minusM:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b)) (T2M a)=(plusM (minusM (T2M a) (T2M b)) (T2M b)) ->
  (eqTerm a (multTerm (divTerm a b nZb) b)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.
Variable maxM:Monome -> Monome ->Monome.
Hypothesis maxM_ppcm:
           (a, b, c:Monome)
           c=(plusM (minusM c a) a) -> c=(plusM (minusM c b) b) ->
           c=(plusM (minusM c (maxM a b)) (maxM a b)).
Hypothesis maxM_prop_l:(a, b:Monome)(maxM a b)=(plusM (minusM (maxM a b) a) a).
Hypothesis maxM_prop_r:(a, b:Monome)(maxM a b)=(plusM (minusM (maxM a b) b) b).

Definition ppc: Term -> Term ->Term.
Intros H; Case H; Intros b2 c2 H'; Case H'; Intros b3 c3; Simpl;
 Exact (A1, (maxM c2 c3)).
Defined.
Hypothesis maxM_com:(a, b:Monome)(maxM a b)=(maxM b a).

Theorem pcc_com: (a, b:Term)(eqTerm (ppc a b) (ppc b a)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem divTerm_ppc:
  (a, b, c:Term)
  (nZa:~ (zeroP a))
  (nZb:~ (zeroP b))
  (nZppab:~ (zeroP (ppc a b)))
  (eqTerm c (multTerm (divTerm c a nZa) a)) ->
  (eqTerm c (multTerm (divTerm c b nZb) b)) ->
  (eqTerm c (multTerm (divTerm c (ppc a b) nZppab) (ppc a b))).
Intros a b c; Case a; Case b; Case c; Simpl; Auto.
Intuition.
Qed.

Theorem divTerm_ppcl:
  (a, b:Term) (nZa:~ (zeroP a)) ~ (zeroP b) ->
  (eqTerm (ppc a b) (multTerm (divTerm (ppc a b) a nZa) a)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem divTerm_ppcr:
  (a, b:Term) ~ (zeroP a) ->
  (nZb:~ (zeroP b))(eqTerm (ppc a b) (multTerm (divTerm (ppc a b) b nZb) b)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.

Theorem ppc_nZ: (a, b, c:Term) ~ (zeroP a) -> ~ (zeroP b) ->~ (zeroP (ppc a b)).
Intros a b; Case a; Case b; Simpl; Auto.
Qed.
Hypothesis divA_rew:
           (a, b:A) (nZ1, nZ2:~ (eqA b A0))(divA a b nZ1)=(divA a b nZ2).

Theorem divTerm_rew:
  (a, b:Term) (nZ1, nZ2:~ (zeroP b))(divTerm a b nZ1)=(divTerm a b nZ2).
Intros a b; Case a; Case b; Simpl; Auto.
Intros a0 m a1 m0 nZ1 nZ2.
Rewrite (divA_rew a1 a0 nZ1 nZ2); Auto.
Qed.
End Pair.

