
(****************************************************************************
                                                                           
          Buchberger : multiplication over polynomials                     
                                                                           
          Laurent Thery March 98 (revised Mai 98)                          
                                                                           
***************************************************************************
 definition of smult*)
Section SMult.
Load "shyp_smult".
Load "mod_smultm_lm".

Fixpoint multpf[p:(list Term)]: (list Term) ->(list Term) :=
   [q:(list Term)]
   (<(list Term)>Case p of
     pO [a:Term] [p':(list Term)](pluspf (multm_lm a q) (multpf p' q)) end).

Theorem canonical_multpf:
  (l1, l2:(list Term)) (canonical l1) -> (canonical l2) ->
  (canonical (multpf l1 l2)).
Intros l1; Elim l1; Simpl; Auto.
Intros a l H' l2 H'0 H'1; Try Assumption.
Apply canonical_pluspf; Auto.
Apply canonical_multlm_lm; Auto.
Apply canonical_nzeroP with p := l; Auto.
Apply H'; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Qed.
Hints Resolve canonical_multpf.

Theorem pluspf_pX:
  (p:(list Term)) (a:Term) (canonical (pX a p)) ->
  (eqP (pX a p) (pluspf (pX a pO) p)).
Intros p; Case p; Clear p; Auto.
Intros a H'; Apply eqp_sym; Auto.
Intros a1 p a0 H'.
Cut (canonical (pX a1 p));
 [Intros Op0 | Apply canonical_imp_canonical with a := a0]; Auto.
Apply eqp_trans with y := (pX a0 (pluspf pO (pX a1 p))); Auto.
Apply eqp_sym; Auto.
Apply pluspf_inv1; Auto.
Apply canonical_imp_order with l := p; Auto.
Qed.

Theorem multpf_disr_pX:
  (p, q:(list Term)) (a:Term) (canonical p) -> (canonical (pX a q)) ->
  (eqP (multpf p (pX a q)) (pluspf (multm_lm a p) (multpf p q))).
Intros p; Elim p; Clear p; Simpl; Auto.
Intros q a H' H'0; Apply eqp_sym; Auto.
Intros b p Rec q a Op0 Op1.
Cut (canonical q); [Intros Op2 | Apply canonical_imp_canonical with a := a];
 Auto.
Cut (canonical p); [Intros Op3 | Apply canonical_imp_canonical with a := b];
 Auto.
Cut ~ (zeroP a); [Intros nZ0 | Apply canonical_nzeroP with p := q]; Auto.
Cut ~ (zeroP b); [Intros nZ1 | Apply canonical_nzeroP with p := p]; Auto.
Cut ~ (zeroP (multTerm b a)); [Intros nZ2 | Idtac]; Auto.
Cut ~ (zeroP (multTerm a b)); [Intros nZ3 | Idtac]; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pX (multTerm b a) (multm_lm b q))
             (pluspf (multm_lm a p) (multpf p q))); Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf (pX (multTerm b a) pO) (multm_lm b q))
             (pluspf (multm_lm a p) (multpf p q))); Auto.
Cut (canonical (pX (multTerm b a) (multm_lm b q))); [Intros Op4 | Idtac]; Auto.
Apply eqp_pluspf_com; Auto.
Apply pluspf_pX; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pX (multTerm b a) pO)
             (pluspf (multm_lm b q) (pluspf (multm_lm a p) (multpf p q)))); Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pX (multTerm b a) pO)
             (pluspf (multm_lm b q) (pluspf (multpf p q) (multm_lm a p)))); Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pX (multTerm b a) pO)
             (pluspf (pluspf (multm_lm b q) (multpf p q)) (multm_lm a p))); Auto.
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pX (multTerm b a) pO)
             (pluspf (multm_lm a p) (pluspf (multm_lm b q) (multpf p q)))); Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf (pX (multTerm b a) pO) (multm_lm a p))
             (pluspf (multm_lm b q) (multpf p q))); Auto.
Apply eqp_sym; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans with y := (pluspf (pX (multTerm a b) pO) (multm_lm a p)); Auto.
Apply eqp_sym; Auto.
Apply pluspf_pX; Auto.
Qed.

Theorem multpf_head:
  (p, q:(list Term)) (a, b:Term) (canonical (pX a p)) -> (canonical (pX b q)) ->
  (Ex [c:(list Term)] (multpf (pX a p) (pX b q))=(pX (multTerm a b) c)).
Intros p; Elim p; Clear p.
Simpl; Intros q a b H' H'0.
Cut (canonical q); [Intros Op0 | Apply canonical_imp_canonical with a := b];
 Auto.
Exists (multm_lm a q); Simpl; Auto.
Rewrite <- (pO_pluspf_inv2
              ? ? T2M zeroP zeroP_dec ltM ltM_nonrefl ltM_trans ltM_dec plusTerm);
 Auto.
Intros a1 p H' q a0 b H'0 H'1; Auto.
Elim (H' q a1 b); Simpl; [Intros c E | Idtac | Idtac]; Auto.
2:Apply canonical_imp_canonical with a := a0; Auto.
Rewrite E.
Exists (pluspf (multm_lm a0 q) (pX (multTerm a1 b) c)); Auto.
Rewrite <- pluspf_inv1_eq; Auto.
Apply multTerm_order_r; Auto.
Apply canonical_imp_order with l := p; Auto.
Qed.

Theorem in_multpf_head:
  (p, q:(list Term)) (a, b:Term) (canonical (pX a p)) -> (canonical (pX b q)) ->
  (In (multTerm a b) (multpf (pX a p) (pX b q))).
Intros p q a b H' H'0.
Elim (multpf_head p q a b); [Intros c E; Rewrite E | Idtac | Idtac]; Simpl; Auto.
Qed.

Theorem multpf_comp:
  (p, r:(list Term)) (eqP p r) ->
  (q, s:(list Term))
  (canonical p) ->
  (canonical q) -> (canonical r) -> (canonical s) -> (eqP q s) ->
  (eqP (multpf p q) (multpf r s)).
Intros p r H'; Elim H'; Simpl; Auto.
Intros ma mb p0 q H'0 H'1 H'2 q0 s H'3 H'4 H'5 H'6 H'7.
Cut (canonical p0);
 [Intros Op0 | Apply canonical_imp_canonical with a := ma; Auto].
Cut (canonical q);
 [Intros Op1 | Apply canonical_imp_canonical with a := mb; Auto].
Cut ~ (zeroP ma); [Intros Z0 | Apply canonical_nzeroP with p := p0; Auto].
Cut ~ (zeroP mb); [Intros Z1 | Apply canonical_nzeroP with p := q]; Auto.
Qed.

Theorem multpf_com:
  (p, q:(list Term)) (canonical p) -> (canonical q) ->
  (eqP (multpf p q) (multpf q p)).
Intros p; Elim p; Simpl; Auto.
Intros q; Elim q; Simpl; Auto.
Intros a l H' H'0 H'1.
Cut (canonical l); [Intros Op0 | Apply canonical_imp_canonical with a := a];
 Auto.
Cut ~ (zeroP a); [Intros Z0 | Apply canonical_nzeroP with p := l]; Auto.
Apply eqp_trans with y := (pluspf pO pO); Auto.
Apply eqp_sym; Auto.
Intros a l H' q H'0 H'1.
Cut (canonical l); [Intros Op0 | Apply canonical_imp_canonical with a := a];
 Auto.
Cut ~ (zeroP a); [Intros Z0 | Apply canonical_nzeroP with p := l]; Auto.
Apply eqp_trans with y := (pluspf (multm_lm a q) (multpf q l)); Auto.
Apply eqp_sym; Auto.
Apply multpf_disr_pX; Auto.
Qed.

Definition smult: poly -> poly ->poly.
Intros sp1 sp2.
Case sp1; Case sp2.
Intros p1 H'1 p2 H'2; Exists (multpf p1 p2); Auto.
Defined.
End SMult.
