
(****************************************************************************
                                                                           
          Buchberger : multiplication by a term                            
                                                                           
          Laurent Thery March 97 (revised Mai 98)                          
                                                                           
***************************************************************************
 definition of smultm_lm *)
Section SMultm_lm.
Load "shyp_smultm_lm".
Load "mod_splus".

Definition multm_lm: Term -> (list Term) ->(list Term).
Intros a p; Elim p; Clear p.
Exact pO.
Intros b p1 p'1.
Exact (pX (multTerm a b) p'1).
Defined.

Theorem multTerm_order_l:
  (m1, m2, m3:Term) (order m1 m2) ->(order (multTerm m3 m1) (multTerm m3 m2)).
Exact multTerm_ltM_l.
Qed.
Hints Resolve multTerm_order_l.

Theorem multTerm_order_r:
  (m1, m2, m3:Term) (order m1 m2) ->(order (multTerm m1 m3) (multTerm m2 m3)).
Exact multTerm_ltM_r.
Qed.
Hints Resolve multTerm_order_r.

Theorem multTerm_eqM:
  (a, b, c, d:Term) (eqM a b) -> (eqM c d) ->(eqM (multTerm a c) (multTerm b d)).
Exact multTerm_eq.
Qed.
Hints Resolve multTerm_eqM.

Theorem invTerm_eqM: (a:Term)(eqM a (invTerm a)).
Exact invTerm_eq.
Qed.
Hints Resolve invTerm_eqM.

Theorem T1_is_min: (a:Term)~ (order a T1).
Exact T1_is_min_ltM.
Qed.
Hints Resolve T1_is_min.

Lemma multm_order_l:
  (l:(list Term)) (m1, m2:Term) ~ (zeroP m1) -> (canonical (pX m2 l)) ->
  (canonical (pX (multTerm m1 m2) (multm_lm m1 l))).
Intros l; Elim l; Simpl; Auto.
Intros m1 m2 H' H'0.
Apply canonical1; Auto.
Red; Intros H'1; Apply H'.
Elim (multTerm_zeroP_div m1 m2); [Intros H'5 | Intros H'5 | Idtac]; Auto.
LApply (canonical_nzeroP m2 pO);
 [Intros H'4; Apply H'4 Orelse Elim H'4 | Idtac]; Auto.
Intros a l0 H' m1 m2 H'0 H'1.
Apply canonical_cons; Auto.
Apply multTerm_order_l; Apply canonical_imp_order with l := l0; Auto.
Red; Intros H'2; Apply H'0.
Elim (multTerm_zeroP_div m1 m2); [Intros H'5 | Intros H'5 | Idtac]; Auto.
LApply (canonical_nzeroP m2 (pX a l0));
 [Intros H'4; Apply H'4 Orelse Elim H'4 | Idtac]; Auto.
Apply H'; Auto.
Apply canonical_imp_canonical with a := m2; Auto.
Qed.

Lemma canonical_multlm_lm:
  (m:Term) (l:(list Term)) ~ (zeroP m) -> (canonical l) ->
  (canonical (multm_lm m l)).
Intros m l; Elim l; Simpl; Auto.
Intros a l0 H' H'0 H'1.
Apply multm_order_l; Auto.
Qed.

Definition tmultm_lm: Term -> (list Term) ->(list Term).
Intros a; Case (zeroP_dec a); Intros Z0.
Intros H'; Exact pO.
Intros p; Exact (multm_lm a p).
Defined.

Theorem tmultlm_lm_zerop_eqp_pO:
  (p:(list Term)) (a:Term) (zeroP a) ->(eqP (tmultm_lm a p) pO).
Intros p a; Unfold tmultm_lm; Case (zeroP_dec a); Auto.
Intros H' H'0; Elim H'; Auto.
Qed.

Theorem multlm_lm_eqp_pO_pO:
  (p:(list Term)) (a:Term) (eqP p pO) ->(eqP (multm_lm a p) pO).
Intros p a H'; Inversion H'; Auto.
Qed.

Theorem multlm_lm_dist1:
  (p:(list Term))
  (a, b:Term)
  (eqM a b) ->
  ~ (zeroP a) -> ~ (zeroP b) -> ~ (zeroP (plusTerm a b)) -> (canonical p) ->
  (eqP (multm_lm (plusTerm a b) p) (pluspf (multm_lm a p) (multm_lm b p))).
Intros p; Elim p; Simpl; Auto.
Intros; Apply eqp_sym; Auto.
Intros a l H' a0 b H'0 H'1 H'2 H'3 H'4.
Cut (canonical l); Try Apply canonical_imp_canonical with a := a; Auto;
 Intros C0.
Apply eqp_trans
     with y :=
          (pX
             (plusTerm (multTerm a0 a) (multTerm b a))
             (pluspf (multm_lm a0 l) (multm_lm b l))); Auto.
Apply eqp_trans
     with y :=
          (pX
             (plusTerm (multTerm a0 a) (multTerm b a))
             (pluspf (multm_lm a0 l) (multm_lm b l))); Auto.
Apply pluspf_inv3b; Auto.
Red; Intros tH'5; Absurd (zeroP (multTerm (plusTerm a0 b) a)); Auto.
Red; Intros H'5; Try Exact H'5.
Elim (multTerm_zeroP_div (plusTerm a0 b) a); [Intros H'9 | Intros H'9 | Idtac];
 Auto.
Absurd (zeroP a); Auto.
Apply canonical_nzeroP with p := l; Auto.
Apply zeroP_comp_eqTerm with a := (plusTerm (multTerm a0 a) (multTerm b a));
 Auto.
Qed.

Theorem multlm_lm_dist2:
  (p:(list Term))
  (a, b:Term)
  (eqM a b) ->
  ~ (zeroP a) -> ~ (zeroP b) -> (zeroP (plusTerm a b)) -> (canonical p) ->
  (eqP pO (pluspf (multm_lm a p) (multm_lm b p))).
Intros p; Elim p; Simpl; Auto.
Intros; Apply eqp_sym; Auto.
Intros a l H' a0 b H'0 H'1 H'2 H'3 H'4.
Cut (canonical l); Try Apply canonical_imp_canonical with a := a; Auto;
 Intros C0.
Apply eqp_trans with y := (pluspf (multm_lm a0 l) (multm_lm b l)); Auto.
Apply pluspf_inv3a; Auto.
Apply zeroP_comp_eqTerm with a := (multTerm (plusTerm a0 b) a); Auto.
Qed.

Theorem multlm_lm_T1:
  (p:(list Term)) (a:Term) (eqTerm a T1) ->(eqP (multm_lm a p) p).
Intros p; Elim p; Auto.
Simpl; Auto.
Qed.

Theorem T1_eqM: (a, b:Term) (eqTerm a T1) ->(eqM b (multTerm a b)).
Exact T1_eq.
Qed.
Hints Resolve T1_eqM.

Theorem eqTerm_T1_eqM: (a:Term) (eqTerm a T1) ->(eqM a T1).
Exact eqTerm_T1_eq.
Qed.
Hints Resolve eqTerm_T1_eqM.

Theorem multlm_lm_invTerm:
  (p:(list Term)) (a:Term) (eqTerm a T1) ->
  (eqP (pluspf p (multm_lm (invTerm a) p)) pO).
Intros p; Elim p; Simpl; Auto.
Intros a l H' a0 H'0.
Apply eqp_trans with y := (pluspf l (multm_lm (invTerm a0) l)); Auto.
Apply eqp_sym; Apply pluspf_inv3a; Auto.
Apply eqM_trans with y := (multTerm a0 a); Auto.
Apply zeroP_comp_eqTerm with a := (plusTerm a (invTerm a)); Auto.
Apply plusTerm_comp_r; Auto.
2:Apply eqTerm_trans with y := (invTerm (multTerm a0 a)); Auto.
Change (eqM a (multTerm (invTerm a0) a)).
Apply eqM_trans with y := (multTerm T1 a); Auto.
Apply eqM_trans with y := (multTerm a0 a); Auto.
Qed.

Theorem multlm_lm_multTerm:
  (p:(list Term)) (a, b:Term)
  (eqP (multm_lm (multTerm a b) p) (multm_lm a (multm_lm b p))).
Intros p; Elim p; Simpl; Auto.
Qed.

Theorem multlm_lm_comA:
  (p:(list Term)) (a, b:Term)
  (eqP (multm_lm (multTerm a b) p) (multm_lm (multTerm b a) p)).
Intros p; Elim p; Simpl; Auto.
Qed.

Theorem multm_lm_comp:
  (a, b:Term) (p, q:(list Term)) (eqTerm a b) -> (eqP p q) ->
  (eqP (multm_lm a p) (multm_lm b q)).
Intros a b p q H' H'0; Elim H'0; Simpl; Auto.
Qed.

Theorem multm_lm_ltP_comp:
  (a:Term) (p, q:(list Term)) (ltP p q) ->(ltP (multm_lm a p) (multm_lm a q)).
Intros a p q H'; Elim H'; Auto.
Simpl; Auto.
Intros x y p0 q0 H'0; Simpl; Apply ltP_hd; Auto.
Intros x y p0 q0 H'0 H'1 H'2; Simpl; Apply ltP_tl; Auto.
Qed.

Theorem zeroP_nonZerop_imp_zeroPl:
  (a, b:Term) ~ (zeroP a) -> (zeroP b) ->(zeroP (multTerm b a)).
Intros a b H' H'0; Elim (multTerm_zeroP_div b a); Auto.
Qed.

Theorem zeroP_nonZerop_imp_zeroPr:
  (a, b:Term) ~ (zeroP a) -> (zeroP b) ->(zeroP (multTerm a b)).
Intros a b H' H'0; Elim (multTerm_zeroP_div a b); Auto.
Qed.

Theorem nzeroP_nonZerop_imp_zeroPl:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->~ (zeroP (multTerm a b)).
Intros a b H' H'0; Red; Intros H'1; Elim (multTerm_zeroP_div a b); Auto.
Qed.
Hints Resolve zeroP_nonZerop_imp_zeroPl zeroP_nonZerop_imp_zeroPr
     nzeroP_nonZerop_imp_zeroPl.

Theorem multlm_comp_canonical:
  (p:(list Term)) (a, b:Term) (canonical (pX a p)) -> ~ (zeroP b) ->
  (canonical (pX (multTerm b a) (multm_lm b p))).
Intros p a b H' H'0; Generalize (canonical_multlm_lm b (pX a p)); Simpl; Auto.
Qed.
Hints Resolve multlm_comp_canonical.

Local ffst := (fst (list Term) (list Term)).

Local ssnd := (snd (list Term) (list Term)).

Local ppair := (pair (list Term) (list Term)).

Definition twoP_ind:
  (P:(list Term) -> (list Term) ->Prop)
  ((p, q:(list Term)) ((r, s:(list Term)) (lessP (r, s) (p, q)) ->(P r s)) ->
   (P p q)) ->(p, q:(list Term))(P p q).
Intros P H' p q; Try Assumption.
Change ([pq:(list Term) * (list Term)](P (ffst pq) (ssnd pq)) (p, q)).
Cut (Ex [x:(list Term) * (list Term)] x=(ppair p q)).
Unfold ppair; Intros H'0; Elim H'0; Intros x E; Rewrite <- E; Clear H'0.
Pattern x;
 Apply well_founded_ind
      with A := (list Term) * (list Term) R := lessP 1 := wf_lessP; Auto.
Intros x0 H'0; Apply H'; Auto.
Intros r s H'1.
Apply (H'0 (r, s)).
Generalize H'1; Case x0; Simpl; Auto.
Exists (ppair p q); Auto.
Qed.

Theorem multlm_lm_dist_pluspf:
  (p, q:(list Term)) (a:Term) (canonical p) -> (canonical q) -> ~ (zeroP a) ->
  (eqP (multm_lm a (pluspf p q)) (pluspf (multm_lm a p) (multm_lm a q))).
Intros p q; Pattern p q.
Apply twoP_ind; Auto.
Intros p0; Case p0.
Intros q0 H' a H'0 H'1 H'2.
Apply eqp_trans with y := (multm_lm a q0); Auto.
Apply multm_lm_comp; Auto.
Apply eqp_sym; Auto.
Intros a l q0.
Case q0.
Intros H' a0 H'0 H'1 H'2.
Apply eqp_trans with y := (multm_lm a0 (pX a l)); Auto.
Apply multm_lm_comp; Auto.
Apply eqp_sym; Auto.
Intros a0 l0 H' a1 H'0 H'1 H'2; Simpl.
Case (order_dec a a0); [Intros H0; Case H0; Clear H0 | Idtac]; Intros H0.
Cut (order (multTerm a1 a) (multTerm a1 a0)); [Intros Od0 | Idtac]; Auto.
Apply eqp_trans with y := (multm_lm a1 (pX a0 (pluspf (pX a l) l0))); Auto.
Apply eqp_sym; Auto.
Apply multm_lm_comp; Auto.
Simpl.
Apply eqp_trans
     with y :=
          (pX
             (multTerm a1 a0)
             (pluspf (pX (multTerm a1 a) (multm_lm a1 l)) (multm_lm a1 l0)));
 Auto.
Apply eqpP1; Auto.
Apply eqp_trans with y := (pluspf (multm_lm a1 (pX a l)) (multm_lm a1 l0)); Auto.
Apply H'; Simpl; Auto.
Red; Red; Simpl; Rewrite <- plus_n_Sm; Auto.
Apply canonical_imp_canonical with a := a0; Auto.
Cut (order (multTerm a1 a0) (multTerm a1 a)); [Intros Od0 | Idtac]; Auto.
Apply eqp_trans with y := (multm_lm a1 (pX a (pluspf l (pX a0 l0)))); Auto.
Apply eqp_sym; Auto.
Apply multm_lm_comp; Auto.
Simpl.
Apply eqp_trans
     with y :=
          (pX
             (multTerm a1 a)
             (pluspf (multm_lm a1 l) (pX (multTerm a1 a0) (multm_lm a1 l0))));
 Auto.
Apply eqpP1; Auto.
Apply eqp_trans with y := (pluspf (multm_lm a1 l) (multm_lm a1 (pX a0 l0)));
 Auto.
Apply H'; Simpl; Auto.
Red; Red; Simpl; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Cut (eqM (multTerm a1 a) (multTerm a1 a0)); [Intros Od0 | Idtac]; Auto.
Case (zeroP_dec (plusTerm a a0)); Intros Z0.
Apply eqp_trans with y := (multm_lm a1 (pluspf l l0)); Auto.
Apply eqp_sym; Auto.
Apply multm_lm_comp; Auto.
Cut (zeroP (plusTerm (multTerm a1 a) (multTerm a1 a0))); [Intros Od1 | Idtac];
 Auto.
Apply eqp_trans with y := (pluspf (multm_lm a1 l) (multm_lm a1 l0)); Auto.
Apply H'; Simpl; Auto.
Red; Red; Simpl; Rewrite <- plus_n_Sm; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_imp_canonical with a := a0; Auto.
Apply zeroP_comp_eqTerm with a := (multTerm a1 (plusTerm a a0)).
Apply zeroP_multTerm_r; Auto.
Apply eqTerm_sym; Auto.
Cut ~ (zeroP (plusTerm (multTerm a1 a) (multTerm a1 a0))); [Intros Od1 | Idtac];
 Auto.
Apply eqp_trans with y := (multm_lm a1 (pX (plusTerm a a0) (pluspf l l0))); Auto.
Apply multm_lm_comp; Auto.
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pX
             (plusTerm (multTerm a1 a) (multTerm a1 a0))
             (pluspf (multm_lm a1 l) (multm_lm a1 l0))); Auto.
Simpl.
Apply eqpP1; Auto.
Apply H'; Simpl; Auto.
Red; Red; Simpl; Rewrite <- plus_n_Sm; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_imp_canonical with a := a0; Auto.
Apply nzeroP_comp_eqTerm with a := (multTerm a1 (plusTerm a a0)); Auto.
Qed.
Hints Resolve multlm_lm_dist_pluspf.
End SMultm_lm.

