
(****************************************************************************
                                                                           
          Buchberger : one step divisison  terms                           
                                                                           
          Laurent Thery April 97 (revised Mai 98)                          
                                                                           
****************************************************************************)
Section Spminus.
Load "shyp_spminus".
Load "mod_sminus".

Mutual Inductive divP: Term -> Term ->Prop :=
     divTerm_def:
       (a, b:Term) ~ (zeroP a) ->
       (nZb:~ (zeroP b)) (eqTerm a (multTerm (divTerm a b nZb) b)) ->(divP a b) .

Theorem divP_inv1: (a, b:Term) (divP a b) ->~ (zeroP a).
Intros a b H; Inversion H; Auto.
Qed.

Theorem divP_inv2: (a, b:Term) (divP a b) ->~ (zeroP b).
Intros a b H; Inversion H; Auto.
Qed.

Theorem divP_inv3:
  (a, b:Term) (nZb:~ (zeroP b)) (divP a b) ->
  (eqTerm a (multTerm (divTerm a b nZb) b)).
Intros a b nZb H; Inversion H; Auto.
Apply eqTerm_trans with y := (multTerm (divTerm a b nZb0) b); Auto.
Qed.
Hints Resolve divP_inv1 divP_inv2 divP_inv3.

Theorem divP_plusTerm:
  (a, b, c:Term)
  (divP a c) -> (divP b c) -> (eqM a b) -> ~ (zeroP (plusTerm a b)) ->
  (divP (plusTerm a b) c).
Intros a b c H' H'0; Inversion H'0; Inversion H'.
Intros H'1 H'2; Apply divTerm_def with nZb := nZb0; Auto.
Apply eqTerm_trans
     with y :=
          (plusTerm
             (multTerm (divTerm a c nZb0) c) (multTerm (divTerm b c nZb0) c));
 Auto.
Apply eqTerm_trans
     with y := (multTerm (plusTerm (divTerm a c nZb0) (divTerm b c nZb0)) c);
 Auto.
Qed.
Hints Resolve divP_plusTerm.

Theorem divP_invTerm_l: (a, b:Term) (divP a b) ->(divP (invTerm a) b).
Intros a b H'; Inversion H'; Auto.
Apply divTerm_def with nZb := nZb; Auto.
Apply eqTerm_trans with y := (multTerm (invTerm (divTerm a b nZb)) b); Auto.
Apply eqTerm_trans with y := (invTerm (multTerm (divTerm a b nZb) b)); Auto.
Qed.
Hints Resolve divP_invTerm_l.

Theorem divP_invTerm_r: (a, b:Term) (divP a b) ->(divP a (invTerm b)).
Intros a b H'; Inversion H'.
Cut ~ (zeroP (invTerm b)); [Intros nZib | Auto].
Apply divTerm_def with nZb := nZib; Auto.
Apply eqTerm_trans with y := (multTerm (invTerm (divTerm a b nZb)) (invTerm b));
 Auto.
Apply eqTerm_trans with y := (invTerm (multTerm (divTerm a b nZb) (invTerm b)));
 Auto.
Apply eqTerm_trans with y := (invTerm (invTerm (multTerm (divTerm a b nZb) b)));
 Auto.
Apply eqTerm_trans with y := (multTerm (divTerm a b nZb) b); Auto.
Qed.
Hints Resolve divP_invTerm_r.

Theorem divP_is_not_order: (a, b:Term) (divP a b) ->~ (order a b).
Intros a b H'; Inversion H'.
Case (order_dec T1 (divTerm a b nZb));
 [Intros tmp; Case tmp; Clear tmp | Idtac]; Intros H'2; Auto.
Apply order_not_order; Auto.
Apply order_eqM_l with a := (multTerm (divTerm a b nZb) b); Auto.
Apply order_eqM_r with a := (multTerm T1 b); Auto.
Elim (T1_is_min (divTerm a b nZb)); Auto.
Apply order_not_eqM.
Apply eqM_trans with y := (multTerm (divTerm a b nZb) b); Auto.
Apply eqM_trans with y := (multTerm T1 b); Auto.
Apply eqM_sym; Auto.
Qed.
Hints Resolve divP_is_not_order.

Theorem divTerm_multTerml:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->(divP (multTerm a b) a).
Intros a b nZa nZb.
Apply divTerm_def with nZb := nZa; Auto.
Apply eqTerm_trans with y := (multTerm (multTerm (divTerm a a nZa) b) a); Auto.
Apply eqTerm_trans with y := (multTerm (multTerm T1 b) a); Auto.
Apply eqTerm_trans with y := (multTerm b a); Auto.
Qed.

Theorem divTerm_multTermr:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->(divP (multTerm a b) b).
Intros a b nZa nZb.
Apply divTerm_def with nZb := nZb; Auto.
Apply eqTerm_trans with y := (multTerm (multTerm a (divTerm b b nZb)) b); Auto.
Qed.
Hints Resolve divTerm_multTermr divTerm_multTerml.

Theorem divP_trans: (transitive Term divP).
Red; Intros a b c H' H'0.
Cut ~ (zeroP c); [Intros nZc | Auto].
Apply divTerm_def with nZb := nZc; Auto.
Apply divP_inv1 with b := b; Auto.
2:Apply divP_inv2 with a := b; Auto.
Cut ~ (zeroP b); [Intros nZb | Auto].
Apply eqTerm_trans
     with y := (multTerm (multTerm (divTerm a b nZb) (divTerm b c nZc)) c); Auto.
Apply eqTerm_trans
     with y := (multTerm (divTerm a b nZb) (multTerm (divTerm b c nZc) c)); Auto.
Apply eqTerm_trans with y := (multTerm (divTerm a b nZb) b); Auto.
Apply eqTerm_multTerm_comp; Auto.
Inversion H'0; Inversion H'.
Apply eqTerm_trans with y := (divTerm (multTerm (divTerm a b nZb) b) c nZc);
 Auto.
Apply divP_inv2 with a := a; Auto.
Qed.
Hints Resolve divP_trans.

Local eqM_divTerm:
  (a, b, c, d:Term)
  (nZc:~ (zeroP c)) (nZd:~ (zeroP d)) (eqM a b) -> (eqM c d) ->
  (eqM (divTerm a c nZc) (divTerm b d nZd)) := eq_divTerm.
Hints Resolve eqM_divTerm.

Theorem divP_nZero:
  (a, b:Term) (nZb:~ (zeroP b)) (divP a b) ->~ (zeroP (divTerm a b nZb)).
Intros a b nZb H'; Inversion H'; Auto.
Red; Intros H'0; Apply H.
Apply zeroP_comp_eqTerm with a := (multTerm (divTerm a b nZb) b); Auto.
Qed.
Hints Resolve divP_nZero.

Theorem divP_eqTerm_comp:
  (a, b, c:Term) (divP a c) -> (eqTerm a b) ->(divP b c).
Intros a b c H' H'0.
Cut ~ (zeroP c); [Intros nZc | Auto].
Apply divTerm_def with nZb := nZc; Auto.
Apply nzeroP_comp_eqTerm with a := a; Auto.
Apply divP_inv1 with b := c; Auto.
2:Apply divP_inv2 with a := a; Auto.
Apply eqTerm_trans with y := (multTerm (divTerm a c nZc) c); Auto.
Apply eqTerm_trans with y := a; Auto.
Qed.

Theorem divP_order_comp:
  (a, b:Term) (p:(list Term)) (canonical (pX b p)) -> (divP a b) ->
  (canonical (pX a p)).
Intros a b p; Case p.
Intros H' H'0.
Apply canonical1; Auto.
Apply divP_inv1 with b := b; Auto.
Intros a0 l H' H'0; Apply canonical_cons; Auto.
Case (order_dec a b); [Intros tmp; Case tmp; Clear tmp | Idtac]; Intros H'2;
 Auto.
Elim (divP_is_not_order a b); Auto.
Apply order_trans with y := b; Auto.
Apply canonical_imp_order with l := l; Auto.
Apply order_eqM_l with a := b; Auto.
Apply eqM_sym; Auto.
Apply canonical_imp_order with l := l; Auto.
Apply divP_inv1 with b := b; Auto.
Apply canonical_imp_canonical with a := b; Auto.
Qed.

Definition spminusf :=
   [a, b:Term] [nZb:~ (zeroP b)] [p, q:(list Term)]
   (minuspf p (multm_lm (divTerm a b nZb) q)).

Theorem canonical_spminusf:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term)) (canonical p) -> (canonical q) -> (divP a b) ->
  (canonical (spminusf a b nZb p q)).
Unfold spminusf.
Intros a b nZb p q H' H'0 H'1; Apply canonical_minuspf; Auto.
Qed.
Hints Resolve canonical_spminusf.

Theorem spminusf_extend:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term))
  (divP a b) -> (canonical (pX a p)) -> (canonical (pX b q)) ->
  (eqP (spminusf a b nZb p q) (spminusf a b nZb (pX a p) (pX b q))).
Intros a b nZb p q H'; Unfold spminusf; Simpl; Inversion H'.
Intros H'0 H'1;
 Apply eqp_trans
      with y := (minuspf (pX a p) (pX a (multm_lm (divTerm a b nZb) q))); Auto.
Apply eqp_sym; Auto.
Apply minuspf_comp; Auto.
Apply canonical_pX_eqM with a := (multTerm (divTerm a b nZb) b); Auto.
Qed.

Theorem canonical_spminusf_full:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term))
  (canonical (pX a p)) -> (canonical (pX b q)) -> (divP a b) ->
  (canonical (spminusf a b nZb p q)).
Intros a b nZb p q H' H'0 H'1; Apply canonical_spminusf; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_imp_canonical with a := b; Auto.
Qed.
Hints Resolve canonical_spminusf_full.

Theorem canonical_spminus_full_t:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term))
  (canonical (pX a p)) -> (canonical (pX b q)) -> (divP a b) ->
  (canonical (pX a (spminusf a b nZb p q))).
Unfold spminusf.
Intros a b nZb p q H' H'0 H'1; Try Assumption.
Inversion H'1.
Apply order_minuspf; Auto.
Apply canonical_pX_eqM with a := (multTerm (divTerm a b nZb) b); Auto.
Qed.
Hints Resolve canonical_spminus_full_t.

Theorem spminusf_pluspf:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (pluspf p q) r) (pluspf (spminusf a b nZb p r) q)).
Intros a b nZb p q r H' H'0 H'1 H'2; Unfold spminusf.
Inversion H'2.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p q) (multm_lm (invTerm T1) (multm_lm (divTerm a b nZb) r)));
 Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p (pluspf q (multm_lm (invTerm T1) (multm_lm (divTerm a b nZb) r))));
 Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p (pluspf (multm_lm (invTerm T1) (multm_lm (divTerm a b nZb) r)) q));
 Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p (multm_lm (invTerm T1) (multm_lm (divTerm a b nZb) r))) q);
 Apply eqp_sym; Auto.
Apply eqp_pluspf_com; Auto.
Qed.
Hints Resolve spminusf_pluspf.

Theorem eqptail_spminusf_com:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (eqP p q) -> (divP a b) ->
  (eqP (spminusf a b nZb p r) (spminusf a b nZb q r)).
Unfold spminusf; Auto.
Qed.

Theorem eqTerm_spminusf_com:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q:(list Term))
  (canonical p) -> (canonical q) -> (eqTerm a b) -> (divP a c) ->
  (eqP (spminusf a c nZc p q) (spminusf b c nZc p q)).
Intros a b c nZc p q H' H'0 H'1 H'2.
Cut (divP b c); [Intros H'3 | Auto].
Unfold spminusf; Auto.
Apply divP_eqTerm_comp with a := a; Auto.
Qed.

Theorem eqp_spminusf_com:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) -> (canonical r) -> (eqP p q) -> (eqTerm a b) -> (divP a c) ->
  (eqP (spminusf a c nZc p r) (spminusf b c nZc q r)).
Intros a b c nZc p q r H' H'0 H'1 H'2 H'3 H'4.
Apply eqp_trans with y := (spminusf b c nZc p r); Auto.
Apply eqTerm_spminusf_com; Auto.
Apply eqptail_spminusf_com; Auto.
Apply divP_eqTerm_comp with a := a; Auto.
Qed.
Hints Resolve eqTerm_spminusf_com eqp_spminusf_com eqp_spminusf_com.

Theorem spminusf_minuspf:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (minuspf p q) r) (minuspf (spminusf a b nZb p r) q)).
Intros a b nZb p q r H' H'0 H'1 H'2; Try Assumption.
Apply eqp_trans
     with y := (spminusf a b nZb (pluspf p (multm_lm (invTerm T1) q)) r); Auto.
Apply eqp_trans
     with y := (pluspf (spminusf a b nZb p r) (multm_lm (invTerm T1) q)); Auto.
Apply eqp_sym; Auto.
Qed.

Theorem spminusf_plusTerm:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  (divP a c) -> (divP b c) -> (eqM a b) -> ~ (zeroP (plusTerm a b)) ->
  (eqP
     (spminusf (plusTerm a b) c nZc (pluspf p q) r)
     (pluspf (spminusf a c nZc p r) (spminusf b c nZc q r))).
Intros a b c nZc p q r H' H'0 H'1 H'2 H'3 H'4 H'5; Unfold spminusf.
Apply eqp_sym;
 Apply eqp_trans
      with y :=
           (pluspf
              (pluspf p (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
              (pluspf q (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))));
 [Auto | Idtac].
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))
                (pluspf
                   q (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))));
 [Auto | Idtac].
Apply pluspf_assoc; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (pluspf
                   (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)) q)
                (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_sym; Auto 8.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (pluspf
                   q (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
                (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))));
 [Auto 8 | Idtac].
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                q
                (pluspf
                   (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))
                   (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))));
 [Auto 10 | Idtac].
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p q)
             (pluspf
                (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))
                (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_sym; Auto.
Apply pluspf_assoc; Auto.
Apply eqp_sym;
 Apply eqp_trans
      with y :=
           (pluspf
              (pluspf p q)
              (multm_lm
                 (invTerm T1) (multm_lm (divTerm (plusTerm a b) c nZc) r)));
 [Auto | Idtac].
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (multm_lm
             (invTerm T1)
             (pluspf
                (multm_lm (divTerm a c nZc) r) (multm_lm (divTerm b c nZc) r))).
Apply multm_lm_comp; Auto.
Apply eqp_trans
     with y := (multm_lm (plusTerm (divTerm a c nZc) (divTerm b c nZc)) r).
Apply multm_lm_comp.
Apply eq_divTerm_plusTerm; Auto.
Inversion H'2; Auto.
Inversion H'3; Auto.
Auto.
Apply multlm_lm_dist1; Auto.
Inversion H'2; Inversion H'3; Auto.
Apply nzeroP_comp_eqTerm with a := (divTerm (plusTerm a b) c nZc); Auto.
Inversion H'2; Inversion H'3; Auto.
Qed.

Theorem spminusf_multTerm:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q:(list Term))
  (canonical p) -> (canonical q) -> (divP b c) -> ~ (zeroP a) ->
  (eqP
     (spminusf (multTerm a b) c nZc (multm_lm a p) q)
     (multm_lm a (spminusf b c nZc p q))).
Intros a b c nZc p q H' H'0 H'1 H'2; Unfold spminusf.
Apply eqp_trans
     with y :=
          (minuspf (multm_lm a p) (multm_lm (multTerm a (divTerm b c nZc)) q));
 [Auto | Idtac].
Apply minuspf_comp; Auto.
Apply canonical_multlm_lm; Auto.
Apply nzeroP_comp_eqTerm with a := (multTerm a (divTerm b c nZc)); Auto.
Inversion H'1; Auto.
Inversion H'1; Auto.
Apply eqp_trans
     with y :=
          (minuspf (multm_lm a p) (multm_lm a (multm_lm (divTerm b c nZc) q)));
 [Auto | Idtac].
Apply eqp_sym; Auto.
Qed.
Hints Resolve spminusf_plusTerm spminusf_multTerm.

Theorem spminusf_minusTerm_l:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (minuspf p q) r) (minuspf (spminusf a b nZb p r) q)).
Intros a b nZb p q r H' H'0 H'1 H'2.
Apply eqp_trans
     with y := (spminusf a b nZb (pluspf p (multm_lm (invTerm T1) q)) r);
 [Auto | Idtac].
Apply eqp_trans
     with y := (pluspf (spminusf a b nZb p r) (multm_lm (invTerm T1) q));
 [Auto | Idtac].
Apply eqp_sym; Auto.
Qed.

Theorem spminusf_plusTerm_l:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (pluspf p q) r) (pluspf (spminusf a b nZb p r) q)).
Intros a b nZb p q r H' H'0 H'1 H'2.
Apply eqp_trans with y := (pluspf (spminusf a b nZb p r) q); [Auto | Idtac].
Apply eqp_sym; Auto.
Qed.

Theorem spminusf_minusTerm_r:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP
     (spminusf a b nZb (minuspf p q) r)
     (minuspf p (spminusf (invTerm a) b nZb q r))).
Intros a b nZb p q r H' H'0 H'1 H'2; Try Assumption.
Apply eqp_trans
     with y := (spminusf a b nZb (pluspf p (multm_lm (invTerm T1) q)) r);
 [Auto | Idtac].
Apply eqp_trans
     with y := (spminusf a b nZb (pluspf (multm_lm (invTerm T1) q) p) r);
 [Auto | Idtac].
Apply eqp_trans
     with y := (pluspf (spminusf a b nZb (multm_lm (invTerm T1) q) r) p);
 [Auto | Idtac].
Apply eqp_trans
     with y := (pluspf p (spminusf a b nZb (multm_lm (invTerm T1) q) r));
 [Auto | Idtac].
Apply eqp_sym;
 Apply eqp_trans
      with y :=
           (pluspf p (multm_lm (invTerm T1) (spminusf (invTerm a) b nZb q r)));
 [Auto | Idtac].
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (spminusf
             (multTerm (invTerm T1) (invTerm a)) b nZb (multm_lm (invTerm T1) q)
             r); [Apply eqp_sym; Auto | Idtac].
Apply eqp_trans
     with y :=
          (spminusf
             (invTerm (multTerm T1 (invTerm a))) b nZb (multm_lm (invTerm T1) q)
             r); [Auto | Idtac].
Apply eqTerm_spminusf_com; Auto.
Apply divP_trans with y := (invTerm a); Auto.
Apply divTerm_multTermr; Auto.
Apply nZero_invTerm_nZero; Auto.
Apply divP_inv1 with b := b; Auto.
Apply eqp_trans
     with y :=
          (spminusf (invTerm (invTerm a)) b nZb (multm_lm (invTerm T1) q) r);
 Auto.
Apply eqTerm_spminusf_com; Auto.
Apply divP_trans with y := (invTerm a); Auto.
Apply divP_invTerm_l; Auto.
Apply divTerm_multTermr; Auto.
Apply nZero_invTerm_nZero; Auto.
Apply divP_inv1 with b := b; Auto.
Qed.
Hints Resolve spminusf_minusTerm_r.

Theorem spminusf_plusTerm_r:
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q, r:(list Term))
  (canonical p) -> (canonical q) -> (canonical r) -> (divP a b) ->
  (eqP (spminusf a b nZb (pluspf p q) r) (pluspf p (spminusf a b nZb q r))).
Intros a b nZb p q r H' H'0 H'1 H'2; Try Assumption.
Apply eqp_trans with y := (spminusf a b nZb (pluspf q p) r); [Auto | Idtac].
Apply eqp_trans with y := (pluspf (spminusf a b nZb q r) p); [Auto | Idtac].
Apply eqp_sym; Auto.
Qed.
Hints Resolve spminusf_plusTerm_r.

Theorem divP_minusTerm_comp:
  (a, b, c:Term)
  (divP a c) -> (divP b c) -> (eqM a b) -> ~ (zeroP (minusTerm a b)) ->
  (divP (minusTerm a b) c).
Intros a b c H' H'0 H'1 H'2.
Apply divP_eqTerm_comp with a := (plusTerm a (invTerm b)); Auto.
Apply divP_plusTerm; Auto.
Apply nzeroP_comp_eqTerm with a := (minusTerm a b); Auto.
Qed.
Hints Resolve divP_minusTerm_comp.

Theorem spminusf_minusTerm:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  (divP a c) -> (divP b c) -> (eqM a b) -> ~ (zeroP (minusTerm a b)) ->
  (eqP
     (spminusf (minusTerm a b) c nZc (minuspf p q) r)
     (minuspf (spminusf a c nZc p r) (spminusf b c nZc q r))).
Intros a b c nZc p q r H' H'0 H'1 H'2 H'3 H'4 H'5.
Apply eqp_trans
     with y := (spminusf (plusTerm a (invTerm b)) c nZc (minuspf p q) r);
 [Auto | Idtac].
Apply eqp_trans
     with y :=
          (spminusf
             (plusTerm a (invTerm b)) c nZc (pluspf p (multm_lm (invTerm T1) q))
             r); [Auto | Idtac].
Apply eqp_spminusf_com; Auto.
Apply divP_eqTerm_comp with a := (minusTerm a b); Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (spminusf a c nZc p r)
             (spminusf (invTerm b) c nZc (multm_lm (invTerm T1) q) r));
 [Auto | Idtac].
Apply spminusf_plusTerm; Auto.
Apply nzeroP_comp_eqTerm with a := (minusTerm a b); Auto.
Apply eqp_sym;
 Apply eqp_trans
      with y :=
           (pluspf
              (spminusf a c nZc p r)
              (multm_lm (invTerm T1) (spminusf b c nZc q r))); [Auto | Idtac].
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (spminusf (multTerm (invTerm T1) b) c nZc (multm_lm (invTerm T1) q) r);
 [Auto | Idtac].
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (spminusf (invTerm (multTerm T1 b)) c nZc (multm_lm (invTerm T1) q) r);
 [Auto | Auto].
Apply eqTerm_spminusf_com; Auto.
Apply divP_trans with y := b; Auto.
Apply divTerm_multTermr; Auto.
Apply divP_inv1 with b := c; Auto.
Apply eqTerm_spminusf_com; Auto.
Apply divP_trans with y := b; Auto.
Apply divP_invTerm_l; Auto.
Apply divTerm_multTermr; Auto.
Apply divP_inv1 with b := c; Auto.
Qed.

Theorem spminusf_minusTerm_z:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  (divP a c) -> (divP b c) -> (eqM a b) -> (zeroP (minusTerm a b)) ->
  (eqP (minuspf p q) (minuspf (spminusf a c nZc p r) (spminusf b c nZc q r))).
Intros a b c nZc p q r H' H'0 H'1 H'2 H'3 H'4 H'5.
Apply eqp_sym;
 Apply eqp_trans
      with y :=
           (minuspf
              (minuspf p (multm_lm (divTerm a c nZc) r))
              (minuspf q (multm_lm (divTerm b c nZc) r))); [Auto | Idtac].
Apply eqp_trans
     with y :=
          (minuspf
             (pluspf p (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
             (pluspf q (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply minuspf_comp; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
             (multm_lm
                (invTerm T1)
                (pluspf
                   q (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))));
 [Auto | Idtac].
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
             (pluspf
                (multm_lm (invTerm T1) q)
                (multm_lm
                   (invTerm T1)
                   (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))));
 [Auto | Idtac].
Apply eqp_pluspf_com; Auto.
Apply canonical_multlm_lm; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))
                (pluspf
                   (multm_lm (invTerm T1) q)
                   (multm_lm
                      (invTerm T1)
                      (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))))).
Apply pluspf_assoc; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (pluspf
                   (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))
                   (multm_lm (invTerm T1) q))
                (multm_lm
                   (invTerm T1)
                   (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply canonical_pluspf; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_sym; Apply pluspf_assoc; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (pluspf
                   (multm_lm (invTerm T1) q)
                   (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
                (multm_lm
                   (invTerm T1)
                   (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf
                p
                (pluspf
                   (multm_lm (invTerm T1) q)
                   (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))))
             (multm_lm
                (invTerm T1)
                (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_sym; Apply pluspf_assoc; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf
                (minuspf p q)
                (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
             (multm_lm
                (invTerm T1)
                (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p (multm_lm (invTerm T1) q))
             (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))).
Apply eqp_sym; Apply pluspf_assoc; Auto.
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf p q)
             (pluspf
                (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))
                (multm_lm
                   (invTerm T1)
                   (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))))).
Apply pluspf_assoc; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf p q)
             (multm_lm
                (invTerm T1)
                (pluspf
                   (multm_lm (divTerm a c nZc) r)
                   (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply canonical_multlm_lm; Auto.
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (minuspf p q)
             (multm_lm
                (invTerm T1)
                (minuspf
                   (multm_lm (divTerm a c nZc) r) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_pluspf_com; Auto.
Apply canonical_multlm_lm; Auto.
Apply eqp_sym; Auto.
Apply eqp_trans with y := (pluspf (minuspf p q) (multm_lm (invTerm T1) pO)).
Apply eqp_pluspf_com; Auto.
Apply multm_lm_comp; Auto.
Apply eqp_sym; Apply multlm_lm_pO; Auto.
Cut (eqTerm (divTerm a c nZc) (divTerm b c nZc)); [Intros eqTerm0 | Idtac].
Inversion H'2; Inversion H'3; Auto.
Inversion H'2; Inversion H'3; Auto.
Apply zeroP_comp_eqTerm
     with a := (plusTerm (divTerm a c nZc) (invTerm (divTerm b c nZc))); Auto.
Apply zeroP_comp_eqTerm
     with a := (plusTerm (divTerm a c nZc) (invTerm (divTerm a c nZc))); Auto.
Apply eqTerm_plusTerm_comp; Auto.
Change (eqM (divTerm a c nZc) (invTerm (divTerm b c nZc))).
Apply eqM_trans with y := (divTerm b c nZc); Auto.
Inversion H'2; Inversion H'3; Auto.
Inversion H'2; Inversion H'3; Auto.
Qed.
Hints Resolve spminusf_minusTerm.

Theorem spminusf_plusTerm_z:
  (a, b, c:Term)
  (nZc:~ (zeroP c))
  (p, q, r:(list Term))
  (canonical p) ->
  (canonical q) ->
  (canonical r) ->
  (divP a c) -> (divP b c) -> (eqM a b) -> (zeroP (plusTerm a b)) ->
  (eqP (pluspf p q) (pluspf (spminusf a c nZc p r) (spminusf b c nZc q r))).
Intros a b c nZc p q r H' H'0 H'1 H'2 H'3 H'4 H'5.
Apply eqp_sym;
 Apply eqp_trans
      with y :=
           (pluspf
              (minuspf p (multm_lm (divTerm a c nZc) r))
              (minuspf q (multm_lm (divTerm b c nZc) r))); [Auto | Idtac].
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
             (pluspf q (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))
                (pluspf
                   q (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))))).
Apply pluspf_assoc; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (pluspf
                   (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)) q)
                (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_sym; Apply pluspf_assoc; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             p
             (pluspf
                (pluspf
                   q (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
                (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf
                p
                (pluspf
                   q (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))))
             (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))).
Apply eqp_sym; Apply pluspf_assoc; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf
                (pluspf p q)
                (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r)))
             (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r))).
Apply eqp_pluspf_com; Auto.
Apply canonical_pluspf; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p q) (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))).
Apply eqp_sym; Apply pluspf_assoc; Auto.
Apply eqp_sym; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p q)
             (pluspf
                (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))
                (multm_lm (invTerm T1) (multm_lm (divTerm b c nZc) r)))).
Apply pluspf_assoc; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (pluspf p q)
             (multm_lm
                (invTerm T1)
                (pluspf
                   (multm_lm (divTerm a c nZc) r) (multm_lm (divTerm b c nZc) r)))).
Apply eqp_pluspf_com; Auto.
Apply eqp_sym; Auto.
Apply eqp_trans with y := (pluspf (pluspf p q) (multm_lm (invTerm T1) pO)).
Apply eqp_pluspf_com; Auto.
Apply multm_lm_comp; Auto.
Cut (eqTerm (divTerm b c nZc) (invTerm (divTerm a c nZc)));
 [Intros eqTerm0 | Idtac].
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (divTerm a c nZc) r)
             (multm_lm (invTerm (divTerm a c nZc)) r)).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (divTerm a c nZc) r)
             (multm_lm (invTerm (multTerm T1 (divTerm a c nZc))) r)).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (divTerm a c nZc) r)
             (multm_lm (multTerm (invTerm T1) (divTerm a c nZc)) r)).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (pluspf
             (multm_lm (divTerm a c nZc) r)
             (multm_lm (invTerm T1) (multm_lm (divTerm a c nZc) r))).
Apply eqp_pluspf_com; Auto.
Apply eqp_trans
     with y :=
          (minuspf
             (multm_lm (divTerm a c nZc) r) (multm_lm (divTerm a c nZc) r));
 Auto.
Apply eqp_sym; Auto.
Inversion H'2; Inversion H'3; Auto.
Apply eqTerm_trans with y := (divTerm (invTerm a) c nZc); Auto.
Apply eqTerm_divTerm_com; Auto.
Apply zerop_is_eqTerm; Auto.
Apply eqM_trans with y := a; Auto.
Apply eqM_sym; Auto.
Apply zeroP_comp_eqTerm with a := (plusTerm b (invTerm (invTerm a))); Auto.
Apply zeroP_comp_eqTerm with a := (plusTerm b a); Auto.
Apply zeroP_comp_eqTerm with a := (plusTerm a b); Auto.
Apply eqTerm_plusTerm_comp; Auto.
Rewrite <- invTerm_eq; Rewrite <- invTerm_eq; Auto.
Simpl; Auto.
Qed.
Hints Resolve spminusf_minusTerm.

Theorem ltP_divP_pX:
  (a, b:Term)
  (p, q:(list Term))
  (canonical (pX a p)) -> (canonical (pX b q)) -> (divP a b) ->(ltP q (pX a p)).
Intros a b p; Case p; Auto.
Intros q H' H'0 H'1; Try Assumption.
Apply canonical_pX_ltP; Apply divP_order_comp with b := b; Auto.
Intros a0 l q H' H'0 H'1; Apply ltP_trans with y := (pX a pO); Auto.
Apply canonical_pX_ltP; Apply divP_order_comp with b := b; Auto.
Qed.
End Spminus.
