
(****************************************************************************
                                                                           
          Buchberger : reduction                                           
                                                                           
          Laurent Thery Mai 97 (revised Mai 98)                            
                                                                           
****************************************************************************)
Section Sreduce.
Load "shyp_sreduce".
Load "mod_spminus".

Theorem divP_dec:
  (a, b:Term) ~ (zeroP a) -> ~ (zeroP b) ->{(divP a b)}+{~ (divP a b)}.
Intros a b nZa nZb.
Elim (divTerm_dec a b nZa nZb); Auto.
Intros H'1; Left; Apply divTerm_def with nZb := nZb; Auto.
Qed.

Theorem divP_eqM:
  (a, b, c:Term) (eqM a b) -> ~ (zeroP b) -> (divP a c) ->(divP b c).
Intros a b c H' nZb H'1; Inversion H'1.
Apply divTerm_def with nZb := nZb0; Auto.
Apply divTerm_eq with a := a; Auto.
Qed.

Mutual Inductive inPolySet: (list Term) -> (list poly) ->Prop :=
     incons:
       (a:Term) (p:(list Term)) (H:(canonical (pX a p))) (P:(list poly))
       (inPolySet
          (pX a p)
          (cons (exist (list Term) [l0:(list Term)](canonical l0) (pX a p) H) P))
    | inskip:
        (a:poly) (p:(list Term)) (P:(list poly)) (inPolySet p P) ->
        (inPolySet p (cons a P)) .
Hints Resolve incons inskip.

Lemma inPolySet_imp_canonical:
  (p:(list Term)) (L:(list poly)) (inPolySet p L) ->(canonical p).
Intros p L H'; Elim H'; Auto.
Qed.
Hints Resolve inPolySet_imp_canonical.

Lemma inPolySet_is_pX:
  (p:(list Term)) (L:(list poly)) (inPolySet p L) ->
  (Ex [a:Term] (Ex [q:(list Term)] p=(pX a q))).
Intros p L H'; Elim H'; Auto.
Intros a p0 H'0 H'1; Exists a; Exists p0; Auto.
Qed.

Mutual Inductive pickinSetp: Term -> (list Term) -> (list poly) ->Prop :=
     pickinSeteqp:
       (a, b:Term)
       (p:(list Term)) (H:(canonical (pX b p))) (P:(list poly)) (divP a b) ->
       (pickinSetp
          a (pX b p)
          (cons (exist (list Term) [l0:(list Term)](canonical l0) (pX b p) H) P))
    | pickinSetskip:
        (a, b:Term)
        (p:(list Term)) (q:poly) (P:(list poly)) (pickinSetp a p P) ->
        (pickinSetp a p (cons q P)) .
Hints Resolve pickinSeteqp pickinSetskip.

Lemma pickin_is_pX:
  (a:Term) (p:(list Term)) (Q:(list poly)) (pickinSetp a p Q) ->
  (Ex [b:Term] (Ex [q:(list Term)] p=(pX b q))).
Intros a p Q H'; Elim H'; Auto.
Intros a0 b p0 H'0 H'1 H'2; Exists b; Exists p0; Auto.
Qed.

Mutual Inductive reduce[Q:(list poly)]: (list Term) -> (list Term) ->Prop :=
     reducetop:
       (a, b:Term)
       (nZb:~ (zeroP b))
       (p, q, r:(list Term))
       (inPolySet (pX b q) Q) -> (divP a b) -> (eqP (spminusf a b nZb p q) r) ->
       (reduce Q (pX a p) r)
    | reduceskip:
        (a, b:Term) (p, q:(list Term)) (reduce Q p q) -> (eqTerm a b) ->
        (reduce Q (pX a p) (pX b q)) .
Hints Resolve reduceskip.

Mutual Inductive reducehead[Q:(list poly)]: (list Term) -> (list Term) ->Prop :=
     reduceheadO:
       (a, b:Term)
       (nZb:~ (zeroP b)) (p, q:(list Term)) (pickinSetp a (pX b q) Q) ->
       (reducehead Q (pX a p) (spminusf a b nZb p q)) .
Hints Resolve reduceheadO.

Lemma pick_inv_in:
  (Q:(list poly)) (a:Term) (p:(list Term)) (pickinSetp a p Q) ->(inPolySet p Q).
Intros Q a p H'; Elim H'; Auto.
Qed.

Lemma pick_inv_eqM_lem:
  (Q:(list poly)) (a:Term) (p:(list Term)) (pickinSetp a p Q) ->
  (b:Term) (q:(list Term)) p=(pX b q) ->(divP a b).
Intros Q a p H'; Elim H'; Auto.
Intros a0 b p0 H'0 H'1 H'2 b0 q H'3; Injection H'3; Auto.
Intros H'4 H'5; Rewrite <- H'5; Auto.
Qed.

Lemma pick_inv_eqM:
  (Q:(list poly)) (a, b:Term) (p, q:(list Term)) (pickinSetp a (pX b q) Q) ->
  (divP a b).
Intros Q a b H' q H'0.
Apply pick_inv_eqM_lem with Q := Q p := (pX b q) q := q; Auto.
Qed.

Theorem reducehead_imp_reduce:
  (Q:(list poly)) (p, q:(list Term)) (reducehead Q p q) ->(reduce Q p q).
Intros Q p q H'; Elim H'; Auto.
Intros a b nZb p0 q0 H'0.
Apply reducetop with b := b nZb := nZb q := q0; Auto.
Apply pick_inv_in with a := a; Auto.
Apply pick_inv_eqM with Q := Q q := q0; Auto.
Qed.

Theorem ltP_reduce:
  (Q:(list poly)) (p, q:(list Term)) (reduce Q p q) -> (canonical p) ->(ltP q p).
Intros Q p q H'; Elim H'; Auto.
2:Intros a b p0 q0 H'0 H'1 H'2 H'3; Apply ltP_tl; Auto.
2:Apply H'1; Apply canonical_imp_canonical with a := a; Auto.
Intros a b nZb p0; Case p0; Auto.
Intros q0 r H'0 H'1 H'2 H'3.
Apply canonical_pX_ltP; Auto.
Apply eqp_imp_canonical with p := (pX a (spminusf a b nZb (nil Term) q0)); Auto.
Apply canonical_spminusf_full_t; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Intros t l q0 r H'0 H'1 H'2 H'3; Apply ltP_trans with y := (pX a pO); Auto.
Apply canonical_pX_ltP; Auto.
Apply eqp_imp_canonical with p := (pX a (spminusf a b nZb (cons t l) q0)); Auto.
Apply canonical_spminusf_full_t; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Qed.

Theorem canonical_reduce:
  (Q:(list poly)) (p, q:(list Term)) (reduce Q p q) -> (canonical p) ->
  (canonical q).
Intros Q p q H'; Elim H'; Auto.
Intros a b nZb p0 q0 r H'0 H'1 H'2 H'3.
Apply eqp_imp_canonical with p := (spminusf a b nZb p0 q0); Auto.
Apply canonical_spminusf; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_imp_canonical with a := b; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Intros a b p0 q0 H'0 H'1 H'2 H'3.
Apply eqp_imp_canonical with p := (pX a q0); Auto.
Cut (canonical p0); [Intros C1 | Apply canonical_imp_canonical with a := a];
 Auto.
Apply ltP_pX_canonical; Auto.
Apply canonical_nzeroP with p := p0; Auto.
Apply ltP_trans with y := p0.
Apply ltP_reduce with Q := Q; Auto.
Apply canonical_pX_ltP; Auto.
Qed.

Theorem reduce_eqp_com:
  (Q:(list poly))
  (p, q, r, s:(list Term))
  (reduce Q p q) -> (canonical p) -> (eqP p r) -> (eqP q s) ->(reduce Q r s).
Intros Q p q r s H'; Generalize r s; Clear r s; Elim H'; Auto.
Intros a b nZb p0 q0 r H'0 H'1 H'2 r0 s H'3 H'4 H'5.
Inversion H'4.
Simpl in H3; Simpl in H1; Simpl in H0; Simpl in H.
Apply reducetop with b := b nZb := nZb q := q0; Auto.
Apply divP_eqTerm_comp with a := a; Auto.
Apply eqp_trans with y := (spminusf a b nZb p0 q0); Auto.
Cut (canonical p0);
 [Intros Cp0 | Apply canonical_imp_canonical with a := a; Auto].
Apply eqp_spminusf_com; Auto.
Apply eqp_imp_canonical with p := p0; Auto.
Apply canonical_imp_canonical with a := b; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Apply eqp_sym; Auto.
Apply divP_eqTerm_comp with a := a; Auto.
Apply eqp_trans with y := r; Auto.
Intros a b p0 q0 H'0 H'1 H'2 r s H'3 H'4 H'5.
Inversion H'4.
Simpl in H3; Simpl in H1; Simpl in H0; Simpl in H.
Inversion H'5.
Simpl in H8; Simpl in H6; Simpl in H5; Simpl in H4.
Apply reduceskip; Auto.
Apply H'1; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply eqTerm_trans with y := a; Auto.
Apply eqTerm_trans with y := b; Auto.
Qed.

Mutual Inductive reduceplus[Q:(list poly)]: (list Term) -> (list Term) ->Prop :=
     Rstar_0: (x, y:(list Term)) (eqP x y) ->(reduceplus Q x y)
    | Rstar_n:
        (x, y, z:(list Term)) (reduce Q x y) -> (reduceplus Q y z) ->
        (reduceplus Q x z) .
Hints Resolve Rstar_0.

Theorem reduceplus_eqp_com:
  (Q:(list poly))
  (p, q, r, s:(list Term))
  (reduceplus Q p q) -> (canonical p) -> (eqP p r) -> (eqP q s) ->
  (reduceplus Q r s).
Intros Q p q r s H'; Generalize r s; Clear r s; Elim H'; Auto.
Intros x y H'0 r s H'1 H'2 H'3.
Apply Rstar_0; Auto.
Apply eqp_trans with y := y; Auto.
Apply eqp_trans with y := x; Auto.
Apply eqp_sym; Auto.
Intros x y z H'0 H'1 H'2 r s H'3 H'4 H'5.
Apply Rstar_n with y := y; Auto.
Apply reduce_eqp_com with p := x q := y; Auto.
Apply H'2; Auto.
Apply canonical_reduce with Q := Q p := x; Auto.
Qed.

Theorem reduceplus_trans:
  (Q:(list poly))
  (x, y, z:(list Term))
  (reduceplus Q x y) -> (canonical x) -> (reduceplus Q y z) ->(reduceplus Q x z).
Intros Q x y z H'; Elim H'; Auto.
Intros x0 y0 H'0 H'1 H'2.
Apply reduceplus_eqp_com with p := y0 q := z; Auto.
Apply eqp_imp_canonical with p := x0; Auto.
Apply eqp_sym; Auto.
Intros x0 y0 z0 H'0 H'1 H'2 H'3 H'4.
Apply Rstar_n with y := y0; Auto.
Apply H'2; Auto.
Apply canonical_reduce with Q := Q p := x0; Auto.
Qed.

Theorem canonical_reduceplus:
  (Q:(list poly)) (p, q:(list Term)) (reduceplus Q p q) -> (canonical p) ->
  (canonical q).
Intros Q p q H'; Elim H'; Auto.
Intros x y H'0 H'1.
Apply eqp_imp_canonical with p := x; Auto.
Intros x y z H'0 H'1 H'2 H'3.
Apply H'2; Auto.
Apply canonical_reduce with Q := Q p := x; Auto.
Qed.

Theorem order_reduceplus:
  (Q:(list poly)) (p, q:(list Term)) (reduceplus Q p q) -> (canonical p) ->
  (ltP q p) \/ (eqP p q).
Intros Q p q H'; Elim H'; Auto.
Intros x y z H'0 H'1 H'2 H'3; Left; Try Assumption.
Elim H'2;
 [Intros H'5; Try Exact H'5; Clear H'2 | Intros H'5; Clear H'2 | Clear H'2].
Apply ltP_trans with y := y; Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply ltp_eqp_comp with p := y q := x; Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply canonical_reduce with Q := Q p := x; Auto.
Apply canonical_reduce with Q := Q p := x; Auto.
Qed.

Theorem reduceplus_skip:
  (Q:(list poly))
  (a, b:Term)
  (p, q:(list Term))
  (reduceplus Q p q) -> (canonical (pX a p)) -> (eqTerm a b) ->
  (reduceplus Q (pX a p) (pX b q)).
Intros Q a b p q H'; Generalize a b; Clear a b; Elim H'; Auto.
Intros x y z H'0 H'1 H'2 a b H'3 H'4.
Apply Rstar_n with y := (pX b y); Auto.
Apply H'2; Auto.
Apply ltP_pX_canonical; Auto.
Apply canonical_reduce with Q := Q p := x; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_nzeroP with p := x; Auto.
Apply eqp_imp_canonical with p := (pX a x); Auto.
Apply ltP_trans with y := x.
Apply ltP_reduce with Q := Q; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_pX_ltP; Auto.
Apply eqp_imp_canonical with p := (pX a x); Auto.
Qed.
Hints Resolve reduceplus_skip.

Definition irreducible :=
   [Q:(list poly)] [p:(list Term)] (q:(list Term))~ (reduce Q p q).

Theorem irreducible_eqp_com:
  (Q:(list poly))
  (p, q:(list Term)) (irreducible Q p) -> (canonical p) -> (eqP p q) ->
  (irreducible Q q).
Unfold irreducible.
Intros Q p q H' H'0 H'1 q0; Red; Intros H'2.
Specialize 1 H' with q := q0; Intros H'3; Apply H'3; Clear H'3; Auto.
Apply reduce_eqp_com with p := q q := q0; Auto.
Apply eqp_imp_canonical with p := p; Auto.
Apply eqp_sym; Auto.
Qed.

Mutual Inductive reducestar[Q:(list poly)]: (list Term) -> (list Term) ->Prop :=
     reducestar0:
       (p, q:(list Term)) (reduceplus Q p q) -> (irreducible Q q) ->
       (reducestar Q p q) .

Theorem reducestar_eqp_com:
  (Q:(list poly))
  (p, q, r, s:(list Term))
  (reducestar Q p q) -> (canonical p) -> (eqP p r) -> (eqP q s) ->
  (reducestar Q r s).
Intros Q p q r s H' H'0 H'1 H'2; Inversion H'.
Apply reducestar0; Auto.
Apply reduceplus_eqp_com with p := p q := q; Auto.
Apply irreducible_eqp_com with p := q; Auto.
Apply canonical_reduceplus with Q := Q p := p; Auto.
Qed.

Theorem reduce_multm:
  (Q:(list poly))
  (a:Term) (p, q:(list Term)) (reduce Q p q) -> (canonical p) -> ~ (zeroP a) ->
  (reduce Q (multm_lm a p) (multm_lm a q)).
Intros Q a p q H'; Generalize a; Elim H'; Clear a H'; Auto.
Simpl; Auto.
Intros a b nZb p0 q0 r H' H'0 H'1 a0 H'2 H'3.
Cut ~ (zeroP (multTerm a0 a)); [Intros nZ0 | Idtac]; Auto.
Apply reducetop with b := b nZb := nZb q := q0; Auto.
Apply divTerm_def with nZb := nZb; Auto.
Apply eqTerm_trans with y := (multTerm (multTerm a0 (divTerm a b nZb)) b);
 [Auto | Idtac].
Apply eqTerm_trans with y := (multTerm a0 (multTerm (divTerm a b nZb) b));
 [Auto | Idtac].
Apply eqTerm_sym; Auto.
Apply eqTerm_sym; Auto.
Apply eqp_trans with y := (multm_lm a0 (spminusf a b nZb p0 q0)); Auto.
Apply spminusf_multTerm; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_imp_canonical with a := b; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Apply nzeroP_nonZerop_imp_zeroPl; Auto.
Apply canonical_nzeroP with p := p0; Auto.
Intros a b p0 q0 H' H'0 H'1 a0 H'2 H'3.
Simpl; Apply reduceskip; Auto.
Apply H'0; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Qed.
Hints Resolve reduce_multm.

Lemma canonical_multm_inv:
  (p:(list Term)) (a:Term) ~ (zeroP a) -> (canonical (multm_lm a p)) ->
  (canonical p).
Intros p; Elim p; Simpl; Auto.
Intros a l; Case l; Simpl; Auto.
Intros H' a0 H'0 H'1.
Apply canonical1; Auto.
Red; Intros H'2; Absurd (zeroP (multTerm a0 a)); Auto.
Apply canonical_nzeroP with p := pO; Auto.
Intros a0 l0 H' a1 H'0 H'1.
Apply canonical_cons; Auto.
Case (order_dec a0 a); [Intros temp; Case temp; Clear temp | Idtac]; Intros H;
 Auto.
Absurd (order (multTerm a1 a) (multTerm a1 a0)); Auto.
Apply order_not_order; Auto.
Apply canonical_imp_order with l := (multm_lm a1 l0); Auto.
Absurd (eqM (multTerm a1 a0) (multTerm a1 a)); Auto.
Apply eqM_not_order; Auto.
Apply canonical_imp_order with l := (multm_lm a1 l0); Auto.
Red; Intros H'2; Absurd (zeroP (multTerm a1 a)); Auto.
Apply canonical_nzeroP with p := (multm_lm a1 l0); Auto.
Apply canonical_skip_fst with b := (multTerm a1 a0); Auto.
Apply H' with a := a1; Auto.
Apply canonical_imp_canonical with a := (multTerm a1 a); Auto.
Qed.

Theorem eqTerm_multTerm_imp_eqTerm:
  (a, b, c:Term) ~ (zeroP c) -> (eqTerm (multTerm c a) (multTerm c b)) ->
  (eqTerm a b).
Intros a b c nZc H'0.
Apply eqTerm_trans with y := (divTerm (multTerm c a) c nZc); Auto.
2:Apply eqTerm_trans with y := (divTerm (multTerm c b) c nZc); Auto.
2:Apply eqTerm_trans with y := (multTerm (divTerm c c nZc) b); Auto.
Apply eqTerm_trans with y := (multTerm (divTerm c c nZc) a); Auto.
Qed.

Theorem reduce_multm_inv_lem:
  (Q:(list poly)) (p, q:(list Term)) (reduce Q p q) ->
  (r:(list Term)) (canonical r) -> p=(multm_lm (invTerm T1) r) ->
  (reduce Q r (multm_lm (invTerm T1) q)).
Intros Q p q H'; Elim H'; Clear H'; Auto.
2:Intros a b p0 q0 H' H'0 H'1 r; Case r; Auto.
2:Intros H'1 H'2; Inversion H'2; Auto.
2:Inversion H'3.
2:Intros a0 l H'2 H'3.
2:Simpl in H'3; Inversion H'3; Auto.
2:Simpl in H1; Simpl in H0.
2:Simpl.
2:Apply reduceskip; Auto.
2:Apply H'0; Auto.
2:Apply canonical_imp_canonical with a := a0; Auto.
2:Apply eqTerm_trans with y := (multTerm (invTerm T1) a); Auto.
2:Rewrite H0.
2:Apply eqTerm_trans
       with y := (multTerm (invTerm T1) (multTerm (invTerm T1) a0)); Auto.
2:Apply eqTerm_trans
       with y := (multTerm (multTerm (invTerm T1) (invTerm T1)) a0); Auto.
Intros a b nZb p0 q0 r H' H'0 H'1 r0; Case r0; Simpl; Auto.
Intros H'2 H'3; Inversion H'3.
Intros a0 l H'2 H'3.
Cut (canonical l); Try Apply canonical_imp_canonical with a := a0; Auto;
 Intros C0.
Apply reducetop with b := b nZb := nZb q := q0; Auto.
Apply divP_eqM with a := a; Auto.
Rewrite (pX_invl a (multTerm (invTerm T1) a0) p0 (multm_lm (invTerm T1) l));
 Auto.
Apply eqM_sym; Apply eqM_trans with y := (multTerm T1 a0); Auto.
Apply canonical_nzeroP with p := l; Auto.
Cut (canonical (pX b q0)); Try Apply inPolySet_imp_canonical with L := Q; Auto;
 Intros C1.
Cut (canonical q0); Try Apply canonical_imp_canonical with a := b; Auto;
 Intros C2.
Cut (canonical (pX a p0)); [Intros C3 | Idtac]; Auto.
Cut (canonical p0); Try Apply canonical_imp_canonical with a := a; Auto;
 Intros C4.
Cut (divP a0 b); [Intros divP0 | Idtac].
Apply eqp_trans with y := (multm_lm (invTerm T1) (spminusf a b nZb p0 q0)).
Apply eqp_sym;
 Apply eqp_trans
      with y :=
           (spminusf
              (multTerm (invTerm T1) a) b nZb (multm_lm (invTerm T1) p0) q0);
 Auto.
Apply eqp_sym.
Apply spminusf_multTerm; Auto.
Apply eqp_trans with y := (spminusf a0 b nZb (multm_lm (invTerm T1) p0) q0);
 Auto.
Apply eqTerm_spminusf_com; Auto.
Rewrite (pX_invl a (multTerm (invTerm T1) a0) p0 (multm_lm (invTerm T1) l));
 Auto.
Apply eqTerm_trans with (multTerm (multTerm (invTerm T1) (invTerm T1)) a0); Auto.
Apply divP_eqM with a := a; Auto.
Apply nzeroP_comp_eqTerm with a := (invTerm a); Auto.
Apply nZero_invTerm_nZero; Auto.
Apply canonical_nzeroP with p := p0; Auto.
Apply eqTerm_trans with (invTerm (multTerm T1 a)); Auto.
Apply eqp_spminusf_com; Auto.
Rewrite (pX_invr p0 (multm_lm (invTerm T1) l) a (multTerm (invTerm T1) a0));
 Auto.
Apply eqp_trans with y := (multm_lm (multTerm (invTerm T1) (invTerm T1)) l);
 Auto.
Apply eqp_sym; Auto.
Apply multm_lm_comp; Auto.
Apply divP_eqM with a := a; Auto.
Rewrite (pX_invl a (multTerm (invTerm T1) a0) p0 (multm_lm (invTerm T1) l));
 Auto.
Apply eqM_sym; Apply eqM_trans with y := (multTerm T1 a0); Auto.
Apply nzeroP_comp_eqTerm with a := (invTerm a); Auto.
Apply nZero_invTerm_nZero; Auto.
Apply canonical_nzeroP with p := p0; Auto.
Rewrite (pX_invl a (multTerm (invTerm T1) a0) p0 (multm_lm (invTerm T1) l));
 Auto.
Apply eqTerm_trans with (multTerm (invTerm (invTerm T1)) a0); Auto.
Rewrite H'3; Change (canonical (multm_lm (invTerm T1) (cons a0 l))); Auto.
Qed.

Theorem reduce_multm_invr:
  (Q:(list poly))
  (p, q:(list Term)) (reduce Q (multm_lm (invTerm T1) p) q) -> (canonical p) ->
  (reduce Q p (multm_lm (invTerm T1) q)).
Intros Q p q H' H'0.
Apply reduce_multm_inv_lem with p := (multm_lm (invTerm T1) p); Auto.
Qed.

Theorem ltP_order_comp:
  (a, b, c:Term) (p, q:(list Term)) (ltP (pX b p) (pX a q)) -> (order a c) ->
  (order b c).
Intros a b c p q H1; Inversion H1.
Simpl in H0.
Intros; Apply order_trans with y := a; Auto.
Simpl in H3; Intros; Apply order_eqM_r with a := a; Auto.
Apply eqM_sym; Auto.
Qed.
Hints Resolve ltP_order_comp.

Theorem eqp_invT1_pO_is_pO:
  (p:(list Term)) (eqP (multm_lm (invTerm T1) p) pO) ->(eqP p pO).
Intros p; Case p; Simpl; Auto.
Intros a l H'; Inversion H'.
Qed.

Theorem reduce_imp_reduceplus:
  (Q:(list poly)) (p, q:(list Term)) (reduce Q p q) ->(reduceplus Q p q).
Intros Q p q H'.
Apply Rstar_n with y := q; Auto.
Qed.
Hints Resolve reduce_imp_reduceplus.

Theorem rep_plus_reduce:
  (Q:(list poly))
  (p, q:(list Term))
  (s:(list Term)) (reduce Q (pluspf p q) s) -> (canonical p) -> (canonical q) ->
  (Ex [p1:(list Term)]
  (Ex [q1:(list Term)]
  (reduceplus Q p p1) /\ ((reduceplus Q q q1) /\ (eqP s (pluspf p1 q1))))).
Intros Q p q; Pattern p q.
Apply Opm_ind; Intros; Auto.
Exists pO; Exists s; Split; [Idtac | Split]; Auto.
Apply reduce_imp_reduceplus.
Apply reduce_eqp_com with p := (pluspf pO p0) q := s; Auto.
Apply eqp_sym; Auto.
Exists s; Exists pO; Split; [Idtac | Split]; Auto.
Apply reduce_imp_reduceplus.
Apply reduce_eqp_com with p := (pluspf p0 pO) q := s; Auto.
Apply eqp_sym; Auto.
Cut (canonical p0);
 [Intros C0 | Apply canonical_imp_canonical with a := a; Auto].
Cut (canonical q0);
 [Intros C1 | Apply canonical_imp_canonical with a := b; Auto].
Cut ~ (zeroP a); [Intros nZ1 | Apply canonical_nzeroP with p := p0; Auto].
Cut ~ (zeroP b); [Intros nZ2 | Apply canonical_nzeroP with p := q0; Auto].
Inversion H1.
Cut (canonical (pX b0 q1));
 [Intros C2 | Apply inPolySet_imp_canonical with L := Q; Auto].
Cut (canonical q1);
 [Intros C3 | Apply canonical_imp_canonical with a := b0; Auto].
Exists (pX a p0); Exists (spminusf b b0 nZb q0 q1); Split; [Idtac | Split]; Auto.
Apply reduce_imp_reduceplus.
Apply reducetop with b := b0 nZb := nZb q := q1; Auto.
Cut (eqTerm a0 b); [Intros eqb | Idtac].
Apply divP_eqM with a := a0; Auto.
Rewrite <- (pluspf_inv2_eq a b p0 q0) in H4; Auto.
Rewrite (pX_invl a0 b p1 (pluspf (pX a p0) q0)); Auto.
Apply eqp_trans with y := (spminusf a0 b0 nZb p1 q1).
Apply eqp_sym; Auto.
Rewrite <- (pluspf_inv2_eq a b p0 q0) in H4; Auto.
Rewrite (pX_invl a0 b p1 (pluspf (pX a p0) q0)); Auto.
Rewrite (pX_invr p1 (pluspf (pX a p0) q0) a0 b); Auto.
Apply spminusf_plusTerm_r; Auto.
Rewrite <- (pX_invl a0 b p1 (pluspf (pX a p0) q0)); Auto.
Elim (H q1);
 [Intros p2 E; Elim E; Intros q2 E0; Elim E0; Intros H'4 H'5; Elim H'5;
   Intros H'6 H'7; Clear H'5 E0 E | Idtac | Idtac | Idtac]; Auto.
Exists p2; Exists (pX b q2); Split; [Idtac | Split]; Auto.
Apply eqp_trans with y := (pX a0 (pluspf p2 q2)); Auto.
Rewrite <- (pluspf_inv2_eq a b p0 q0) in H4; Auto.
Rewrite (pX_invl a0 b p1 (pluspf (pX a p0) q0)); Auto.
Rewrite pluspf_inv2_eqa; Auto.
Apply ltP_pX_canonical; Auto.
Apply canonical_reduceplus with Q := Q p := (pX a p0); Auto.
Elim (order_reduceplus Q (pX a p0) p2);
 [Intros H'5 | Intros H'5 | Idtac | Idtac]; Auto.
Apply ltP_trans with y := (pX a p0); Auto.
Apply ltp_eqp_comp with p := (pX a p0) q := (pX b pO); Auto.
Apply ltP_pX_canonical; Auto.
Apply canonical_reduceplus with Q := Q p := q0; Auto.
Elim (order_reduceplus Q q0 q2); [Intros H'5 | Intros H'5 | Idtac | Idtac]; Auto.
Apply ltP_trans with y := q0; Auto.
Apply ltp_eqp_comp with p := q0 q := (pX b pO); Auto.
Apply reduce_eqp_com with p := p1 q := q1; Auto.
Apply canonical_imp_canonical with a := a0; Auto.
Rewrite H4; Auto.
Rewrite <- (pluspf_inv2_eq a b p0 q0) in H4; Auto.
Rewrite (pX_invr p1 (pluspf (pX a p0) q0) a0 b); Auto.
Rewrite <- (pluspf_inv1_eq a b p0 q0) in H1.
Cut (canonical p0);
 [Intros C0 | Apply canonical_imp_canonical with a := a; Auto].
Cut (canonical q0);
 [Intros C1 | Apply canonical_imp_canonical with a := b; Auto].
Cut ~ (zeroP a); [Intros nZ1 | Apply canonical_nzeroP with p := p0; Auto].
Cut ~ (zeroP b); [Intros nZ2 | Apply canonical_nzeroP with p := q0; Auto].
Inversion H1.
Simpl in H5; Simpl in H4.
Cut (canonical (pX b0 q1));
 [Intros C2 | Apply inPolySet_imp_canonical with L := Q; Auto].
Cut (canonical q1);
 [Intros C3 | Apply canonical_imp_canonical with a := b0; Auto].
Cut ~ (zeroP b0); [Intros nZ3 | Apply canonical_nzeroP with p := q1; Auto].
Exists (spminusf a b0 nZb p0 q1); Exists (pX b q0); Split; [Idtac | Split]; Auto.
Apply reduce_imp_reduceplus; Auto.
Apply reducetop with b := b0 nZb := nZb q := q1; Auto.
Rewrite <- H4; Auto.
Apply eqp_trans with y := (spminusf a0 b0 nZb p1 q1).
Apply eqp_sym; Auto.
Rewrite H5; Rewrite <- H4; Auto.
Simpl in H5; Simpl in H4.
Rewrite H5 in H6.
Elim (H q1);
 [Intros p2 E; Elim E; Intros q2 E0; Elim E0; Intros H'4 H'5; Elim H'5;
   Intros H'6 H'7; Clear H'5 E0 E | Idtac | Idtac | Idtac]; Auto.
Exists (pX a p2); Exists q2; Split; [Idtac | Split]; Auto.
Apply eqp_trans with y := (pX a (pluspf p2 q2)); Auto.
Apply eqpP1; Auto.
Rewrite <- H4; Auto.
Rewrite pluspf_inv1_eqa; Auto.
Apply ltP_pX_canonical; Auto.
Apply canonical_reduceplus with Q := Q p := p0; Auto.
Elim (order_reduceplus Q p0 p2); [Intros H'14 | Intros H'14 | Idtac | Idtac];
 Auto.
Apply ltP_trans with y := p0; Auto.
Apply ltp_eqp_comp with p := p0 q := (pX a pO); Auto.
Apply ltP_pX_canonical; Auto.
Apply canonical_reduceplus with Q := Q p := (pX b q0); Auto.
Elim (order_reduceplus Q (pX b q0) q2);
 [Intros H'14 | Intros H'14 | Idtac | Idtac]; Auto.
Apply ltP_trans with y := (pX b q0); Auto.
Apply ltp_eqp_comp with p := (pX b q0) q := (pX a pO); Auto.
Auto.
Cut (canonical p0);
 [Intros C0 | Apply canonical_imp_canonical with a := a; Auto].
Cut (canonical q0);
 [Intros C1 | Apply canonical_imp_canonical with a := b; Auto].
Cut ~ (zeroP a); [Intros nZ1 | Apply canonical_nzeroP with p := p0; Auto].
Cut ~ (zeroP b); [Intros nZ2 | Apply canonical_nzeroP with p := q0; Auto].
Case (zeroP_dec (plusTerm a b)); Intros Z0.
Elim (H s);
 [Intros p1 E; Elim E; Intros q1 E0; Elim E0; Intros H'3 H'5; Elim H'5;
   Intros H'6 H'7; Clear H'5 E0 E | Idtac | Idtac | Idtac]; Auto.
Exists (pX a p1); Exists (pX b q1); Split; [Idtac | Split]; Auto.
Apply eqp_trans with y := (pluspf p1 q1); Auto.
Rewrite (pluspf_inv3a_eq a b p0 q0); Auto.
Rewrite <- (pluspf_inv3b_eq a b p0 q0) in H1; Auto.
Inversion H1.
Simpl in H5; Simpl in H4.
Cut (canonical (pX b0 q1));
 [Intros C2 | Apply inPolySet_imp_canonical with L := Q; Auto].
Cut (canonical q1);
 [Intros C3 | Apply canonical_imp_canonical with a := b0; Auto].
Cut (divP a b0); [Intros div0 | Idtac].
Cut (divP b b0); [Intros div1 | Idtac].
Exists (spminusf a b0 nZb p0 q1); Exists (spminusf b b0 nZb q0 q1); Split;
 [Idtac | Split]; Auto.
Apply reduce_imp_reduceplus; Auto.
Apply reducetop with b := b0 nZb := nZb q := q1; Auto.
Apply reduce_imp_reduceplus; Auto.
Apply reducetop with b := b0 nZb := nZb q := q1; Auto.
Apply eqp_trans with y := (spminusf a0 b0 nZb p1 q1).
Apply eqp_sym; Auto.
Rewrite H4; Rewrite H5.
Apply spminusf_plusTerm; Auto.
Apply divP_eqM with a := (plusTerm a b); Auto.
Rewrite <- H4; Auto.
Apply divP_eqM with a := (plusTerm a b); Auto.
Rewrite <- H4; Auto.
Simpl in H5; Simpl in H4.
Rewrite H5 in H6.
Elim (H q1);
 [Intros p2 E; Elim E; Intros q2 E0; Elim E0; Intros H'3 H'5; Elim H'5;
   Intros H'6 H'7; Clear H'5 E0 E | Idtac | Idtac | Idtac]; Auto.
Exists (pX a p2); Exists (pX b q2); Split; [Idtac | Split]; Auto.
Apply eqp_trans with y := (pX a0 (pluspf p2 q2)); Auto.
Rewrite H4; Auto.
Qed.

Theorem reduceplus_multm:
  (Q:(list poly))
  (a:Term)
  (p, q:(list Term)) (reduceplus Q p q) -> (canonical p) -> ~ (zeroP a) ->
  (reduceplus Q (multm_lm a p) (multm_lm a q)).
Intros Q a p q H'; Elim H'; Auto.
Intros x y z H'0 H'1 H'2 H'3 H'4.
Apply Rstar_n with y := (multm_lm a y); Auto.
Apply H'2; Auto.
Apply canonical_reduce with Q := Q p := x; Auto.
Qed.

Lemma reduceplus_multm_invr_lem:
  (Q:(list poly)) (p, q:(list Term)) (reduceplus Q p q) ->
  (r:(list Term)) (eqP p (multm_lm (invTerm T1) r)) -> (canonical r) ->
  (reduceplus Q r (multm_lm (invTerm T1) q)).
Intros Q p q H'; Elim H'; Auto.
Intros x y H'0 r H'1 H'2.
Apply Rstar_0; Auto.
Apply eqp_trans with y := (multm_lm (invTerm T1) x); Auto.
Apply eqp_trans with y := (multm_lm (invTerm T1) (multm_lm (invTerm T1) r));
 Auto.
Apply eqp_trans with y := (multm_lm (multTerm (invTerm T1) (invTerm T1)) r);
 Auto.
Apply eqp_trans with y := (multm_lm T1 r); Auto.
Apply eqp_sym; Auto.
Apply eqp_sym; Auto.
Intros x y z H'0 H'1 H'2 r H'3 H'4.
Apply Rstar_n with y := (multm_lm (invTerm T1) y); Auto.
Apply reduce_multm_invr; Auto.
Apply reduce_eqp_com with p := x q := y; Auto.
Apply eqp_imp_canonical with p := (multm_lm (invTerm T1) r); Auto.
Apply eqp_sym; Auto.
Apply reduceplus_multm; Auto.
Apply canonical_reduce with Q := Q p := x; Auto.
Apply eqp_imp_canonical with p := (multm_lm (invTerm T1) r); Auto.
Apply eqp_sym; Auto.
Qed.

Theorem reduceplus_multm_invr:
  (Q:(list poly))
  (p, q:(list Term))
  (reduceplus Q (multm_lm (invTerm T1) p) q) -> (canonical p) ->
  (reduceplus Q p (multm_lm (invTerm T1) q)).
Intros Q p q H' H'0.
Apply reduceplus_multm_invr_lem with p := (multm_lm (invTerm T1) p); Auto.
Qed.

Theorem rep_minus_reduce:
  (Q:(list poly))
  (p, q:(list Term))
  (s:(list Term))
  (reduce Q (minuspf p q) s) -> (canonical p) -> (canonical q) ->
  (Ex [p1:(list Term)]
  (Ex [q1:(list Term)]
  (reduceplus Q p p1) /\ ((reduceplus Q q q1) /\ (eqP s (minuspf p1 q1))))).
Intros Q p q s H' H'0 H'1.
Elim (rep_plus_reduce Q p (multm_lm (invTerm T1) q) s);
 [Intros p1 E; Elim E; Intros q1 E0; Elim E0; Intros H'9 H'10; Elim H'10;
   Intros H'11 H'12; Clear H'10 E0 E | Idtac | Idtac | Idtac]; Auto.
Exists p1; Exists (multm_lm (invTerm T1) q1); Split; [Idtac | Split]; Auto.
Apply reduceplus_multm_invr; Auto.
Cut (canonical p1); [Intros Cp1 | Idtac].
Cut (canonical q1); [Intros Cq1 | Idtac].
Apply eqp_trans with y := (pluspf p1 q1); Auto.
Apply eqp_trans
     with y := (pluspf p1 (multm_lm (invTerm T1) (multm_lm (invTerm T1) q1)));
 Auto.
Apply eqp_trans
     with y := (pluspf p1 (multm_lm (multTerm (invTerm T1) (invTerm T1)) q1));
 Auto.
Apply eqp_trans with y := (pluspf p1 (multm_lm T1 q1)); Auto.
Apply eqp_sym; Auto.
Apply eqp_sym; Auto.
Apply canonical_reduceplus with Q := Q p := (multm_lm (invTerm T1) q); Auto.
Apply canonical_reduceplus with Q := Q p := p; Auto.
Apply reduce_eqp_com with p := (minuspf p q) q := s; Auto.
Qed.

Theorem rep_plus_zero_reduce:
  (Q:(list poly)) (s, t:(list Term)) (reduceplus Q s t) -> (canonical s) ->
  (p, q:(list Term))
  (eqP s (pluspf p q)) -> (canonical p) -> (canonical q) -> (eqP t pO) ->
  (Ex [r1:(list Term)]
  (reduceplus Q p r1) /\ (reduceplus Q q (multm_lm (invTerm T1) r1))).
Intros Q s t H'; Elim H'.
Intros x y eqxy Op p q H'0 H'1 H'2 H'3; Exists p; Split; Auto.
Apply Rstar_0; Auto.
Cut (canonical y); [Intros Cx | Idtac].
Apply eqp_trans with y := (pluspf pO (multm_lm (invTerm T1) p)); Auto.
Apply eqp_trans with y := (pluspf y (multm_lm (invTerm T1) p)); Auto.
Apply eqp_trans with y := (pluspf x (multm_lm (invTerm T1) p)); Auto.
Apply eqp_sym;
 Apply eqp_trans with y := (pluspf (pluspf p q) (multm_lm (invTerm T1) p)); Auto.
Apply eqp_trans with y := (pluspf (pluspf q p) (multm_lm (invTerm T1) p)); Auto.
Apply eqp_trans with y := (pluspf q (pluspf p (multm_lm (invTerm T1) p))); Auto.
Apply eqp_sym; Apply eqp_trans with y := (pluspf q (minuspf p p)); Auto.
Apply eqp_sym; Apply eqp_trans with y := (pluspf q pO); Auto.
Apply eqp_imp_canonical with p := x; Auto.
Intros x y z H'0 H'1 H'2 H'3 p q H'4 H'5 H'6 H'7.
Elim (rep_plus_reduce Q p q y);
 [Intros p1 E; Elim E; Intros q1 E0; Elim E0; Intros H'15 H'16; Elim H'16;
   Intros H'17 H'18; Try Exact H'17; Clear H'16 E0 E | Idtac | Idtac | Idtac];
 Auto.
LApply H'2;
 [Intros H'8; Elim (H'8 p1 q1);
   [Intros r1 E; Elim E; Intros H'16 H'19; Clear E H'2 | Clear H'2 | Clear H'2 |
     Clear H'2 | Clear H'2] | Clear H'2]; Auto.
Exists r1; Split; Auto.
Apply reduceplus_trans with y := p1; Auto.
Apply reduceplus_trans with y := q1; Auto.
Apply canonical_reduceplus with Q := Q p := p; Auto.
Apply canonical_reduceplus with Q := Q p := q; Auto.
Apply eqp_imp_canonical with p := (pluspf p1 q1); Auto.
Apply eqp_sym; Auto.
Apply canonical_pluspf.
Apply canonical_reduceplus with Q := Q p := p; Auto.
Apply canonical_reduceplus with Q := Q p := q; Auto.
Apply reduce_eqp_com with p := x q := y; Auto.
Qed.

Theorem red_plus_zero_reduce:
  (Q:(list poly))
  (p, q:(list Term))
  (canonical p) -> (canonical q) -> (reduceplus Q (pluspf p q) pO) ->
  (Ex [r1:(list Term)]
  (reduceplus Q p r1) /\ (reduceplus Q q (multm_lm (invTerm T1) r1))).
Intros Q p q H' H'0 H'1.
Apply rep_plus_zero_reduce with s := (pluspf p q) t := pO; Auto.
Qed.

Theorem red_minus_zero_reduce:
  (Q:(list poly))
  (p, q:(list Term))
  (canonical p) -> (canonical q) -> (reduceplus Q (minuspf p q) pO) ->
  (Ex [r1:(list Term)] (reduceplus Q p r1) /\ (reduceplus Q q r1)).
Intros Q p q H' H'0 H'1.
Elim (red_plus_zero_reduce Q p (multm_lm (invTerm T1) q));
 [Intros r1 E; Elim E; Intros H'8 H'9; Clear E | Idtac | Idtac | Idtac]; Auto.
Exists r1; Split; Auto.
Apply reduceplus_eqp_com
     with p := q q := (multm_lm (invTerm T1) (multm_lm (invTerm T1) r1)); Auto.
Apply reduceplus_multm_invr; Auto.
Apply eqp_trans with y := (multm_lm (multTerm (invTerm T1) (invTerm T1)) r1);
 Auto.
Apply eqp_sym; Auto.
Apply reduceplus_eqp_com with p := (minuspf p q) q := pO; Auto.
Qed.

Theorem reduce_plus_top:
  (Q:(list poly)) (r:(list Term)) (canonical r) ->
  (a, b:Term)
  (nZb:~ (zeroP b))
  (p, q:(list Term))
  (inPolySet (pX b q) Q) -> (canonical (pX a p)) -> (divP a b) ->
  (ex
     (list Term)
     [s:(list Term)]
     (reduceplus Q (pluspf (pX a p) r) s) /\
     (reduceplus Q (pluspf (spminusf a b nZb p q) r) s)).
Intros Q r; Elim r; Auto.
Intros H' a b nZb p q H'0 H'1 H'2; Exists (spminusf a b nZb p q); Split; Auto.
Apply reduceplus_eqp_com with p := (pX a p) q := (spminusf a b nZb p q); Auto.
Apply reduce_imp_reduceplus; Auto.
Apply reducetop with b := b nZb := nZb q := q; Auto.
Apply eqp_sym; Auto.
Clear r; Intros c r H' H'0 a b nZb p q H'1 H'2 H'3.
Cut (canonical p); [Intros C0 | Apply canonical_imp_canonical with a := a; Auto].
Cut (canonical (pX b q));
 [Intros C1 | Apply inPolySet_imp_canonical with L := Q; Auto].
Cut (canonical q); [Intros C2 | Apply canonical_imp_canonical with a := b; Auto].
Cut (canonical r); [Intros C3 | Apply canonical_imp_canonical with a := c; Auto].
Cut ~ (zeroP a); [Intros nZ1 | Apply canonical_nzeroP with p := p; Auto].
Cut ~ (zeroP c); [Intros nZ3 | Apply canonical_nzeroP with p := r; Auto].
Case (order_dec a c); [Intros temp; Case temp; Clear temp | Idtac]; Intros test.
LApply H';
 [Intros H'4; Elim (H'4 a b nZb p q);
   [Intros s E; Elim E; Intros H'13 H'14; Clear E H' | Clear H' | Clear H' |
     Clear H'] | Clear H']; Auto.
Exists (pX c s); Split; Auto.
Change (reduceplus Q (pluspf (pX a p) (pX c r)) (pX c s)).
Rewrite <- (pluspf_inv2_eq a c p r); Auto.
Change (reduceplus Q (pluspf (spminusf a b nZb p q) (pX c r)) (pX c s)).
Cut (canonical (pX a (spminusf a b nZb p q))); Auto.
Generalize H'14; Case (spminusf a b nZb p q); Simpl; Auto.
Intros H' H'5.
Apply reduceplus_eqp_com with p := (pX c (pluspf pO r)) q := (pX c s); Auto.
Apply eqp_trans with y := (pX c r); Auto.
Apply eqp_sym; Auto.
Intros t l H' H'5.
Change (reduceplus Q (pluspf (pX t l) (pX c r)) (pX c s)).
Rewrite <- (pluspf_inv2_eq t c l r); Auto.
Apply reduceplus_skip; Auto.
Rewrite (pluspf_inv2_eq t c l r); Auto.
Apply canonical_pluspf; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply order_trans with y := a; Auto.
Apply canonical_imp_order with l := l; Auto.
Apply order_trans with y := a; Auto.
Apply canonical_imp_order with l := l; Auto.
Exists (pluspf (spminusf a b nZb p q) (cons c r)); Split; Auto.
Apply reduce_imp_reduceplus.
Change (reduce
          Q (pluspf (pX a p) (pX c r)) (pluspf (spminusf a b nZb p q) (pX c r))).
Rewrite <- (pluspf_inv1_eq a c p r); Auto.
Apply reducetop with b := b nZb := nZb q := q; Auto.
Cut (divP c b); [Intros divc | Apply divP_eqM with a := a]; Auto.
Cut (canonical (pX c (spminusf a b nZb p q)));
 [Intros Ca | Apply canonical_pX_eqM with a := a]; Auto.
Case (zeroP_dec (plusTerm a c)); Intros Z0.
Exists (pluspf p r); Split; Auto.
Change (reduceplus Q (pluspf (pX a p) (pX c r)) (pluspf p r)).
Rewrite <- (pluspf_inv3a_eq a c p r); Auto.
Change (reduceplus Q (pluspf (spminusf a b nZb p q) (pX c r)) (pluspf p r)).
Rewrite <- pluspf_inv2_eqa; Auto.
Apply reduce_imp_reduceplus.
Apply reducetop with b := b nZb := nZb q := q; Auto.
Apply eqp_trans with y := (pluspf (spminusf a b nZb p q) (spminusf c b nZb r q));
 Auto.
Apply eqp_sym; Auto.
Apply spminusf_plusTerm_z; Auto.
Exists (spminusf (plusTerm a c) b nZb (pluspf p r) q); Split; Auto.
Change (reduceplus
          Q (pluspf (pX a p) (pX c r))
          (spminusf (plusTerm a c) b nZb (pluspf p r) q)).
Rewrite <- (pluspf_inv3b_eq a c p r); Auto.
Apply reduce_imp_reduceplus.
Apply reducetop with b := b nZb := nZb q := q; Auto.
Change (reduceplus
          Q (pluspf (spminusf a b nZb p q) (pX c r))
          (spminusf (plusTerm a c) b nZb (pluspf p r) q)).
Rewrite <- (pluspf_inv2_eqa c (spminusf a b nZb p q) r); Auto.
Apply reduce_imp_reduceplus.
Apply reducetop with b := b nZb := nZb q := q; Auto.
Apply eqp_trans with y := (pluspf (spminusf a b nZb p q) (spminusf c b nZb r q));
 Auto.
Apply eqp_sym; Auto.
Qed.

Theorem one_plus_reduceplus:
  (Q:(list poly)) (p1, p2:(list Term)) (reduce Q p1 p2) ->
  (r:(list Term)) (canonical p1) -> (canonical r) ->
  (Ex [s:(list Term)]
  (reduceplus Q (pluspf p1 r) s) /\ (reduceplus Q (pluspf p2 r) s)).
Intros Q p1 p2 H'; Elim H'; Auto.
Intros a b nZb p q r H'0 H'1 H'2 r0 H'3 H'4.
LApply (reduce_plus_top Q r0);
 [Intros H'7; Elim (H'7 a b nZb p q);
   [Intros s E; Elim E; Intros H'16 H'17; Clear E | Idtac | Idtac | Idtac] |
   Idtac]; Auto.
Exists s; Split; Auto.
Cut (canonical (pX b q));
 [Intros C0 | Apply inPolySet_imp_canonical with L := Q; Auto].
Apply reduceplus_eqp_com with p := (pluspf (spminusf a b nZb p q) r0) q := s;
 Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_imp_canonical with p := (spminusf a b nZb p q); Auto.
Intros a b p q H'0 H'1 H'2 r H'3.
Cut (canonical p); [Intros C0 | Apply canonical_imp_canonical with a := a; Auto].
Cut ~ (zeroP a); [Intros Z0 | Apply canonical_nzeroP with p := p; Auto].
Cut (canonical (pX a q)); [Intros C1 | Idtac].
Cut (canonical q); [Intros C2 | Apply canonical_imp_canonical with a := a; Auto].
Elim r; Auto.
Intros H'3; Exists (pX a q); Split; Auto.
Apply reduceplus_eqp_com with p := (pX a p) q := (pX a q); Auto.
Apply eqp_sym; Auto.
Apply reduceplus_eqp_com with p := (pX a q) q := (pX a q); Auto.
Apply eqp_trans with y := (pX b q); Auto.
Apply eqp_sym; Auto.
Clear r; Intros c r H'3 H'4.
Cut (canonical r); [Intros C3 | Apply canonical_imp_canonical with a := c; Auto].
Cut ~ (zeroP c); [Intros Z1 | Apply canonical_nzeroP with p := r; Auto].
Case (order_dec a c); [Intros temp; Case temp; Clear temp | Idtac]; Intros test.
Elim H'4;
 [Intros s E; Elim E; Intros H'7 H'8; Try Exact H'7; Clear E H'4 | Clear H'4];
 Auto.
Exists (pX c s); Split; Auto.
Rewrite <- pluspf_inv2_eqa; Auto.
Cut (canonical (pX b q)); [Intros C4 | Apply canonical_pX_eqM with a := a; Auto].
Rewrite <- pluspf_inv2_eqa; Auto.
Apply reduceplus_eqp_com with p := (pX c (pluspf (pX a q) r)) q := (pX c s);
 Auto.
Apply reduceplus_skip; Auto.
Apply reduceplus_eqp_com with p := (pluspf (pX b q) r) q := s; Auto.
Apply eqp_imp_canonical with p := (pX c (pX a q)); Auto.
Apply nzeroP_comp_eqTerm with a := a; Auto.
Elim (H'1 (pX c r));
 [Intros s E; Elim E; Intros H'9 H'10; Clear E | Idtac | Idtac]; Auto.
Cut (canonical (pX b q)); [Intros C4 | Apply canonical_pX_eqM with a := a; Auto].
Exists (pX a s); Split; Auto.
Rewrite <- pluspf_inv1_eq; Auto.
Apply reduceplus_eqp_com with p := (pluspf (pX a q) (pX c r)) q := (pX a s);
 Auto.
Rewrite <- pluspf_inv1_eq; Auto.
Apply nzeroP_comp_eqTerm with a := a; Auto.
Case (zeroP_dec (plusTerm a c)); Intros Z2.
Elim (H'1 r); [Intros s E; Elim E; Intros H'9 H'10; Clear E | Idtac | Idtac];
 Auto.
Exists s; Split; Auto.
Rewrite <- pluspf_inv3a_eq; Auto.
Apply reduceplus_eqp_com with p := (pluspf (pX a q) (pX c r)) q := s; Auto.
Rewrite <- pluspf_inv3a_eq; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_imp_canonical with p := (pX a q); Auto.
Elim (H'1 r); [Intros s E; Elim E; Intros H'9 H'10; Clear E | Idtac | Idtac];
 Auto.
Exists (pX (plusTerm a c) s); Split; Auto.
Rewrite <- pluspf_inv3b_eq; Auto.
Apply reduceplus_skip; Auto.
Rewrite pluspf_inv3b_eq; Auto.
Apply reduceplus_eqp_com
     with p := (pluspf (pX a q) (pX c r)) q := (pX (plusTerm a c) s); Auto.
Rewrite <- pluspf_inv3b_eq; Auto.
Apply reduceplus_skip; Auto.
Rewrite pluspf_inv3b_eq; Auto.
Apply eqp_pluspf_com; Auto.
Apply eqp_imp_canonical with p := (pX a q); Auto.
Apply canonical_reduce with Q := Q p := (pX a p); Auto.
Qed.

Theorem one_minus_reduceplus:
  (Q:(list poly)) (p1, p2:(list Term)) (reduce Q p1 p2) ->
  (r:(list Term)) (canonical p1) -> (canonical r) ->
  (Ex [s:(list Term)]
  (reduceplus Q (minuspf p1 r) s) /\ (reduceplus Q (minuspf p2 r) s)).
Intros Q p1 p2 H' r H'0 H'1.
LApply (one_plus_reduceplus Q p1 p2);
 [Intros H'5; Elim (H'5 (multm_lm (invTerm T1) r));
   [Intros s E; Elim E; Intros H'9 H'10; Clear E | Idtac | Idtac] | Idtac]; Auto.
Exists s; Split; Auto.
Apply reduceplus_eqp_com with p := (pluspf p1 (multm_lm (invTerm T1) r)) q := s;
 Auto.
Apply eqp_sym; Auto.
Cut (canonical p2); [Intros C0 | Idtac].
Apply reduceplus_eqp_com with p := (pluspf p2 (multm_lm (invTerm T1) r)) q := s;
 Auto.
Apply eqp_sym; Auto.
Apply canonical_reduce with Q := Q p := p1; Auto.
Qed.
Variable selectpoly:
         (Q:(list poly)) (r:(list Term)) (canonical r) ->
         {q:(list Term) | (reduce Q r q)}+{(irreducible Q r)}.

Definition s2p: poly ->(list Term).
Intros H'; Elim H'.
Intros x H'0; Exact x.
Defined.

Definition Reduce0:
  (Q:(list poly)) (p:poly){q:poly | (reducestar Q (s2p p) (s2p q))}.
(** I've tried this relizer it didn't work so I do it in a proof style
Realizer <(list Term)>rec reduce0 :: :: {sorder}
          [p:(list Term)]
           (<(list Term)>Case (selectpoly Q p) of [q:(list Term)](reduce0 q) p end). **)
Intros Q.
Intros p; Pattern p; Apply well_founded_induction with A := poly R := sltP; Auto.
Exact sltp_wf.
Intros x; Elim x; Simpl.
Intros p0 Ordp0 Induct; Try Assumption.
Case (selectpoly Q p0 Ordp0).
Intros H'; Elim H'.
Intros q0 reducep0q0; Try Assumption.
Cut (canonical q0); Auto.
Intros Ordq0; Try Assumption.
Generalize (Induct (exist (list Term) [p:(list Term)](canonical p) q0 Ordq0));
 Simpl; Auto.
Simpl.
Intros H'0; LApply H'0; [Intros H'1; Elim H'1; Clear H'0 | Clear H'0].
Intros r reducestarr; Try Assumption.
Exists r.
Apply reducestar0; Auto.
Inversion reducestarr.
Apply Rstar_n with y := q0; Auto.
Inversion reducestarr; Auto.
Apply ltP_reduce with Q := Q; Auto.
Apply canonical_reduce with Q := Q p := p0; Auto.
Intros H'; Exists (exist (list Term) [p:(list Term)](canonical p) p0 Ordp0);
 Auto.
Simpl; Auto.
Apply reducestar0; Auto; Auto.
Defined.

Definition mks: (p:(list Term)) (canonical p) ->poly.
Intros p H'; Exists p; Exact H'.
Defined.

Theorem reduce0_reducestar:
  (Q:(list poly)) (p:(list Term)) (canonical p) ->
  (Ex [t:(list Term)] (reducestar Q p t)).
Intros Q p H.
Generalize (Reduce0 Q (mks p H)).
Intros H'; Elim H'.
Simpl.
Intros x H'0; Exists (s2p x); Try Assumption.
Qed.

Lemma irreducible_inv_px_l:
  (Q:(list poly))
  (a:Term) (p:(list Term)) (canonical (pX a p)) -> (irreducible Q (pX a p)) ->
  (irreducible Q (pX a pO)).
Unfold irreducible.
Intros Q a p H' H'0 q; Red; Intros H'1.
Inversion H'1.
Simpl in H0; Simpl in H.
Specialize 1 H'0 with q := (spminusf a b nZb p q0); Intros H'2; Apply H'2;
 Clear H'2; Auto.
Apply reducetop with b := b nZb := nZb q := q0; Auto.
Rewrite <- H; Auto.
Simpl in H; Simpl in H0.
Rewrite H0 in H1.
Inversion H1; Auto.
Qed.

Lemma pO_irreducible: (Q:(list poly))(irreducible Q pO).
Unfold irreducible; Auto.
Intros Q q; Red; Intros H'; Inversion H'.
Qed.
Hints Resolve pO_irreducible.

Definition pickinSet:
  (a:Term) ~ (zeroP a) ->
  (Q:(list poly))
  {p:(list Term) | (pickinSetp a p Q)}+{
   (b:Term) (p, q:(list Term)) (inPolySet p Q) -> (divP a b) ->~ p=(pX b q)}.
Intros a Z0 Q; Elim Q.
Right; Intros b p q H'; Inversion H'.
Intros a0; Case a0.
Intros x; Case x.
Intros c l H'; Case H'; Auto.
Intros H'0; Left; Inversion H'0.
Exists x0; Auto.
Intros H'0; Right.
Intros b p q H'1; Inversion H'1.
Intros H'3; Red; Intros H'4.
LApply (H'0 b p q);
 [Intros H'7; LApply H'7; [Intros H'8; Clear H'7 | Clear H'7] | Idtac]; Auto.
Intros a1 l c l0 H'.
Case (divP_dec a a1); Auto.
Apply canonical_nzeroP with p := l; Auto.
Intros H'0; Left; Exists (pX a1 l); Auto.
Intros Hyp.
Case H'; Intros H'1.
Inversion H'1.
Left; Exists x0; Auto.
Right.
Intros b p q H'2; Inversion H'2; Auto.
Intros H'3; Red; Intros H'4; Apply Hyp.
Simpl in H2; Rewrite <- H2.
Injection H'4.
Intros H'5 H'6; Rewrite H'6; Auto.
Defined.

Definition Reduce1:
  (Q:(list poly)) (p:poly){q:poly | (reducestar Q (s2p p) (s2p q))}.
Intros Q p; Pattern p.
Apply well_founded_induction with A := poly R := sltP; Auto.
Exact sltp_wf.
Clear p.
Intros p; Elim p.
Intros p0; Elim p0.
Simpl.
Intros y H'; Try Assumption.
Exists (exist (list Term) [p:(list Term)](canonical p) pO y); Simpl; Auto.
Apply reducestar0; Auto.
Intros a p1 H' y H'0.
Cut ~ (zeroP a); [Intros Z0 | Idtac].
Case (pickinSet a Z0 Q); Intros Pi0.
Inversion Pi0.
Generalize H; Case x.
Intros H'1; Exists p.
Elim (pickin_is_pX a pO Q); [Intros b E; Elim E; Intros q E0; Clear E | Idtac];
 Auto.
Inversion E0.
Intros b q1 H'1.
Cut ~ (zeroP b); [Intros nZb | Auto].
Cut (inPolySet (pX b q1) Q).
Intros IP; Try Assumption.
Cut (divP a b).
Intros EqC; Try Assumption.
Cut (canonical (spminusf a b nZb p1 q1)).
Intros O1; Try Assumption.
LApply (H'0
          (exist
             (list Term) [p:(list Term)](canonical p) (spminusf a b nZb p1 q1)
             O1)); [Intros H'6; Try Exact H'6 | Idtac]; Auto.
Elim H'6.
Simpl.
Intros q2 H'5.
Exists q2.
Inversion H'5.
Apply reducestar0; Auto.
Apply Rstar_n with y := (spminusf a b nZb p1 q1); Auto.
Apply reducetop with b := b nZb := nZb q := q1; Auto.
Simpl; Apply ltP_reduce with Q := Q; Auto.
Apply reducetop with b := b nZb := nZb q := q1; Auto.
Apply canonical_spminusf_full; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Apply pick_inv_eqM with Q := Q q := q1; Auto.
Apply pick_inv_in with a := a; Auto.
Apply canonical_nzeroP with p := q1; Auto.
Apply inPolySet_imp_canonical with L := Q; Auto.
Apply pick_inv_in with a := a; Auto.
Simpl; Cut (canonical p1).
Intros H'2;
 LApply (H'0 (exist (list Term) [p:(list Term)](canonical p) p1 H'2));
 [Intros H'3; Try Exact H'3 | Idtac]; Auto.
Elim H'3.
Simpl; Intros x; Elim x; Simpl.
Intros p2 H'4 H'5; Cut (canonical (pX a p2)).
Intros H'6;
 Exists (exist (list Term) [p:(list Term)](canonical p) (pX a p2) H'6); Simpl;
 Auto.
Inversion H'5.
Apply reducestar0; Auto.
Unfold irreducible.
Intros q0; Red; Intros H'1; Inversion H'1.
Simpl in H3; Simpl in H4.
LApply (Pi0 b (pX b q1) q1);
 [Intros H'10; LApply H'10; [Intros H'11; Apply H'11; Clear H'10 | Clear H'10] |
   Idtac]; Auto.
Rewrite <- H3; Auto.
Simpl in H3; Simpl in H4.
Unfold irreducible in H0.
Apply (H0 q1); Auto.
Rewrite <- H4; Auto.
Apply ltP_pX_canonical; Auto.
Elim (order_reduceplus Q p1 p2); [Intros H'10 | Intros H'10 | Idtac | Idtac];
 Auto.
Apply ltP_trans with y := p1; Auto.
Apply ltp_eqp_comp with p := p1 q := (pX a pO); Auto.
Inversion H'5; Auto.
Simpl.
Apply ltP_refl_pX; Auto.
Apply canonical_imp_canonical with a := a; Auto.
Apply canonical_nzeroP with p := p1; Auto.
Defined.
End Sreduce.
