Section Wf_incl.
Require Import Buch.
Load "shyp_sreduce".
Load "mod_splus".
Variable nf:poly -> (list poly) ->poly.
Variable zerop:poly ->Prop.

Axiom not_wf_seq_decrS:
      (A:Set) (R:A -> A ->Prop) ~ (well_founded A R) ->
      (ex nat ->A [g:nat ->A] (x:nat)(R (g (S x)) (g x))).

Local addEnd: poly -> (list poly) ->(list poly) :=
   (addEnd Term Monome T2M zeroP zeroP_dec ltM).
Variable spO:poly.

Definition lastE: (list poly) ->poly.
Intros H'; Elim H'.
Exact spO.
Intros H'0 H'1; Case H'1.
Intros H'2; Exact H'0.
Intros H'2 H'3 H'4; Exact H'4.
Defined.

Theorem not_addEnd_nil: (a:poly) (p:(list poly))~ (addEnd a p)=(nil poly).
Intros a p; Elim p; Simpl; Auto with datatypes.
Qed.

Theorem lastE_is_add: (a:poly) (p:(list poly))(lastE (addEnd a p))=a.
Intros a p; Elim p; Simpl; Auto.
Intros a0 l; Generalize (not_addEnd_nil a l); Case (addEnd a l); Simpl; Auto.
Intros H'; Apply H' Orelse Case H'; Auto.
Qed.

Mutual Inductive RO: (list poly) -> (list poly) ->Prop :=
     RO0:
       (a:poly) (P:(list poly)) ~ (zerop (nf a P)) ->(RO (addEnd (nf a P) P) P) .

Theorem not_wf_RO_seq:
  ~ (well_founded (list poly) RO) ->
  (Ex [g:nat ->(list poly)] (x:nat)(RO (g (S x)) (g x))).
Intros; Apply (not_wf_seq_decrS (list poly) RO); Auto.
Qed.

Theorem RO_exist:
  (P, Q:(list poly)) (RO P Q) ->
  (Ex [a:poly] ~ (zerop (nf a Q)) /\ P=(addEnd (nf a Q) Q)).
Intros P Q H'; Inversion H'; Auto.
Exists a; Split; Auto.
Qed.
Require Import Monomials.
Variable d:nat.
Variable get_mon:(a:poly)(mon d).
Hypothesis get_is_correct:
           (a, b:poly)
           (P:(list poly)) ~ (zerop (nf a P)) -> (In b P) -> ~ (zerop b) ->
           ~ (mdiv d (get_mon b) (get_mon (nf a P))).
Require Import Nat_Sequences.
Require Import Dixon.

Theorem g_gives_nzero_addEnd:
  (g:nat ->(list poly)) (x:nat) (RO (g (S x)) (g x)) ->
  (Ex [a:poly] ~ (zerop (nf a (g x))) /\ (g (S x))=(addEnd (nf a (g x)) (g x))).
Intros g x H'; Inversion H'.
Exists a; Split; Auto.
Qed.

Theorem g_gives_incl:
  (g:nat ->(list poly)) (x:nat) (RO (g (S x)) (g x)) ->(incl (g x) (g (S x))).
Intros g x H'; Inversion H'; Auto.
Red.
Intros a0 H'0; Apply (addEnd_id2 ? ? T2M zeroP zeroP_dec ltM); Auto.
Qed.
Hints Resolve g_gives_incl.

Theorem in_trans:
  (U:Set) (a:U) (L1, L2:(list U)) (In a L1) -> (incl L1 L2) ->(In a L2).
Intros U a L1 L2 H' H'0; Auto.
Qed.

Theorem g_gives_nzero:
  (g:nat ->(list poly)) (x:nat) (RO (g (S x)) (g x)) ->~ (g (S x))=(nil poly).
Intros g x H'; Inversion H'.
Elim (g x); Simpl; Auto with datatypes.
Qed.

Theorem In_Last_comp: (L:(list poly)) ~ L=(nil poly) ->(In (lastE L) L).
Intros L; Elim L; Auto.
Intros H'; Apply H' Orelse Case H'; Auto.
Intros a l; Case l; Simpl; Auto with datatypes.
Qed.

Theorem g_gives_In:
  (g:nat ->(list poly)) ((x:nat)(RO (g (S x)) (g x))) ->
  (i, j:nat)(In (lastE (g (S i))) (g (S (S (plus j i))))).
Intros g H' i j; Elim j; Simpl; Auto.
Elim (g_gives_nzero_addEnd g (S i));
 [Intros a E; Elim E; Intros H'3 H'4; Try Exact H'3; Clear E | Idtac]; Auto.
Rewrite H'4.
Apply in_trans with U := poly L1 := (g (S i)); Auto.
Apply In_Last_comp; Auto.
Apply g_gives_nzero; Auto.
Rewrite <- H'4; Auto.
Intros n H'0.
Apply in_trans with U := poly L1 := (g (S (S (plus n i)))); Auto.
Qed.

Theorem g_gives_nzero_last_nf:
  (g:nat ->(list poly)) (x:nat) (RO (g (S x)) (g x)) ->
  (Ex [a:poly] ~ (zerop (nf a (g x))) /\ (lastE (g (S x)))=(nf a (g x))).
Intros g x H'.
Elim (g_gives_nzero_addEnd g x);
 [Intros a E; Elim E; Intros H'3 H'4; Clear E | Idtac]; Auto.
Exists a; Split; Auto.
Rewrite H'4; Auto.
Apply lastE_is_add; Auto.
Qed.

Theorem not_Dix_0:
  ~ (well_founded (list poly) RO) ->
  (Ex [g:nat ->(mon d)] (i, j:nat)~ (mdiv d (g i) (g (S (plus i j))))).
Intros H'.
Elim not_wf_RO_seq; [Intros g E | Idtac]; Auto.
Exists [n:nat](get_mon (lastE (g (S n)))).
Intros i j.
Elim (g_gives_nzero_last_nf g (S (plus i j)));
 [Intros a E0; Elim E0; Intros H'3 H'4; Clear E0 | Idtac]; Auto.
Rewrite H'4.
Apply get_is_correct; Auto.
Rewrite (plus_sym i j); Case j; Simpl; Auto.
Apply In_Last_comp; Auto.
Specialize 1 E with x := i; Intros H'0; Inversion H'0; Auto.
Case (g i); Simpl; Auto with datatypes.
Intros n.
Apply g_gives_In; Auto.
Elim (g_gives_nzero_last_nf g i);
 [Intros a0 E0; Elim E0; Intros H'5 H'6; Try Exact H'5; Clear E0 | Idtac]; Auto.
Rewrite H'6; Auto.
Qed.

Theorem wf_incl: (well_founded (list poly) RO).
Apply NNPP.
Red; Intros H'.
Elim not_Dix_0; [Intros g E | Idtac]; Auto.
Elim (Dixon2 d g); Intros i E0; Elim E0; Intros j E1; Elim E1; Intros H'2 H'3;
 Try Exact H'2; Clear E1 E0.
Specialize 2 E with i := i j := (minus j (S i)); Intros H'1;
 Apply H'1 Orelse Case H'1; Try Assumption; Clear H'1.
Change (mdiv d (g i) (g (plus (S i) (minus j (S i))))).
Rewrite (le_plus_minus_r (S i) j); Auto.
Qed.
End Wf_incl.

